import { world, BlockPermutation, system, EntityInventoryComponent, EntityColorComponent, ItemStack, Player, EntityIsTamedComponent, EntityHealthComponent, EntityItemComponent, EntityEquippableComponent, EquipmentSlot, WeatherType, MoonPhase, GameMode, MinecraftDimensionTypes, Entity } from '@minecraft/server';
import { ModalFormData, FormRejectError, FormCancelationReason, MessageFormData, ActionFormData } from '@minecraft/server-ui';

// Based off https://stackoverflow.com/a/1349426/688379
const RANDOM_ID_CHARACTERS = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789';
const makeRandomId = (length = 16) => {
    return [...new Array(length)].map(() => RANDOM_ID_CHARACTERS.charAt(Math.floor(Math.random() * RANDOM_ID_CHARACTERS.length))).join('');
};
const makeRandomNumber = (min, max) => {
    return min + Math.floor(Math.random() * (max - min + 1));
};
const makeRandomNumberDecimal = (min, max) => {
    return min + Math.random() * (max - min + 1);
};

const TRANSLATE = (key, ...params) => ({
    translate: key,
    with: params.length > 0 ? typeof params[0] === 'string' ? params : RAW_TEXT(...params) : undefined
});
const RAW_TEXT = (...rawText) => ({
    rawtext: rawText
});
const TEXT = (value) => ({
    text: value.toString()
});
const TEXT_OR_RAW_TEXT = (value) => {
    if (typeof value === 'string' || typeof value === 'number') {
        return TEXT(value);
    }
    return value;
};
const ITEM = (item) => {
    let key = item.typeId;
    let addName = false;
    if (item.hasComponent("display_name")) {
        addName = true;
    }
    // Removes minecraft prefix if found.
    if (key.startsWith("minecraft:")) {
        addName = true;
        key = key.substring("minecraft:".length);
    }
    // Moves spawn_egg postfix if found
    if (key.endsWith("_spawn_egg")) {
        addName = true;
        key = `spawn_egg.entity.${key.replace("_spawn_egg", "")}`;
    }
    let isBlock;
    try {
        BlockPermutation.resolve(item.typeId);
        isBlock = true;
        addName = true;
    }
    catch (e) {
        isBlock = false;
    }
    if (addName) {
        key = `${key}.name`;
    }
    if (isBlock) {
        key = `tile.${key}`;
    }
    else {
        key = `item.${key}`;
    }
    return TRANSLATE(key);
};
const ENTITY = (entity) => {
    let key = entity;
    if (key.startsWith("minecraft:")) {
        key = `entity.${key.substring("minecraft:".length)}.name`;
    }
    else {
        key = `entity.${key}.name`;
    }
    return TRANSLATE(key);
};
// export const blockTypeToString = (blockType: string | BlockType) => typeof blockType === 'string' ? blockType : blockType.id;
const debug = (what) => {
    const players = world.getPlayers({
        tags: ["debug"]
    });
    if (players.length > 0) {
        const text = typeof what === 'string' ? what : what();
        for (const player of players) {
            void player.dimension.runCommandAsync(`w ${player.name} debug: ${text}`);
        }
    }
};
const getDistance = (start, end) => {
    let x1 = start.x;
    let y1 = start.y;
    let z1 = start.z;
    let x2 = end.x;
    let y2 = end.y;
    let z2 = end.z;
    return Math.ceil(Math.sqrt(Math.pow(x2 - x1, 2) + Math.pow(y2 - y1, 2) + Math.pow(z2 - z1, 2)));
};
const getHorDistance = (start, end) => {
    let x1 = start.x;
    start.y;
    let z1 = start.z;
    let x2 = end.x;
    end.y;
    let z2 = end.z;
    return Math.ceil(Math.sqrt(Math.pow(x2 - x1, 2) + Math.pow(z2 - z1, 2)));
};
const getVerDistance = (start, end) => {
    start.x;
    let y1 = start.y;
    start.z;
    end.x;
    let y2 = end.y;
    end.z;
    return Math.ceil(y1 - y2 + 1);
};
const getDirection = (start, end) => {
    const xDiff = end.x - start.x;
    const yDiff = end.y - start.y;
    const zDiff = start.z - end.z;
    const isSouth = isFarAwayInDirection(zDiff);
    const isWest = isFarAwayInDirection(xDiff);
    const isAbove = isFarAwayInDirection(yDiff);
    const where = [];
    if (isSouth === undefined && isWest === undefined && isAbove === undefined) {
        return [0 /* NEXT_TO_YOU */];
    }
    if (isSouth) {
        where.push(1 /* NORTH */);
    }
    else if (isSouth === false) {
        where.push(2 /* SOUTH */);
    }
    if (isWest) {
        where.push(4 /* EAST */);
    }
    else if (isWest === false) {
        where.push(3 /* WEST */);
    }
    if (isAbove) {
        where.push(5 /* ABOVE */);
    }
    else if (isAbove === false) {
        where.push(6 /* BELOW */);
    }
    return where;
};
const isFarAwayInDirection = (diff) => {
    const howFar = 3;
    if (Math.abs(diff) > howFar) {
        return diff > 0;
    }
    // Means that is next to us in the direction
    return undefined;
};

const getEmptySlot = (container) => {
    for (let i = 0; i < container.size; ++i) {
        if (container.getItem(i) === undefined) {
            return i;
        }
    }
    return undefined;
};
const getSlotByCondition = (container, condition) => {
    for (let i = 0; i < container.size; ++i) {
        const item = container.getItem(i);
        if (item && condition(item)) {
            return i;
        }
    }
    return undefined;
};

const PROP = (...identifier) => `jig_ccomp:${identifier.join(':')}`;
const GET_RAW_PROP = (target, identifier) => {
    return target.getDynamicProperty(identifier);
};
const SET_RAW_PROP = (target, identifier, value) => {
    target.setDynamicProperty(identifier, value);
};
const GET_PROP = (target, identifier, defaultValue) => {
    const value = target.getDynamicProperty(identifier);
    if (typeof value === 'string') {
        return JSON.parse(value);
    }
    if (defaultValue) {
        return defaultValue;
    }
    const message = `Invalid identifier ${identifier}`;
    debug(message);
    throw new Error(message);
};
const SET_PROP = (target, identifier, value) => {
    target.setDynamicProperty(identifier, JSON.stringify(value));
};
// export const mutateDynamicProperty = <T>(target: WithDynamicProperty, identifier: string, mutator: DynamicPropertyMutator<T>) => {
//   const input = target.getDynamicProperty(identifier);
//   const output = mutator(typeof input === 'string' ? JSON.parse(input) : undefined);
//   target.setDynamicProperty(identifier, JSON.stringify(output));
// }
//
// export class DynamicPropertyContext<T> {
//   private _value: T;
//   private target: WithDynamicProperty;
//   private identifier: string;
//
//   constructor(target: WithDynamicProperty, identifier: string, defaultValue: T) {
//     this.target = target;
//     this.identifier = identifier;
//     const value = target.getDynamicProperty(identifier);
//     this._value = typeof value === 'string' ? JSON.parse(value) : undefined;
//   }
//
//   public get value() {
//     return this._value;
//   }
//
//   private set value(value: T) {
//     this._value = value;
//   }
//
//   mutate(mutator: DynamicPropertyMutator<T>) {
//     this.value = mutator(this.value);
//   }
//
//   commit() {
//     this.target.setDynamicProperty(this.identifier, JSON.stringify(this.value));
//   }
// }

// Add operations that we need to do over entities, but entities might not be loaded/online.
// Delay the operations for later.
const FUTURE_OPERATION_ID = 'jig_ccomp:future-operation';
const NO_OPERATION_ID = -1;
const scheduleOperation = (entity, operationRequest) => {
    const time = system.currentTick + operationRequest.timeFromNow;
    const operation = {
        id: makeRandomId(),
        operation: operationRequest.operation,
        params: operationRequest.params,
        timerId: NO_OPERATION_ID,
        time
    };
    operation.timerId = system.runTimeout(() => runOperation(entity, operation), operationRequest.timeFromNow);
    addOperation(entity, operation);
};
const cancelOperations = (entity) => {
    SET_PROP(entity, FUTURE_OPERATION_ID, getOperations(entity).map(operation => {
        if (operation.timerId !== NO_OPERATION_ID) {
            system.clearRun(operation.timerId);
        }
        operation.timerId = NO_OPERATION_ID;
        return operation;
    }));
};
const runPendingOperations = (entity) => {
    const operations = getOperations(entity).map(operation => {
        if (operation.timerId !== NO_OPERATION_ID) {
            system.clearRun(operation.timerId);
        }
        operation.timerId = system.runTimeout(() => runOperation(entity, operation), Math.max(1, operation.time - system.currentTick));
        return operation;
    });
    entity.setDynamicProperty(FUTURE_OPERATION_ID, JSON.stringify(operations));
};
const runOperation = async (entity, operation) => {
    if (entity.isValid()) {
        switch (operation.operation) {
            case "send-email-mob" /* COMPUTER_SEND_EMAIL_FROM_MOB */:
                const { operationSendEmailFromMob } = await Promise.resolve().then(function () { return Email; });
                operationSendEmailFromMob(entity, operation.params);
                break;
            case "enable-npc" /* PLAYER_ENABLE_NPC */:
                const { operationEnableNpc } = await Promise.resolve().then(function () { return Email; });
                operationEnableNpc(entity, operation.params);
                break;
        }
        removeOperation(entity, operation);
    }
};
const getOperations = (entity) => {
    return JSON.parse(entity.getDynamicProperty(FUTURE_OPERATION_ID) ?? '[]');
};
const addOperation = (entity, ...operation) => {
    const operations = getOperations(entity);
    operations.push(...operation);
    entity.setDynamicProperty(FUTURE_OPERATION_ID, JSON.stringify(operations));
};
const removeOperation = (entity, operation) => {
    const operations = getOperations(entity);
    entity.setDynamicProperty(FUTURE_OPERATION_ID, JSON.stringify(operations.filter(o => o.id !== operation.id)));
};

const NUMBERED_ENTITY_COUNT = 16;
const initComputers = () => {
    world.beforeEvents.itemUseOn.subscribe((event) => {
        const getAllValidItemTypes = () => {
            const itemTypes = [];
            VALID_COMPUTERS_TYPES.forEach(type => {
                if (type.endsWith('_lap')) {
                    // Generate numbered types for 'laptop'
                    for (let i = 0; i < NUMBERED_ENTITY_COUNT; i++) {
                        itemTypes.push(`${type}${i}_item`);
                    }
                }
                else {
                    // For other types, just add the base item type
                    itemTypes.push(`${type}_item`);
                }
            });
            return itemTypes;
        };
        const validItemTypes = getAllValidItemTypes();
        if (validItemTypes.includes(event.itemStack.typeId)) {
            event.cancel = true;
            system.runTimeout(() => {
                const targetLocation = {
                    x: event.block.location.x + event.faceLocation.x,
                    y: event.block.location.y + event.faceLocation.y,
                    z: event.block.location.z + event.faceLocation.z
                };
                const extractNumber = (typeId) => {
                    const match = typeId.match(/(\d+)_item$/);
                    return match ? match[1] : "";
                };
                const stripItemSuffixAndNumber = (typeId) => {
                    // Regular expression to match the "_item" suffix and any trailing digits
                    return typeId.replace(/(\d+)?_item$/, "");
                };
                const computer = event.block.dimension.spawnEntity(stripItemSuffixAndNumber(event.itemStack.typeId), targetLocation);
                computer.setRotation({
                    x: computer.getRotation().x,
                    y: (event.source.getRotation().y + 180) % 360,
                });
                const number = extractNumber(event.itemStack.typeId);
                if (number !== undefined && number !== "") {
                    computer.runCommand(`event entity @s jig_ccomp:color_${number}`);
                }
                // Hydrating a computer
                if (event.itemStack.getLore().length === 2) {
                    const computerId = event.itemStack.getLore()[1];
                    computer.setDynamicProperty('jig_ccomp:COMPUTER_ID', computerId);
                    if (GET_RAW_PROP(world, IDENTIFIER_NAME(computerId)) === undefined) {
                        // We lost the data. The computer has an id but not a name in the system.
                        // Mark as lost data and set the name from the lore
                        SET_RAW_PROP(world, IDENTIFIER_LOST_DATA(computerId), true);
                        SET_RAW_PROP(world, IDENTIFIER_NAME(computerId), event.itemStack.getLore()[0]);
                    }
                    SET_RAW_PROP(world, IDENTIFIER_CREATOR(computerId), event.source.name);
                    // No longer an item
                    SET_RAW_PROP(world, IDENTIFIER_ITEMIZED_ON(computerId), undefined);
                    if (GET_RAW_PROP(world, IDENTIFIER_TRIGGER_EMAIL_NOTIFICATION(computerId))) {
                        computer.triggerEvent('jig_ccomp:got_email');
                        SET_RAW_PROP(world, IDENTIFIER_TRIGGER_EMAIL_NOTIFICATION(computerId), undefined);
                    }
                    // Hydrate operations
                    addOperation(computer, ...GET_PROP(world, IDENTIFIER_SAVED_OPERATIONS(computerId), []));
                    SET_RAW_PROP(world, IDENTIFIER_SAVED_OPERATIONS(computerId), undefined);
                    runPendingOperations(computer);
                    const inventory = event.source.getComponent(EntityInventoryComponent.componentId);
                    if (inventory.container) {
                        for (let i = 0; i < inventory.container.size; ++i) {
                            const item = inventory.container.getItem(i);
                            if (item?.typeId === event.itemStack.typeId && item.getLore().length === 2 && item.getLore()[1] === computerId) {
                                inventory.container.setItem(i);
                                break;
                            }
                        }
                    }
                }
                else {
                    // New computer
                    new CraftComputer(computer).setCreator(event.source.name);
                    const inventory = event.source.getComponent(EntityInventoryComponent.componentId);
                    if (inventory.container) {
                        for (let i = 0; i < inventory.container.size; ++i) {
                            const item = inventory.container.getItem(i);
                            if (item?.typeId === event.itemStack.typeId && item.getLore().length === 0) {
                                // decrease one or remove if only one
                                if (item.amount >= 2) {
                                    item.amount -= 1;
                                    inventory.container.setItem(i, item);
                                }
                                else {
                                    inventory.container.setItem(i);
                                }
                                break;
                            }
                        }
                    }
                }
            }, 1);
        }
    });
    world.afterEvents.entityDie.subscribe((event) => {
        if (CraftComputer.isComputer(event.deadEntity)) {
            const computer = new CraftComputer(event.deadEntity);
            let colortostring = ``;
            if (event.deadEntity.hasComponent(EntityColorComponent.componentId)) {
                const colorComponent = event.deadEntity.getComponent(EntityColorComponent.componentId);
                //debug(`color ${colorComponent.value}`);
                colortostring = `${colorComponent.value}`;
                if (colortostring === undefined) {
                    colortostring = ``;
                }
            }
            const computerItem = new ItemStack(computer.computer.typeId + colortostring + '_item', 1);
            computerItem.setLore([
                computer.getName(),
                computer.computerId
            ]);
            const dimension = computer.computer.dimension;
            const location = computer.computer.location;
            dimension.spawnItem(computerItem, location);
            const container = computer.computer.getComponent(EntityInventoryComponent.componentId).container;
            for (let i = 0; i < container.size; i++) {
                const item = container.getItem(i);
                if (item) {
                    dimension.spawnItem(item, location);
                }
            }
            // Mark attachments as downloaded
            computer.setEmails(computer.getEmails().map(e => {
                if (e.attachment !== undefined && e.attachment >= 0) {
                    e.attachment = -1;
                }
                return e;
            }));
            // Cancel & save future operations
            cancelOperations(computer.computer);
            const operations = getOperations(computer.computer);
            SET_PROP(world, IDENTIFIER_SAVED_OPERATIONS(computer.computerId), operations);
            SET_PROP(world, IDENTIFIER_ITEMIZED_ON(computer.computerId), world.getDay());
        }
    });
    // Vacuum at startup and then every 24 hours
    craftComputers.vacuumOldComputers();
    system.runInterval(() => {
        craftComputers.vacuumOldComputers();
    }, 20 * 60 * 60 * 24);
};
const VALID_COMPUTERS_TYPES = [
    'jig_ccomp:computer_old',
    'jig_ccomp:computer_ok',
    'jig_ccomp:computer_good',
    'jig_ccomp:computer_lap'
];
const findBlocked = (senderType, blockType, id) => (blocked) => {
    return blocked.sender_type === senderType && blocked.block_type === blockType && blocked.data.id === id;
};
const isPlayerBlocked = (playerType, player, blocks) => {
    return !!blocks.find(b => {
        if (b.sender_type === playerType) {
            if (b.block_type === "player") {
                if (b.data.id === player) {
                    return true;
                }
            }
        }
        return false;
    });
};
const evaluateBlocks = (blocks, email) => {
    return !!blocks.find(b => {
        if (b.sender_type === email.sender_type) {
            if (b.block_type === "computer") {
                if (b.data.id === email.from_computer_id) {
                    return true;
                }
            }
            else if (b.block_type === "player") {
                if (b.data.id === email.from_player_id) {
                    return true;
                }
            }
            return false;
        }
    });
};
const COMPUTER_VERSION = 2;
const COMPUTER_PREFIX = 'COMP';
const ITEMIZED_ON = 'ITEMIZED_ON';
const IDENTIFIER_NAME = (computerId) => PROP(COMPUTER_PREFIX, computerId, 'NAME');
const IDENTIFIER_VERSION = (computerId) => PROP(COMPUTER_PREFIX, computerId, 'VERSION');
const IDENTIFIER_CREATOR = (computerId) => PROP(COMPUTER_PREFIX, computerId, 'CREATOR');
const IDENTIFIER_EMAILS = (computerId) => PROP(COMPUTER_PREFIX, computerId, 'EMAILS');
const IDENTIFIER_BLOCKS = (computerId) => PROP(COMPUTER_PREFIX, computerId, 'EMAIL_BLOCKS');
const IDENTIFIER_WELCOME_SENT = (computerId) => PROP(COMPUTER_PREFIX, computerId, 'WELCOME_SENT');
const IDENTIFIER_ITEMIZED_ON = (computerId) => PROP(COMPUTER_PREFIX, computerId, ITEMIZED_ON);
const IDENTIFIER_LOST_DATA = (computerId) => PROP(COMPUTER_PREFIX, computerId, 'LOST_DATA');
const IDENTIFIER_SAVED_OPERATIONS = (computerId) => PROP(COMPUTER_PREFIX, computerId, 'SAVED_OPERATIONS');
const IDENTIFIER_TRIGGER_EMAIL_NOTIFICATION = (computerId) => PROP(COMPUTER_PREFIX, computerId, 'TRIGGER_EMAIL_NOTIFICATION');
class CraftComputer {
    computer;
    computerId;
    static isComputer(maybeComputer) {
        return VALID_COMPUTERS_TYPES.includes(maybeComputer.typeId);
    }
    constructor(entity) {
        if (!CraftComputer.isComputer(entity)) {
            throw new Error('Not a computer');
        }
        this.computer = entity;
        this.computerId = this.fetchInternalId();
        this.upgradeFromVersion1();
        const version = this.getVersion();
        if (version < COMPUTER_VERSION) {
            const name = () => {
                if (this.computer.typeId === 'jig_ccomp:computer_old') {
                    return `Old JX${makeRandomNumber(100, 999)}`;
                }
                else if (this.computer.typeId == 'jig_ccomp:computer_lap') {
                    return `LAP U${makeRandomNumber(100, 999)}`;
                }
                else if (this.computer.typeId == 'jig_ccomp:computer_ok') {
                    return `OK JTX${makeRandomNumber(1000, 4000)}`;
                }
                else if (this.computer.typeId == 'jig_ccomp:computer_good') {
                    return `RGB JigForce${makeRandomNumber(5000, 9999)}`;
                }
                return this.computerId;
            };
            switch (version) {
                case 0:
                    this.setName(name());
            }
            this.setVersion(COMPUTER_VERSION);
        }
    }
    fetchInternalId() {
        const computerId = GET_RAW_PROP(this.computer, "jig_ccomp:COMPUTER_ID" /* COMPUTER_ID */);
        if (typeof computerId !== 'string') {
            const newId = makeRandomId();
            //debug(`Generating id ${newId} for new computer`);
            SET_RAW_PROP(this.computer, "jig_ccomp:COMPUTER_ID" /* COMPUTER_ID */, newId);
            return newId;
        }
        return computerId;
    }
    upgradeFromVersion1() {
        const V1_COMPUTER_NAME = 'jig_ccomp:COMPUTER_NAME';
        const V1_EMAILS = 'jig_ccomp:EMAILS';
        const V1_EMAIL_BLOCKS = 'jig_ccomp:EMAIL_BLOCKS';
        const oldComputerName = this.computer.getDynamicProperty(V1_COMPUTER_NAME);
        if (oldComputerName !== undefined) {
            // Found a computer with name, it must be a v1 computer entity, upgrading...
            this.setName(oldComputerName);
            this.setEmails(JSON.parse(this.computer.getDynamicProperty(V1_EMAILS) ?? '[]'));
            this.setEmailBlocks(JSON.parse(this.computer.getDynamicProperty(V1_EMAIL_BLOCKS) ?? '[]'));
            this.setWelcomeSent(true);
            this.computer.setDynamicProperty(V1_COMPUTER_NAME, undefined);
            this.computer.setDynamicProperty(V1_EMAILS, undefined);
            this.computer.setDynamicProperty(V1_EMAIL_BLOCKS, undefined);
            this.setVersion(2);
        }
    }
    getId() {
        return this.computerId;
    }
    getName() {
        return GET_RAW_PROP(world, IDENTIFIER_NAME(this.computerId));
    }
    setName(name) {
        SET_RAW_PROP(world, IDENTIFIER_NAME(this.computerId), name);
    }
    getVersion() {
        return GET_RAW_PROP(world, IDENTIFIER_VERSION(this.computerId)) ?? 0;
    }
    setVersion(version) {
        SET_RAW_PROP(world, IDENTIFIER_VERSION(this.computerId), version);
    }
    getCreator() {
        return GET_RAW_PROP(world, IDENTIFIER_CREATOR(this.computerId));
    }
    setCreator(creator) {
        SET_RAW_PROP(world, IDENTIFIER_CREATOR(this.computerId), creator);
    }
    getWelcomeSent() {
        return !!GET_RAW_PROP(world, IDENTIFIER_WELCOME_SENT(this.computerId));
    }
    setWelcomeSent(welcomeSent) {
        SET_RAW_PROP(world, IDENTIFIER_WELCOME_SENT(this.computerId), welcomeSent);
    }
    getLostData() {
        return GET_RAW_PROP(world, IDENTIFIER_LOST_DATA(this.computerId)) ?? 0;
    }
    getEmails() {
        return GET_PROP(world, IDENTIFIER_EMAILS(this.computerId), []);
    }
    setEmails(emails) {
        SET_PROP(world, IDENTIFIER_EMAILS(this.computerId), emails);
    }
    getEmailBlocks() {
        return GET_PROP(world, IDENTIFIER_BLOCKS(this.computerId), []);
    }
    setEmailBlocks(emailBlocks) {
        SET_PROP(world, IDENTIFIER_BLOCKS(this.computerId), emailBlocks);
    }
    addBlock(emailBlock) {
        if (!this.isBlocked(emailBlock.sender_type, emailBlock.block_type, emailBlock.data.id)) {
            const blocks = this.getEmailBlocks();
            blocks.push(emailBlock);
            this.setEmailBlocks(blocks);
        }
    }
    getBlocked(senderType, blockType, id) {
        const blocks = this.getEmailBlocks();
        return blocks.find(findBlocked(senderType, blockType, id));
    }
    isBlocked(senderType, blockType, id) {
        return !!this.getBlocked(senderType, blockType, id);
    }
    addPlayerBlock(email) {
        this.addBlock({
            id: makeRandomId(),
            sender_type: email.sender_type,
            block_type: 'player',
            data: {
                id: email.from_player_id,
                name: TEXT_OR_RAW_TEXT(email.from_player)
            }
        });
    }
    removePlayerBlock(email) {
        this.removeBlockByEmail(email, "player");
    }
    removeComputerBlock(email) {
        this.removeBlockByEmail(email, "computer");
    }
    removeBlockByEmail(email, blockType) {
        const blocks = this.getEmailBlocks();
        blocks.filter(block => {
            return block.sender_type !== email.sender_type && block.block_type !== blockType && block.data.id !== email.from_player_id;
        });
        this.setEmailBlocks(blocks);
    }
    addComputerBlock(email) {
        this.addBlock({
            id: makeRandomId(),
            sender_type: email.sender_type,
            block_type: 'computer',
            data: {
                id: email.from_computer_id,
                name: TEXT_OR_RAW_TEXT(email.from_computer_name)
            }
        });
    }
    removeBlock(blockId) {
        const blocks = this.getEmailBlocks();
        this.setEmailBlocks(blocks.filter(b => blockId !== b.id));
    }
    addEmail(email, attachment, forceSend = false) {
        const blocks = this.getEmailBlocks();
        const blockFound = evaluateBlocks(blocks, email);
        if (blockFound) {
            //debug(`Computer blocked email from: ${typeof email.from_player === 'string' ? email.from_player : email.from_player_id} to computer ${email.from_computer_name} was blocked`);
            return 'blocked';
        }
        if (attachment) {
            const inventory = this.computer.getComponent(EntityInventoryComponent.componentId);
            if (inventory.container) {
                const container = inventory.container;
                const emptySlot = getEmptySlot(container);
                if (emptySlot !== undefined) {
                    if (attachment.container && attachment.index !== undefined) {
                        attachment.container.setItem(attachment.index);
                    }
                    container.setItem(emptySlot, attachment.item);
                    email.attachment = emptySlot;
                }
                else {
                    if (!forceSend) {
                        return false;
                    }
                }
            }
        }
        this.setEmails([...this.getEmails(), {
                ...email,
                id: makeRandomId(),
            }]);
        this.computer.triggerEvent('jig_ccomp:got_email');
        return true;
    }
    downloadAttachment(emailId, player) {
        // Todo: this is simpler in later versions
        const playerContainer = player.getComponent(EntityInventoryComponent.componentId)?.container;
        const computerContainer = this.computer.getComponent(EntityInventoryComponent.componentId)?.container;
        const emails = this.getEmails();
        const email = emails.find(e => e.id === emailId);
        if (email.attachment !== undefined && email.attachment >= 0 && playerContainer && computerContainer) {
            const emptySlot = getEmptySlot(playerContainer);
            if (emptySlot !== undefined) {
                const item = computerContainer.getItem(email.attachment);
                computerContainer.setItem(email.attachment);
                playerContainer.setItem(emptySlot, item);
                email.attachment = -1;
                this.setEmails(emails);
                //player.runCommand('/say you got'+test)
                player.runCommand('/playsound jig_common.suckpop');
                this.computer.runCommand(`/dialogue open @s "${player.name}" jig_ccomp:old_pc_emdo`);
                this.computer.triggerEvent('jig_ccomp:typing');
            }
            else {
                this.computer.runCommand(`/dialogue open @s "${player.name}" jig_ccomp:old_at_emdo`);
                this.computer.triggerEvent('jig_ccomp:typing');
            }
        }
        return false;
    }
    deleteEmail(emailId) {
        const computerContainer = this.computer.getComponent(EntityInventoryComponent.componentId)?.container;
        const emails = this.getEmails();
        const emailIndex = emails.findIndex(e => e.id === emailId);
        const email = emails[emailIndex];
        if (email.attachment !== undefined && email.attachment >= 0 && computerContainer) {
            computerContainer.setItem(email.attachment);
        }
        emails.splice(emailIndex, 1);
        this.setEmails(emails);
    }
    markAsRead(emailId) {
        const emails = this.getEmails();
        const email = emails.find(e => e.id === emailId);
        email.isRead = true;
        this.setEmails(emails);
    }
    getEmailById(emailId) {
        const emails = this.getEmails();
        return emails.find(e => e.id === emailId);
    }
    say(message, player) {
        void this.computer.runCommandAsync(`w ${player.name} ${message}`);
    }
    getProp(property, defaultValue) {
        return this.computer.getDynamicProperty(property) ?? defaultValue;
    }
    setProp(property, value) {
        this.computer.setDynamicProperty(property, value);
    }
}
class CraftComputers {
    computers = {};
    dimensions = [];
    init(dimension) {
        if (this.dimensions.includes(dimension)) {
            return;
        }
        this.dimensions.push(dimension);
        for (const computerType of VALID_COMPUTERS_TYPES) {
            dimension.getEntities({
                type: computerType
            }).forEach((computer) => {
                this.addComputer(computer);
            });
        }
    }
    addComputer(entity) {
        if (!CraftComputer.isComputer(entity)) {
            return;
        }
        const computer = new CraftComputer(entity);
        this.computers[computer.getId()] = computer;
    }
    removeComputer(entity) {
        if (!CraftComputer.isComputer(entity)) {
            return;
        }
        if (entity.isValid()) {
            const computer = new CraftComputer(entity);
            const computerId = computer.getId();
            delete this.computers[computerId];
            //debug(`Removing computer ${computerId}`);
        }
    }
    getComputers() {
        return Object.values(this.computers).filter(c => c.computer.isValid());
    }
    getComputer(id) {
        if (id in this.computers && this.computers[id].computer.isValid()) {
            return this.computers[id];
        }
        return undefined;
    }
    isComputerAvailable(id) {
        if (id in this.computers && this.computers[id].computer.isValid()) {
            return true;
        }
        const reachableComputers = craftComputers.getReachableStoredComputerRecipients();
        if (reachableComputers.find(c => c.id === id)) {
            return true;
        }
        return false;
    }
    getUnlockedNpcs() {
        return new Set(this.getComputers().flatMap(c => {
            return c.getEmails()
                .filter(e => e.sender_type === "npc")
                .map(e => e.from_player_id);
        }));
    }
    broadcastEmail(email) {
        const attachment = email.attachment ? {
            item: new ItemStack(email.attachment, email.attachmentCount ?? 1)
        } : undefined;
        this.getComputers().filter(c => filterComputer(c, email.filter)).forEach(c => {
            c.addEmail({
                sender_type: 'script',
                from_player_id: email.addon,
                from_computer_id: email.computer,
                from_player: email.addon_name,
                from_computer_name: email.computer_name,
                isRead: false,
                message: email.message,
                printable: false,
                subject: email.subject
            }, attachment, true);
        });
    }
    getReachableStoredComputerRecipients() {
        const prefix = `${PROP(COMPUTER_PREFIX, '')}`;
        const postfix = `:${ITEMIZED_ON}`;
        const itemized_ids = world.getDynamicPropertyIds().filter(p => p.startsWith(prefix) && p.endsWith(postfix));
        return itemized_ids
            .filter(itemized_on_id => {
            const day = parseInt(GET_RAW_PROP(world, itemized_on_id));
            return world.getDay() <= day + 1;
        })
            .map(itemized_on_id => {
            const path = itemized_on_id.split(':');
            return path[path.length - 2];
        })
            .map(computer_id => {
            return {
                id: computer_id,
                name: GET_RAW_PROP(world, IDENTIFIER_NAME(computer_id))
            };
        });
    }
    vacuumOldComputers() {
        //debug(`Running itemized computer vacuum`);
        const VACUUM_AFTER_COMPUTERS = 200; // only vacuum when we hit this number of itemized computers
        const VACUUM_OLDER_THAN = 2000; // minecraft days 2000 is around 30 natural days
        const prefix = `${PROP(COMPUTER_PREFIX, '')}`;
        const postfix = `:${ITEMIZED_ON}`;
        const itemized_ids = world.getDynamicPropertyIds().filter(p => p.startsWith(prefix) && p.endsWith(postfix));
        //debug(`Found ${itemized_ids.length} itemized entries`);
        if (itemized_ids.length > VACUUM_AFTER_COMPUTERS) {
            const vacuum_ids = itemized_ids.filter(last_seen_id => {
                const last_seen = parseInt(GET_RAW_PROP(world, last_seen_id));
                if (last_seen + VACUUM_OLDER_THAN < world.getDay()) {
                    return true;
                }
            }).map(itemized_on_id => {
                const path = itemized_on_id.split(':');
                return path[path.length - 2];
            });
            debug(`Vacuuming ${vacuum_ids.length} itemized computers.`);
            // Vacuum ids
            vacuum_ids.forEach(id => {
                SET_RAW_PROP(world, IDENTIFIER_NAME(id), undefined);
                SET_RAW_PROP(world, IDENTIFIER_VERSION(id), undefined);
                SET_RAW_PROP(world, IDENTIFIER_EMAILS(id), undefined);
                SET_RAW_PROP(world, IDENTIFIER_BLOCKS(id), undefined);
                SET_RAW_PROP(world, IDENTIFIER_WELCOME_SENT(id), undefined);
                SET_RAW_PROP(world, IDENTIFIER_ITEMIZED_ON(id), undefined);
                SET_RAW_PROP(world, IDENTIFIER_LOST_DATA(id), undefined);
                SET_RAW_PROP(world, IDENTIFIER_SAVED_OPERATIONS(id), undefined);
            });
        }
    }
}
const filterComputer = (computer, filter) => {
    if (filter) {
        if (filter.players) {
            const creator = computer.getCreator();
            if (!creator) {
                return false;
            }
            if (!filter.players.includes(creator)) {
                return false;
            }
        }
    }
    return true;
};
const craftComputers = new CraftComputers();

const asyncWait = async (ticks) => {
    return new Promise((resolve) => {
        system.runTimeout(() => {
            resolve();
        }, ticks);
    });
};
// Use a builder pattern
// builder pattern: https://medium.com/geekculture/implementing-a-type-safe-object-builder-in-typescript-e973f5ecfb9c
// Needs updating to support optional types
// class ObjectBuilder {
//   public static new<Target>(): IWith<Target, {}> {
//     return new Builder<Target, {}>({});
//   }
// }
//
// interface IWith<Target, Supplied> {
//   with<T extends Omit<Target, keyof Supplied>, K extends keyof T>(
//     key: K,
//     value: T[K],
//   ): keyof Omit<Omit<Target, keyof Supplied>, K> extends never
//     ? IBuild<Target>
//     : IWith<Target, Supplied & Pick<T, K>>;
// }
//
// interface IBuild<Target> {
//   build(): Target;
// }
//
// class Builder<Target, Supplied> implements IBuild<Target>, IWith<Target, Supplied> {
//   constructor(private target: Partial<Target>) {}
//
//   with<T extends Omit<Target, keyof Supplied>, K extends keyof T>(key: K, value: T[K]) {
//     const target: Partial<Target> = { ...this.target, [key]: value };
//
//     return new Builder<Target, Supplied & Pick<T, K>>(target);
//   }
//
//   build() {
//     return this.target as Target;
//   }
// }

const DefaultShowDialogOptions = Object.freeze({
    lockPlayerCamera: true,
    busyRetriesCount: 5,
    busyRetriesTick: 5,
});
/**
 * Base class for all the script dialogues
 *
 * @category Script dialogue
 */
class ScriptDialogue {
    constructor() { }
    /**
     * Opens the script dialogue. It requires the {@link ShowDialogueOptions#player player} as one of the options.
     *
     * You can configure displaying features by setting the optional values in the {@link ShowDialogueOptions options}
     * @param options used to configure what and how to display the script dialogue.
     *
     * @see {@link ShowDialogueOptions}
     */
    async open(options) {
        const resolvedOptions = this.resolveShowDialogueOptions(options);
        try {
            if (resolvedOptions.lockPlayerCamera) {
                this.lockPlayerCamera(resolvedOptions);
            }
            try {
                const showable = await this.getShowable(resolvedOptions);
                const response = await this.show(showable, resolvedOptions);
                if (response.canceled) {
                    return new DialogueCanceledResponse(response.cancelationReason);
                }
                return await this.processResponse(response, resolvedOptions);
            }
            catch (e) {
                if (e && e instanceof FormRejectError) {
                    return new DialogueRejectedResponse(e.reason, e);
                }
                else {
                    return new DialogueRejectedResponse(undefined, e);
                }
            }
        }
        finally {
            if (resolvedOptions.lockPlayerCamera) {
                this.unlockPlayerCamera(resolvedOptions);
            }
        }
    }
    async show(showable, options) {
        let i = 0;
        for (;;) {
            const response = await showable.show(options.player);
            if (response.canceled && response.cancelationReason === FormCancelationReason.UserBusy) {
                if (i < options.busyRetriesCount) {
                    i++;
                    await asyncWait(options.busyRetriesTick);
                    if (options.player.isValid()) {
                        continue;
                    }
                }
            }
            return response;
        }
    }
    lockPlayerCamera(options) {
        options.player.runCommand(`inputpermission set "${options.player.name}" camera disabled`);
        options.player.runCommand(`inputpermission set "${options.player.name}" movement disabled`);
    }
    unlockPlayerCamera(options) {
        options.player.runCommand(`inputpermission set "${options.player.name}" camera enabled`);
        options.player.runCommand(`inputpermission set "${options.player.name}" movement enabled`);
    }
    resolveShowDialogueOptions(options) {
        return {
            ...DefaultShowDialogOptions,
            ...options,
        };
    }
}
/**
 * Base class for script dialogue responses
 *
 * @category Responses
 */
class ScriptDialogueResponse {
}
/**
 * Dialogue response when the script dialogue was canceled.
 *
 * Typically this happens if the player was busy when trying to open the script dialogue or if the user
 * pressed the close button (x).
 *
 * Valid reasons can be seen in {@link @minecraft/server-ui!FormCancelationReason FormCancelationReason}
 *
 * @category Responses
 */
class DialogueCanceledResponse extends ScriptDialogueResponse {
    /**
     * Reason the script dialogue was cancel.
     */
    reason;
    /**
     * @internal
     */
    constructor(reason) {
        super();
        this.reason = reason;
    }
}
/**
 * Dialogue response when the script dialogue was rejected, throwing an exception.
 *
 * Reasons can be seen in {@link @minecraft/server-ui!FormRejectReason FormRejectReason} if any reason could be
 * determined.
 *
 * Known reasons can be seen in {@link @minecraft/server-ui!FormRejectReason FormRejectReason}
 * @category Responses
 */
class DialogueRejectedResponse extends ScriptDialogueResponse {
    /**
     * Reason the script dialogue was rejected if there was any, else the {@link exception} needs to be checked
     * to further determine what was the error.
     */
    reason;
    /**
     * Exception that was throw to rejected the script dialogue.
     */
    exception;
    /**
     * @internal
     */
    constructor(reason, exception) {
        super();
        this.reason = reason;
        this.exception = exception;
    }
}
/**
 * Response for button script dialogues.
 *
 * Contains the selected button's name.
 *
 * @category Responses
 * @category Dual button script dialogue
 * @category Multi button script dialogue
 */
class ButtonDialogueResponse extends ScriptDialogueResponse {
    /**
     * Selected button's name.
     *
     * @see {@link DualButton#name}
     * @see {@link MultiButton#name}
     */
    selected;
    callback;
    /**
     * @internal
     */
    constructor(selected, callback) {
        super();
        this.selected = selected;
        this.callback = callback;
    }
}
class MissingButtonsException extends Error {
    constructor() {
        super('Missing buttons');
    }
}

/**
 * Creates a new dual button script dialogue
 *
 * @category Creation
 * @category Dual button script dialogue
 * @param title Title of the dual button script dialogue
 * @param topButton Contents of the top button
 * @param bottomButton Contents of the bottom button
 */
const dualButtonScriptDialogue = (title, topButton, bottomButton) => {
    return new DualButtonScriptDialogue(title, undefined, topButton, bottomButton);
};
/**
 * Dual button script dialogue class.
 *
 * User's don't need to instantiate this class directly, instead they can use {@link dualButtonScriptDialogue}.
 *
 * Allows the users to optionally set a value for the body by calling {@link DualButtonScriptDialogue#setBody setBody}.
 *
 * @category Dual button script dialogue
 */
class DualButtonScriptDialogue extends ScriptDialogue {
    title;
    body;
    topButton;
    bottomButton;
    /**
     * @internal
     */
    constructor(title, body, topButton, bottomButton) {
        super();
        this.title = title;
        this.body = body;
        this.topButton = topButton;
        this.bottomButton = bottomButton;
    }
    /**
     * Sets content of the script dialogue
     * @param body
     */
    setBody(body) {
        return new DualButtonScriptDialogue(this.title, body, this.topButton, this.bottomButton);
    }
    getShowable(_options) {
        const data = new MessageFormData();
        data.title(this.title);
        if (this.body) {
            data.body(this.body);
        }
        data.button1(this.bottomButton.text);
        data.button2(this.topButton.text);
        return data;
    }
    async processResponse(response, _options) {
        const selectedButton = response.selection === 0 ? this.bottomButton : this.topButton;
        let callbackResponse = undefined;
        if (selectedButton.callback) {
            callbackResponse = await selectedButton.callback(selectedButton.name);
        }
        return new ButtonDialogueResponse(selectedButton.name, callbackResponse);
    }
}

/**
 * Initializes a empty multi button script dialogue.
 *
 * Buttons needs to be added using {@link MultiButtonDialogue#addButton} or {@link MultiButtonDialogue#addButtons}
 *
 * @category Creation
 * @category Multi button script dialogue
 * @param title
 */
const multiButtonScriptDialogue = (title) => {
    return new MultiButtonDialogue(title, undefined, []);
};
/**
 * Class used to build multi button script dialogues.
 *
 * Use {@link multiButtonScriptDialogue} to initialize one.
 *
 * @category Multi button script dialogue
 * @see {@link multiButtonScriptDialogue}
 */
class MultiButtonDialogue extends ScriptDialogue {
    title;
    body;
    buttons;
    /**
     * @internal
     */
    constructor(title, body, buttons) {
        super();
        this.title = title;
        this.body = body;
        this.buttons = buttons;
    }
    /**
     * Sets the content body of the multi button dialogue
     * @param body
     */
    setBody(body) {
        return new MultiButtonDialogue(this.title, body, this.buttons);
    }
    /**
     * Adds a button to the multi button script dialogue.
     * @param name name of the button
     * @param text content of the button
     * @param iconPath path to an icon to show in the button
     */
    addButton(name, text, iconPath, callback) {
        return new MultiButtonDialogue(this.title, this.body, [
            ...this.buttons,
            {
                name,
                text,
                iconPath,
                callback,
            },
        ]);
    }
    /**
     * Adds multiple buttons to the multi button script dialogue.
     * @param buttons array of buttons
     */
    addButtons(buttons) {
        return new MultiButtonDialogue(this.title, this.body, [
            ...this.buttons,
            ...buttons,
        ]);
    }
    getShowable(_options) {
        if (this.buttons.length === 0) {
            throw new MissingButtonsException();
        }
        const formData = new ActionFormData();
        formData.title(this.title);
        if (this.body) {
            formData.body(this.body);
        }
        this.buttons.forEach((button) => {
            formData.button(button.text, button.iconPath);
        });
        return formData;
    }
    async processResponse(response, _options) {
        const selectedButton = this.buttons[response.selection];
        let callbackResponse = undefined;
        if (selectedButton.callback) {
            callbackResponse = await selectedButton.callback(selectedButton.name);
        }
        return new ButtonDialogueResponse(selectedButton.name, callbackResponse);
    }
}

/**
 * Initializes a empty input script dialogue.
 *
 * Elements needs to be added using {@link InputScriptDialogue#addElement} or {@link InputScriptDialogue#addElements}
 *
 * @category Creation
 * @category Input script dialogue
 *
 * @param title Title for the script dialogue
 */
const inputScriptDialogue = (title) => {
    return new InputScriptDialogue(title, []);
};
/**
 * Creates a new dropdown element to use in a input script dialogue.
 *
 * @category Input script dialogue
 * @param name
 * @param label
 */
const inputDropdown = (name, label) => {
    return new InputDropdown(name, label, [], 0);
};
/**
 * Creates a new slider element to use in a input script dialogue.
 *
 * @category Input script dialogue
 * @param name
 * @param label
 * @param minimumValue
 * @param maximumValue
 * @param valueStep
 * @param defaultValue
 */
const inputSlider = (name, label, minimumValue, maximumValue, valueStep, defaultValue) => {
    return new InputSlider(name, label, minimumValue, maximumValue, valueStep, defaultValue);
};
/**
 * Creates a new text field element to use in a input script dialogue.
 *
 * @category Input script dialogue
 * @param name
 * @param label
 * @param placeholderText
 * @param defaultValue
 */
const inputText = (name, label, placeholderText, defaultValue) => {
    return new InputText(name, label, placeholderText, defaultValue);
};
/**
 * Creates a new toggle element to use in a input script dialogue.
 *
 * @category Input script dialogue
 * @param name
 * @param label
 * @param defaultValue
 */
const inputToggle = (name, label, defaultValue) => {
    return new InputToggle(name, label, defaultValue);
};
/**
 * Base for all input elements displayed in the input script dialogue.
 *
 * You don't need to instantiate this class directly, instead you can use any of the builder function
 * depending the type you want to use.
 *
 * @category Input script dialogue
 * @see {@link inputDropdown}
 * @see {@link inputSlider}
 * @see {@link inputToggle}
 * @see {@link inputText}
 */
class InputElement {
    /**
     * @internal
     */
    name;
    /**
     * @internal
     */
    label;
    /**
     * @internal
     */
    constructor(name, label) {
        this.name = name;
        this.label = label;
    }
}
/**
 * Base for all input elements that have a fixed default value.
 *
 * You don't need to instantiate this class directly, instead you can use any of the builder function
 * depending the type you want to use.
 *
 * @category Input script dialogue
 * @see {@link inputDropdown}
 * @see {@link inputSlider}
 * @see {@link inputToggle}
 * @see {@link inputText}
 */
class InputWithDefaultValue extends InputElement {
    /**
     * @internal
     */
    defaultValue;
    /**
     * @internal
     */
    constructor(name, label, defaultValue) {
        super(name, label);
        this.defaultValue = defaultValue;
    }
}
/**
 * @internal
 */
class InputDropdownOption {
    label;
    value;
    constructor(label, value) {
        this.label = label;
        this.value = value;
    }
}
/**
 * Input element's representation of a dropdown.
 *
 * Instantiate by using {@link inputDropdown}
 *
 * @category Input script dialogue
 * @see {@link inputDropdown}
 */
class InputDropdown extends InputElement {
    /**
     * @internal
     */
    options;
    /**
     * @internal
     */
    defaultValueIndex;
    /**
     * @internal
     */
    constructor(name, label, options, defaultValueIndex) {
        super(name, label);
        this.defaultValueIndex = defaultValueIndex ?? 0;
        this.options = options;
    }
    /**
     * Sets the default index of the option you would like to use
     * @param defaultValueIndex
     */
    setDefaultValueIndex(defaultValueIndex) {
        return new InputDropdown(this.name, this.label, [...this.options], defaultValueIndex);
    }
    /**
     * Adds an option to the dropdown
     * @param label
     * @param value
     */
    addOption(label, value) {
        return new InputDropdown(this.name, this.label, [...this.options, new InputDropdownOption(label, value)], this.defaultValueIndex);
    }
}
/**
 * Input element's representation of a slider.
 *
 * Instantiate by using {@link inputSlider}
 *
 * @category Input script dialogue
 * @see {@link inputSlider}
 */
class InputSlider extends InputWithDefaultValue {
    /**
     * @internal
     */
    minimumValue;
    /**
     * @internal
     */
    maximumValue;
    /**
     * @internal
     */
    valueStep;
    /**
     * @internal
     */
    constructor(name, label, minimumValue, maximumValue, valueStep, defaultValue) {
        super(name, label, defaultValue ?? minimumValue);
        this.minimumValue = minimumValue;
        this.maximumValue = maximumValue;
        this.valueStep = valueStep;
    }
}
/**
 * Input element's representation of text field.
 *
 * Instantiate by using {@link inputText}
 *
 * @category Input script dialogue
 * @see {@link inputText}
 */
class InputText extends InputWithDefaultValue {
    /**
     * @internal
     */
    placeholderText;
    /**
     * @internal
     */
    constructor(name, label, placeholderText, defaultValue) {
        super(name, label, defaultValue ?? '');
        this.placeholderText = placeholderText;
    }
}
/**
 * Input element's representation of toggle.
 *
 * Instantiate by using {@link inputToggle}
 *
 * @category Input script dialogue
 * @see {@link inputToggle}
 */
class InputToggle extends InputWithDefaultValue {
    /**
     * @internal
     */
    constructor(name, label, defaultValue) {
        super(name, label, !!defaultValue);
    }
}
/**
 * Class used to build input script dialogues.
 *
 * Use {@link inputScriptDialogue} to initialize one.
 *
 * @category Input script dialogue
 * @see {@link inputScriptDialogue}
 */
class InputScriptDialogue extends ScriptDialogue {
    elements;
    title;
    /**
     * @internal
     */
    constructor(title, elements) {
        super();
        this.title = title;
        this.elements = elements;
    }
    /**
     * Adds an input element to the input script dialogue.
     * @param element
     *
     * @see {@link inputDropdown}
     * @see {@link inputSlider}
     * @see {@link inputToggle}
     * @see {@link inputText}
     */
    addElement(element) {
        return new InputScriptDialogue(this.title, [...this.elements, element]);
    }
    /**
     * Adds multiple input element to the input script dialogue.
     * @param elements
     *
     * @see {@link inputDropdown}
     * @see {@link inputSlider}
     * @see {@link inputToggle}
     * @see {@link inputText}
     */
    addElements(elements) {
        return new InputScriptDialogue(this.title, [...this.elements, ...elements]);
    }
    getShowable(_options) {
        if (this.elements.length === 0) {
            throw new MissingElementsError();
        }
        const data = new ModalFormData();
        data.title(this.title);
        this.elements.forEach((element) => {
            if (element instanceof (InputDropdown)) {
                data.dropdown(element.label, element.options.map((o) => o.label), element.defaultValueIndex);
            }
            else if (element instanceof (InputSlider)) {
                data.slider(element.label, element.minimumValue, element.maximumValue, element.valueStep, element.defaultValue);
            }
            else if (element instanceof (InputText)) {
                data.textField(element.label, element.placeholderText, element.defaultValue);
            }
            else if (element instanceof (InputToggle)) {
                data.toggle(element.label, element.defaultValue);
            }
        });
        return data;
    }
    async processResponse(response, _options) {
        const formValues = response.formValues ?? this.elements.map((_e) => undefined);
        const values = this.elements.map((element, index) => {
            const name = element.name;
            let value = 0;
            const formValue = formValues[index];
            if (element instanceof (InputDropdown)) {
                value = element.options[element.defaultValueIndex].value;
                if (formValue !== undefined) {
                    value = element.options[formValue].value;
                }
            }
            else if (element instanceof (InputSlider) ||
                element instanceof (InputText) ||
                element instanceof (InputToggle)) {
                value = element.defaultValue;
                if (formValue !== undefined) {
                    value = formValue;
                }
            }
            return {
                [name]: value,
            };
        });
        return new InputScriptDialogueResponse(Object.assign({}, ...values));
    }
}
/**
 * Script dialogue response from a input script dialogue.
 * Holds a map with the values used by each of the element's name.
 *
 * @category Input script dialogue
 * @category Responses
 */
class InputScriptDialogueResponse {
    /**
     * Map with the values of the script dialogue..
     */
    values;
    constructor(values) {
        this.values = values;
    }
}
class MissingElementsError extends Error {
    constructor() {
        super('Missing input elements');
    }
}

const setPcIdDialogue = (oldName) => {
    return inputScriptDialogue(TRANSLATE("jig_ccomp:pc-id-set.title"))
        .addElement(inputText("name", TRANSLATE("jig_ccomp:pc-id-set.name.label", "\n", "\n\n"), TRANSLATE("jig_ccomp:pc-id-set.name.placeholder"), oldName));
};

const WORLD_INFO_OBJECTIVE = 'jig_ccomp:world_info';
const COMPUTER_INFO_OBJECTIVE = 'jig_ccomp:jig_computer.addon_stats';
const CUSTOM_OBJECTIVES = 'jig_computer.addon_stats';

const viewBoardsList = (boards) => {
    let boardsElement = inputDropdown("board", RAW_TEXT(TEXT("\n"), TRANSLATE("jig_ccomp:view-stat-list.board", "\n\n")));
    boards.forEach(t => {
        boardsElement = boardsElement.addOption(t, t);
    });
    const toIndex = boards.findIndex(t => t === WORLD_INFO_OBJECTIVE);
    boardsElement = boardsElement.setDefaultValueIndex(toIndex === -1 ? 0 : toIndex);
    return inputScriptDialogue(TRANSLATE("jig_ccomp:view-stat-list.title"))
        .addElement(boardsElement);
};

const getScoreOr = (scoreboardObjectiveOrName, score, defaultValue) => {
    try {
        const value = getScoreInternal(scoreboardObjectiveOrName, score);
        if (value === undefined) {
            return defaultValue;
        }
        return value;
    }
    catch (e) {
        return defaultValue;
    }
};
const getScoreInternal = (scoreboardObjectiveOrName, score) => {
    const scoreboardObjective = typeof scoreboardObjectiveOrName === 'string' ? world.scoreboard.getObjective(scoreboardObjectiveOrName) : scoreboardObjectiveOrName;
    if (!scoreboardObjective) {
        return undefined;
    }
    return scoreboardObjective.getScore(score);
};

const viewStats = (board, TimeOfDay, worldDay, installedDays, stats) => {
    return dualButtonScriptDialogue(RAW_TEXT(TRANSLATE("jig_ccomp:view-stats-board.title")), {
        name: "other-stats",
        text: TRANSLATE("jig_ccomp:view-stats-board.other-stats")
    }, {
        name: "quit",
        text: TRANSLATE("jig_ccomp:quit"),
    }).setBody(RAW_TEXT(TEXT("§l"), TRANSLATE(board), TEXT("§r"), ...isItWorldObjectives(board, TimeOfDay, worldDay, installedDays, stats), TEXT("\n\n"), ...isThereNone(board), ...buildStats(board, stats), TEXT("\n"), TRANSLATE(getCreditsKey(board), "\n")));
};
const getCreditsKey = (board) => {
    if (board === WORLD_INFO_OBJECTIVE) {
        return '';
    }
    return board.replace(CUSTOM_OBJECTIVES, "credits.for.computers");
};
const isItWorldObjectives = (board, TimeOfDay, worldDay, installedDays, stats) => {
    if (board === WORLD_INFO_OBJECTIVE) {
        const worldInfoRawText = [
            TEXT("\n"), TRANSLATE("jig_ccomp:current_time"), TEXT(TimeOfDay),
            TEXT("\n"), TRANSLATE("jig_ccomp:world_age"), TRANSLATE("jig_ccomp:days"), TEXT(worldDay),
            TEXT("\n"), TRANSLATE("jig_ccomp:installed_age"), TRANSLATE("jig_ccomp:days"), TEXT(installedDays),
        ];
        if (stats.length > 0) {
            worldInfoRawText.push(TEXT("\n"), TRANSLATE("jig_ccomp:since-installed-disclaimer"));
        }
        return worldInfoRawText;
    }
    return [];
};
const isThereNone = (board) => {
    const objective = world.scoreboard.getObjective(board);
    if (!objective) {
        debug(`Couldn't find objective ${board}`);
        return [];
    }
    let test = objective.getParticipants();
    if (test.length == 0) {
        return [
            TRANSLATE("jig_ccomp:there_isnt_any_stats_yet"), TEXT("\n")
        ];
    }
    return [];
};
const buildStats = (board, stats) => {
    const objective = world.scoreboard.getObjective(board);
    if (!objective) {
        debug(`Couldn't find objective ${board}`);
        return [];
    }
    return stats.map(stat => RAW_TEXT(TRANSLATE(stat), TEXT(":\n"), TEXT(prettyNumberFormat(getScoreOr(objective, stat, 0))), TEXT("\n")));
};
const prettyNumberFormat = (value) => {
    const stringNumber = value.toString().split('.');
    let integerPart = stringNumber[0];
    for (let i = integerPart.length - 3; i > 0; i -= 3) {
        integerPart = integerPart.substring(0, i) + ',' + integerPart.substring(i);
    }
    stringNumber[0] = integerPart;
    return stringNumber.join('.');
};

const defaultScriptDialogueOptions = {
    busyRetriesCount: 5,
    busyRetriesTick: 5,
    lockPlayerCamera: true,
};

const Config = {
    usePredefined: false,
    favoriteEmailAttachtment: 'jig_ccomp:favorite_email',
    JigarBoxNpcId: 'JigarBox',
    world: {
        installedTime: 'jig_ccomp:installed_time'
    },
    player: {
        jignetContacted: 'jig_ccomp:jignet_contacted'
    },
    forum: {
        post_capacity: 50,
        random_npc_post_ticks: 60 * 60 * 20 // every hour
    },
    radar: {
        //Radar modification tags
        tags: {
            // for chainmail simulation
            chain_protect: 'jig_ccomp:chain_protect',
            // for hidden completely
            radar_protect: 'jig_ccomp:radar_protect'
        }
    },
    predefinedPcId: [
        'Ancient PC 1',
        'Ancient PC 2',
        'Ancient PC 3',
        'Ancient PC 4',
        'Ancient PC 5',
        'OK PC 1',
        'OK PC 2',
        'OK PC 3',
        'OK PC 4',
        'OK PC 5',
        'RGB PC 1',
        'RGB PC 2',
        'RGB PC 3',
        'RGB PC 4',
        'RGB PC 5',
        'Home 1',
        'Home 2',
        'Home 3',
        'Spare 1',
        'Spare 2',
        'Spare 3',
        'Farm PC',
        'Basement PC',
        'Upstairs PC',
        'Castle PC',
        'Mom',
        'Dad',
        'Sister 1',
        'Sister 2',
        'Sister 3',
        'Sister 4',
        'Brother 1',
        'Brother 2',
        'Brother 3',
        'Brother 4',
        'Friend 1',
        'Friend 2',
        'Friend 3',
        'Friend 4',
        'Friend 5',
        'Friend 6',
        'Friend 7',
        'Best Friend',
        'Basalt Delta PC',
        'Beach PC',
        'Birch PC',
        'Caves PC',
        'Cherry Grove PC',
        'Crimson Forest PC',
        'Dark Oak Forest PC',
        'Deep Ocean PC',
        'Deep Dark PC',
        'Desert PC',
        'Dripstone Caves PC',
        'Edge PC',
        'Extreme Hills PC',
        'Flower Forest PC',
        'Forest PC',
        'Frozen PC',
        'Frozen Peaks PC',
        'Grove PC',
        'Hills PC',
        'Ice PC',
        'Ice Plains PC',
        'Jagged Peaks PC',
        'Jungle PC',
        'Lakes PC',
        'Lush Cave PC',
        'Mangrove Swamp PC',
        'Mega PC',
        'Mesa PC',
        'Mooshroom Island PC',
        'Mountain PC',
        'Mutated Biome PC',
        'Nether PC',
        'Nether Wastes PC',
        'Netherwart Forest PC',
        'Ocean PC',
        'Plains PC',
        'Plateau PC',
        'River PC',
        'Roofed Forest PC',
        'Savanna PC',
        'Shore PC',
        'Snowy Slopes PC',
        'Soulsand Valley PC',
        'Stone Beach PC',
        'Swamp PC',
        'Taiga PC',
        'The End PC',
        'Warped Forest PC',
    ]
};

class WorldInfo {
    constructor() {
    }
    getObjectives() {
        return [
            WORLD_INFO_OBJECTIVE,
            COMPUTER_INFO_OBJECTIVE,
            ...world
                .scoreboard
                .getObjectives()
                .filter(o => o.id !== COMPUTER_INFO_OBJECTIVE)
                .filter(o => o.id.endsWith(CUSTOM_OBJECTIVES))
                .map(o => o.id)
        ];
    }
    getStats(objective) {
        return world.scoreboard.getObjective(objective)?.getParticipants().map(p => p.displayName) ?? [];
    }
}
const worldInfo = new WorldInfo();
const initWorldInfo = () => {
    // Init trackers
    world.afterEvents.entityDie.subscribe(async () => {
        world.scoreboard.getObjective('jig_ccomp:world_info')?.addScore('jig_ccomp:deadentities', 1);
    });
    //world.afterEvents.entityHurt.subscribe((event) => {
    //  const sourceEntity = event.hurtEntity;
    //	let health = (sourceEntity.getComponent(EntityHealthComponent.componentId) as EntityHealthComponent).currentValue
    //	let inventory = (sourceEntity.getComponent(EntityInventoryComponent.componentId) as EntityInventoryComponent).container
    //
    //	world.sendMessage('my health is '+health)
    //	world.sendMessage('my inventory is '+inventory)
    //});
    //world.afterEvents.entityDie.subscribe(async (event) => {
    //  const sourceEntity = event.deadEntity;
    //	let health = (sourceEntity.getComponent(EntityHealthComponent.componentId) as EntityHealthComponent).currentValue
    //	let inventory = (sourceEntity.getComponent(EntityInventoryComponent.componentId) as EntityInventoryComponent).container
    //	world.sendMessage('now im dead '+health)
    //});
    world.afterEvents.playerPlaceBlock.subscribe(async () => {
        world.scoreboard.getObjective('jig_ccomp:world_info')?.addScore('jig_ccomp:playerplaceblock', 1);
    });
    world.afterEvents.playerBreakBlock.subscribe(async () => {
        world.scoreboard.getObjective('jig_ccomp:world_info')?.addScore('jig_ccomp:playerbreakblock', 1);
    });
    world.afterEvents.buttonPush.subscribe(async () => {
        world.scoreboard.getObjective('jig_ccomp:world_info')?.addScore('jig_ccomp:buttonpush', 1);
    });
    world.afterEvents.leverAction.subscribe(async () => {
        world.scoreboard.getObjective('jig_ccomp:world_info')?.addScore('jig_ccomp:leveraction', 1);
    });
    world.afterEvents.pressurePlatePush.subscribe(async () => {
        world.scoreboard.getObjective('jig_ccomp:world_info')?.addScore('jig_ccomp:pressureplatepush', 1);
    });
    world.afterEvents.tripWireTrip.subscribe(async () => {
        world.scoreboard.getObjective('jig_ccomp:world_info')?.addScore('jig_ccomp:tripwiretrip', 1);
    });
    function convertToDays(timeUnits) {
        const secondsPerUnit = 3.6;
        return Math.floor(secondsPerUnit / (60 * 60 * 24));
    }
    function convertToTimeOfDay(timeUnits) {
        const baseTime = new Date(0);
        const baseTimeHours = 6;
        const millisecondsPerUnit = 3.6 * 1000; // 1 unit is 3.6 seconds
        const totalMilliseconds = timeUnits * millisecondsPerUnit;
        const shiftedTime = new Date(baseTime.getTime() + totalMilliseconds);
        const hours = ((shiftedTime.getUTCHours() + baseTimeHours) % 24).toString().padStart(2, '0');
        const minutes = shiftedTime.getUTCMinutes().toString().padStart(2, '0');
        const seconds = shiftedTime.getUTCSeconds().toString().padStart(2, '0');
        const milliseconds = shiftedTime.getUTCMilliseconds().toString().padStart(2, '0').slice(0, 2);
        return `${hours}§7:§r${minutes}§7:§r${seconds}§7.§r${milliseconds}`;
    }
    // Init custom world info
    system.afterEvents.scriptEventReceive.subscribe((event) => {
        const player = event.initiator;
        const sourceEntity = event.sourceEntity;
        let timeUnits = world.getTimeOfDay();
        let TimeOfDay = convertToTimeOfDay(timeUnits);
        let worldDay = world.getDay();
        const timeAfterInstalling = convertToDays(world.getAbsoluteTime() - GET_RAW_PROP(world, Config.world.installedTime));
        if (player && player instanceof Player && sourceEntity && CraftComputer.isComputer(sourceEntity)) {
            const computer = new CraftComputer(sourceEntity);
            if (event.id === 'jig_ccomp:show_stats') {
                computer.computer.triggerEvent('jig_ccomp:typing');
                player.playSound('jig_common.robot_bleeps');
                system.runTimeout(async () => {
                    while (true) {
                        const response = await viewBoardsList(worldInfo.getObjectives()).open({ ...defaultScriptDialogueOptions, player });
                        if (response instanceof InputScriptDialogueResponse) {
                            const { board } = response.values;
                            const objective = world.scoreboard.getObjective(board);
                            if (!objective) {
                                return [];
                            }
                            let test = objective.getParticipants();
                            if (test.length == 0) {
                                sourceEntity?.runCommand("/event entity @s jig_ccomp:screen_stac");
                            }
                            else {
                                player.playSound('jig_common.robot_bleeps');
                                sourceEntity?.runCommand("/event entity @s jig_ccomp:screen_stab");
                            }
                            sourceEntity?.runCommand("/event entity @s jig_ccomp:typing");
                            if (board && board) {
                                const innerResponse = await viewStats(board, TimeOfDay, worldDay, timeAfterInstalling, worldInfo.getStats(board)).open({ ...defaultScriptDialogueOptions, player });
                                computer.computer.triggerEvent('jig_ccomp:typing');
                                if (innerResponse instanceof ButtonDialogueResponse) {
                                    if (innerResponse.selected === "other-stats") {
                                        player.playSound('jig_common.robot_bleeps');
                                        sourceEntity?.runCommand("/event entity @s jig_ccomp:screen_stat");
                                        sourceEntity?.runCommand("/event entity @s jig_ccomp:typing");
                                        continue;
                                    }
                                }
                            }
                        }
                        break;
                    }
                }, 10);
            }
        }
    }, {
        namespaces: ['jig_ccomp']
    });
};

const viewSelectRadar = () => {
    return inputScriptDialogue(TRANSLATE("jig_ccomp:select-radar.title"))
        .addElement(inputDropdown("radar", RAW_TEXT(TEXT("\n"), TRANSLATE("jig_ccomp:select-radar.radar", "\n\n")))
        .addOption(TRANSLATE("jig_ccomp:select-radar.radar.computer"), "computer")
        .addOption(TRANSLATE("jig_ccomp:select-radar.radar.everything"), "everything")
        .addOption(TRANSLATE("jig_ccomp:select-radar.radar.item"), "item")
        .addOption(TRANSLATE("jig_ccomp:select-radar.radar.mob"), "mob")
        .addOption(TRANSLATE("jig_ccomp:select-radar.radar.monster"), "monster")
        .addOption(TRANSLATE("jig_ccomp:select-radar.radar.playercheck"), "playercheck")
        .addOption(TRANSLATE("jig_ccomp:select-radar.radar.rideable"), "rideable")
        .addOption(TRANSLATE("jig_ccomp:select-radar.radar.tamed"), "tamed")
        .addOption(TRANSLATE("jig_ccomp:select-radar.radar.villager"), "villager")
        .setDefaultValueIndex(4))
        .addElement(inputToggle("sort", TRANSLATE("jig_ccomp:select-radar.sort-closest"), false))
        .addElement(inputSlider("count", TRANSLATE("jig_ccomp:select-radar.count"), 1, 50, 1, 10));
};

const viewRadar = (radar, entries, player) => {
    const showPrintButton = (entries.filter(e => !e.protected).length > 0);
    let moreOptionsButtons = multiButtonScriptDialogue(TRANSLATE("jig_ccomp:view-radar.more-options.title"))
        .setBody(RAW_TEXT(TEXT("\n"), TRANSLATE(`jig_ccomp:view-radar.more-options.body`)))
        .addButton("other-radar", TRANSLATE("jig_ccomp:view-radar.other-radar"), "textures/jig/ccomp/myui/radar.png", async (selected) => {
        return {
            selectedOther: true
        };
    });
    if (showPrintButton) {
        moreOptionsButtons = moreOptionsButtons.addButton("print-entry", TRANSLATE("jig_ccomp:view-radar.print-entry"), "textures/jig/ccomp/myui/print.png", async (selected) => {
            const response = await multiButtonScriptDialogue(TRANSLATE("jig_ccomp:view-radar.print-entry.title"))
                .setBody(RAW_TEXT(TEXT("\n"), TRANSLATE("jig_ccomp:view-radar.print-entry.body", "\n")))
                .addButtons(entries
                .map((e, index) => ({ ...e, index }))
                .filter(e => !e.protected)
                .map((e) => ({
                name: e.index.toString(),
                text: RAW_TEXT(TEXT(`${e.index + 1}: `), getEntityRawMessage(e), TEXT(` Distance: ${e.distance}`))
            }))).open({ player });
            if (response instanceof ButtonDialogueResponse) {
                return {
                    selectedPrintEntry: response.selected
                };
            }
        });
    }
    moreOptionsButtons = moreOptionsButtons.addButton("quit", TRANSLATE("jig_ccomp:quit"));
    return dualButtonScriptDialogue(RAW_TEXT(TRANSLATE("jig_ccomp:view-radar.title")), {
        name: "more-options",
        text: TRANSLATE("jig_ccomp:view-radar.more-options"),
        callback: async (selected) => {
            const response = await moreOptionsButtons.open({
                ...defaultScriptDialogueOptions,
                player
            });
            if (response instanceof ButtonDialogueResponse) {
                return {
                    ...response.callback,
                    selectedOptions: "more-options"
                };
            }
        }
    }, {
        name: "quit",
        text: TRANSLATE("jig_ccomp:quit"),
    })
        .setBody(RAW_TEXT(TEXT("§l"), TRANSLATE(`jig_ccomp:view-radar.${radar}`), TEXT("§r"), TEXT("\n"), ...buildEntries(entries), TEXT("\n"), entries.some(e => e.protected) ? TRANSLATE("jig_ccomp:view-radar.chainmail-warning") : TEXT("")));
};
const moreround = (value) => {
    return Math.ceil(value);
};
const buildEntries = (entries) => {
    if (entries.length === 0) {
        return [
            TRANSLATE("jig_ccomp:view-radar.not-found")
        ];
    }
    return entries.map(e => buildEntry(e));
};
const getEntityRawMessage = (entry) => {
    if (entry.id == "minecraft:player") {
        return ENTITY("jig_ccomp:fake_player");
    }
    if (entry.item) {
        const message = [
            ENTITY(entry.id),
            TEXT(" "),
            ITEM(entry.item),
        ];
        if (entry.item.isStackable) {
            message.push(TEXT(` x${entry.item.amount}`));
        }
        return RAW_TEXT(...message);
    }
    return ENTITY(entry.id);
};
const buildEntry = (entry) => {
    const messages = [];
    if (entry.protected) {
        messages.push(TEXT("§k"));
    }
    messages.push(getEntityRawMessage(entry));
    if (entry.nametag) {
        messages.push(TEXT(" "));
        messages.push(TRANSLATE("jig_ccomp:view-radar.entity.name"));
        messages.push(TEXT(" " + entry.nametag));
    }
    if (entry.health !== undefined) {
        messages.push(TEXT(` \uE10C ${moreround(entry.health)}`));
    }
    messages.push(TEXT("\n"));
    messages.push(TEXT(`x: ${moreround(entry.location.x)}, y: ${moreround(entry.location.y)}, z: ${moreround(entry.location.z)}`));
    messages.push(TEXT(" "));
    messages.push(TRANSLATE("jig_ccomp:view-radar.distance"));
    messages.push(TEXT(` ${moreround(entry.distance)}`));
    messages.push(TEXT("\n\n"));
    if (entry.protected) {
        messages.push(TEXT("§r"));
    }
    return RAW_TEXT(...messages);
};

const CHARS_PER_LINE = 40;
const printContentFavoriteEmail = (title, content, computer, paper) => {
    const printedPaper = new ItemStack(paper, 1);
    printedPaper.nameTag = title;
    writeContent(printedPaper, content);
    return printedPaper;
};
const printContent = (title, content, computer, player, paper) => {
    computer.computer.triggerEvent('jig_ccomp:typing');
    const printers = computer.computer.dimension.getEntities({
        families: ["jig_ccomp:printer"],
        closest: 1,
        maxDistance: 10,
        location: computer.computer.location
    });
    if (printers.length === 0) {
        computer.computer.runCommand(`/dialogue open @s "${player.name}" jig_ccomp:printer_none`);
        debug("Could not find a printer");
        return;
    }
    const printer = printers[0];
    const container = printer.getComponent(EntityInventoryComponent.componentId).container;
    const hasEmptySpace = container.emptySlotsCount > 0;
    const paperSlot = getSlotByCondition(container, (item) => {
        if (item.typeId === "minecraft:paper" && item.getLore().length === 0) {
            return hasEmptySpace || item.amount === 1;
        }
        return false;
    });
    if (paperSlot === undefined) {
        if (hasEmptySpace) {
            computer.computer.runCommand(`/dialogue open @s "${player.name}" jig_ccomp:printer_pape`);
            debug("The printer is out of paper");
        }
        else {
            computer.computer.runCommand(`/dialogue open @s "${player.name}" jig_ccomp:printer_full`);
            debug("Found no space to put the new paper");
        }
        return;
    }
    const item = container.getItem(paperSlot);
    const printedPaper = new ItemStack(paper, 1);
    if (item.amount === 1) {
        container.setItem(paperSlot);
    }
    else {
        item.amount = item.amount - 1;
        container.setItem(paperSlot, item);
    }
    printedPaper.nameTag = title;
    writeContent(printedPaper, content);
    printer.triggerEvent('jig_ccomp:printing');
    const emptySlot = getEmptySlot(container);
    container.setItem(emptySlot, printedPaper);
    return printedPaper;
};
const writeContent = (paper, content) => {
    const loreList = [];
    content.split("\n").forEach(line => {
        do {
            const result = getNextLine(line);
            line = result[1];
            loreList.push(result[0]);
        } while (line.length > 0);
    });
    paper.setLore(loreList);
};
const getNextLine = (content) => {
    let nextLine = '';
    const words = content.split(' ');
    while (words.length > 0) {
        if (nextLine.length + words[0].length <= CHARS_PER_LINE) {
            nextLine = `${nextLine} ${words.shift()}`.trim();
        }
        else if (words[0].length > CHARS_PER_LINE) {
            nextLine = `${nextLine} ${words[0].substring(0, CHARS_PER_LINE)}`.trim();
            words[0] = words[0].substring(CHARS_PER_LINE);
        }
        else {
            break;
        }
    }
    return [nextLine, words.join(' ')];
};

const ADMIN_SETTINGS_KEY = 'jig_ccomp:admin-settings';
const SettingsKeys = [
    'radar',
    'weather',
    'triviaPrizes',
    'mobEmails'
];
const getSettings = () => {
    const settings = GET_PROP(world, ADMIN_SETTINGS_KEY, {});
    SettingsKeys.forEach(sk => {
        if (settings[sk] === undefined) {
            settings[sk] = true;
        }
    });
    return settings;
};
const initAdminConfigMenu = () => {
    system.afterEvents.scriptEventReceive.subscribe((event) => {
        debug(JSON.stringify(event));
        if (event.sourceEntity instanceof Player) {
            if (event.id === 'jig_ccomp:admin-menu') {
                void openMenu(event.sourceEntity);
            }
        }
    });
};
const openMenu = async (player) => {
    const settings = getSettings();
    const response = await adminSettingsMenuDialogue(player, settings);
    if (response instanceof InputScriptDialogueResponse) {
        SET_PROP(world, ADMIN_SETTINGS_KEY, response.values);
    }
};
const adminSettingsMenuDialogue = async (player, settings) => {
    return await inputScriptDialogue(TRANSLATE('jig_ccomp:admin-menu.title'))
        .addElements(SettingsKeys.map(sk => inputToggle(sk, TRANSLATE(`jig_ccomp:admin-menu.setting.${sk}`), !!settings[sk])))
        .open({
        player,
        ...defaultScriptDialogueOptions
    });
};
const isAdminSettingsEnabled = (key) => {
    const settings = getSettings();
    return settings[key];
};
const disabledAdminDialogue = async (key, player) => {
    return multiButtonScriptDialogue(TRANSLATE("jig_ccomp:admin-menu.disabled.title"))
        .setBody(TRANSLATE("jig_ccomp:admin-menu.disabled.body", TEXT("\n\n"), TRANSLATE(`jig_ccomp:admin-menu.disabled.${key}`)))
        .addButton("back", TRANSLATE("jig_ccomp:admin-menu.disabled.back"))
        .open({
        player,
        ...defaultScriptDialogueOptions
    });
};

const MOB_PAPER = "jig_ccomp:mob_paper";
const MOBN_PAPER = "jig_ccomp:mobn_paper";
const MOBE_PAPER = "jig_ccomp:mobe_paper";
const GOOD_BUG_PAPER = "jig_ccomp:good_bug_paper";
const OK_BUG_PAPER = "jig_ccomp:ok_bug_paper";
const OLD_BUG_PAPER = "jig_ccomp:old_bug_paper";
const TRIVIA_JIG_PAPER = "jig_ccomp:trivia_jig_paper";
const TRIVIA_MIN_PAPER = "jig_ccomp:trivia_min_paper";
const TNT_SWEEPER_1 = "jig_ccomp:tnt_paper_1";
const TNT_SWEEPER_2 = "jig_ccomp:tnt_paper_2";
const TNT_SWEEPER_3 = "jig_ccomp:tnt_paper_3";
const validPaperTypes = [MOB_PAPER, MOBN_PAPER, MOBE_PAPER];
const particleradar = "radar";
const particleradarn = "radarn";
const particleradare = "radare";
let radarchoice = particleradar;
let PAPERCHOICE = MOB_PAPER;
const initRadars = () => {
    world.afterEvents.itemUse.subscribe((event) => {
        let player = event.source;
        const { source, itemStack } = event;
        if (validPaperTypes.includes(itemStack.typeId)) {
            for (const line of itemStack.getLore()) {
                if (line.startsWith("Dimension: ")) {
                    const dimensionName = line.substring("Dimension: ".length);
                    if (`minecraft:${dimensionName}` !== player.dimension.id) {
                        player.runCommandAsync(`/playsound jig_ccomp.radar @a ~ ~ ~ 0.5 0.5`);
                        player.runCommandAsync(`/tellraw @s {"rawtext":[{"text":"§e${itemStack.getLore()[0]} was in the §a${dimensionName}"}]}`);
                        break;
                    }
                }
                if (line.startsWith("x:")) {
                    const indexOfY = line.indexOf("y:");
                    const indexOfZ = line.indexOf("z:");
                    const x = parseInt(line.substring("x:".length, indexOfY - 2));
                    const y = parseInt(line.substring(indexOfY + 2, indexOfZ - 2));
                    const z = parseInt(line.substring(indexOfZ + 2));
                    const target = {
                        x,
                        y,
                        z
                    };
                    getDistance(target, event.source.location);
                    const verdistance = getVerDistance(target, event.source.location);
                    const hordistance = getHorDistance(target, event.source.location);
                    const direction = getDirection(event.source.location, target);
                    let directionString = '';
                    if (direction.includes(0 /* NEXT_TO_YOU */)) {
                        directionString = 'Next to you!';
                    }
                    else {
                        if (direction.includes(5 /* ABOVE */)) {
                            if (directionString !== "") {
                                directionString += " and ";
                            }
                            directionString += verdistance + " blocks above you";
                        }
                        if (direction.includes(6 /* BELOW */)) {
                            if (directionString !== "") {
                                directionString += " and ";
                            }
                            directionString += verdistance + " blocks below you";
                        }
                        if (hordistance >= 4) {
                            let and = "";
                            if (itemStack.typeId === MOB_PAPER) {
                                radarchoice = particleradar;
                            }
                            else if (itemStack.typeId === MOBN_PAPER) {
                                radarchoice = particleradarn;
                            }
                            else if (itemStack.typeId === MOBE_PAPER) {
                                radarchoice = particleradare;
                            }
                            if ((verdistance < 0 && verdistance <= -2) || (verdistance > 0 && verdistance >= 5)) {
                                and = " and ";
                            }
                            directionString += and + hordistance + " blocks to the ";
                            if (direction.includes(1 /* NORTH */)) {
                                directionString += "north";
                                if (direction.includes(4 /* EAST */)) {
                                    player.runCommandAsync("/particle jig_ccomp:" + radarchoice + "_ne");
                                }
                                else if (direction.includes(3 /* WEST */)) {
                                    player.runCommandAsync("/particle jig_ccomp:" + radarchoice + "_nw");
                                }
                                else {
                                    player.runCommandAsync("/particle jig_ccomp:" + radarchoice + "_n");
                                }
                            }
                            if (direction.includes(2 /* SOUTH */)) {
                                directionString += "south";
                                if (direction.includes(4 /* EAST */)) {
                                    player.runCommandAsync("/particle jig_ccomp:" + radarchoice + "_se");
                                }
                                else if (direction.includes(3 /* WEST */)) {
                                    player.runCommandAsync("/particle jig_ccomp:" + radarchoice + "_sw");
                                }
                                else {
                                    player.runCommandAsync("/particle jig_ccomp:" + radarchoice + "_s");
                                }
                            }
                            if (direction.includes(3 /* WEST */)) {
                                directionString += "west";
                                if (direction.includes(1 /* NORTH */)) {
                                    player.runCommandAsync("/particle jig_ccomp:" + radarchoice + "_nw");
                                }
                                else if (direction.includes(2 /* SOUTH */)) {
                                    player.runCommandAsync("/particle jig_ccomp:" + radarchoice + "_sw");
                                }
                                else {
                                    player.runCommandAsync("/particle jig_ccomp:" + radarchoice + "_w");
                                }
                            }
                            if (direction.includes(4 /* EAST */)) {
                                directionString += "east";
                                if (direction.includes(1 /* NORTH */)) {
                                    player.runCommandAsync("/particle jig_ccomp:" + radarchoice + "_ne");
                                }
                                else if (direction.includes(2 /* SOUTH */)) {
                                    player.runCommandAsync("/particle jig_ccomp:" + radarchoice + "_se");
                                }
                                else {
                                    player.runCommandAsync("/particle jig_ccomp:" + radarchoice + "_e");
                                }
                            }
                        }
                    }
                    player.runCommandAsync(`/tellraw @s {"rawtext":[{"text":"§e${itemStack.getLore()[0]} was at: \n§a${line}. \n§6Direction: §r${directionString}."}]}`);
                    player.runCommandAsync(`/playsound jig_ccomp.radar @a ~ ~ ~ 0.5 1.5`);
                    break;
                }
            }
        }
        else if (itemStack.typeId === GOOD_BUG_PAPER) {
            for (const line of itemStack.getLore()) {
                player.playSound("jig_ccomp.bug_scorecard");
            }
        }
        else if (itemStack.typeId === OK_BUG_PAPER) {
            for (const line of itemStack.getLore()) {
                player.playSound("jig_ccomp.bug_hit");
            }
        }
        else if (itemStack.typeId === OLD_BUG_PAPER) {
            for (const line of itemStack.getLore()) {
                player.playSound("jig_ccomp.bug_lose");
            }
        }
        else if (itemStack.typeId === TRIVIA_JIG_PAPER) {
            for (const line of itemStack.getLore()) {
                player.playSound("jig_common.dojig");
            }
        }
        else if (itemStack.typeId === TRIVIA_MIN_PAPER) {
            for (const line of itemStack.getLore()) {
                player.playSound("jig_ccomp.trivia_start");
            }
        }
        else if (itemStack.typeId === TNT_SWEEPER_1) {
            for (const line of itemStack.getLore()) {
                player.playSound("mob.creeper.say");
            }
        }
        else if (itemStack.typeId === TNT_SWEEPER_2) {
            for (const line of itemStack.getLore()) {
                player.dimension.spawnParticle("minecraft:large_explosion", player.location);
                player.dimension.spawnParticle("minecraft:lava_particle", player.location);
                player.playSound("random.explode");
            }
        }
        else if (itemStack.typeId === TNT_SWEEPER_3) {
            for (const line of itemStack.getLore()) {
                player.playSound("random.levelup");
            }
        }
    });
    system.afterEvents.scriptEventReceive.subscribe((event) => {
        const player = event.initiator;
        const sourceEntity = event.sourceEntity;
        if (player && player instanceof Player && sourceEntity && CraftComputer.isComputer(sourceEntity)) {
            if (event.id === 'jig_ccomp:show_radar') {
                let playerSoundOptions = {
                    pitch: makeRandomNumberDecimal(1, 1.5),
                    volume: 1.0,
                };
                player.playSound('jig_ccomp.radar', playerSoundOptions);
                sourceEntity?.runCommand("/event entity @s jig_ccomp:typing");
                system.runTimeout(async () => {
                    if (!isAdminSettingsEnabled("radar")) {
                        disabledAdminDialogue("radar", player);
                        return;
                    }
                    while (true) {
                        const response = await viewSelectRadar().open({ ...defaultScriptDialogueOptions, player });
                        if (response instanceof InputScriptDialogueResponse) {
                            let playerSoundOptions = {
                                pitch: makeRandomNumberDecimal(1, 1.5),
                                volume: 1.0,
                            };
                            player.playSound('jig_ccomp.radar', playerSoundOptions);
                            sourceEntity?.runCommand("/event entity @s jig_ccomp:typing");
                            const { radar, sort, count } = response.values;
                            const selectedRadar = radar;
                            let radarEntries;
                            switch (selectedRadar) {
                                case "computer":
                                    radarEntries = getComputers(sourceEntity, count);
                                    break;
                                case "everything":
                                    radarEntries = getEverything(sourceEntity, count);
                                    break;
                                case "item":
                                    radarEntries = getItems(sourceEntity, count);
                                    break;
                                case "mob":
                                    radarEntries = getMob(sourceEntity, count);
                                    break;
                                case "monster":
                                    radarEntries = getMonsters(sourceEntity, count);
                                    break;
                                case "tamed":
                                    radarEntries = getTamed(sourceEntity, count);
                                    break;
                                case "villager":
                                    radarEntries = getVillager(sourceEntity, count);
                                    break;
                                case "rideable":
                                    radarEntries = getRidable(sourceEntity, count);
                                    break;
                                case "playercheck":
                                    radarEntries = getPlayers(sourceEntity, count);
                                    break;
                                default:
                                    radarEntries = [];
                            }
                            if (sort === true) {
                                radarEntries = radarEntries.reverse();
                            }
                            const innerResponse = await viewRadar(selectedRadar, radarEntries, player).open({
                                ...defaultScriptDialogueOptions,
                                player
                            });
                            if (innerResponse instanceof ButtonDialogueResponse) {
                                sourceEntity?.runCommand("/event entity @s jig_ccomp:typing");
                                if (innerResponse.selected === "more-options") {
                                    let playerSoundOptions = {
                                        pitch: makeRandomNumberDecimal(1, 1.5),
                                        volume: 1.0,
                                    };
                                    player.playSound('jig_ccomp.radar', playerSoundOptions);
                                    if (innerResponse.callback && "selectedOther" in innerResponse.callback) {
                                        continue;
                                    }
                                    else if (innerResponse.callback && "selectedPrintEntry" in innerResponse.callback) {
                                        sourceEntity?.runCommand("/event entity @s jig_ccomp:typing");
                                        if (player.dimension.id === "minecraft:nether") {
                                            PAPERCHOICE = MOB_PAPER;
                                        }
                                        if (player.dimension.id === "minecraft:nether") {
                                            PAPERCHOICE = MOBN_PAPER;
                                        }
                                        if (player.dimension.id === "minecraft:the_end") {
                                            PAPERCHOICE = MOBE_PAPER;
                                        }
                                        const entryToPrint = radarEntries[parseInt(innerResponse.callback.selectedPrintEntry)];
                                        printContent(`Radar Result`, `Entity ID ${entryToPrint.id.split(":")[1]} ${entryToPrint.item ? getEntryItemPrint(entryToPrint.item) : ''}\nDimension: ${player.dimension.id.replace("minecraft:", "")}\nx: ${Math.ceil(entryToPrint.location.x)}, y: ${Math.ceil(entryToPrint.location.y)}, z: ${Math.ceil(entryToPrint.location.z)}`, new CraftComputer(sourceEntity), player, PAPERCHOICE);
                                    }
                                }
                            }
                        }
                        break;
                    }
                }, 10);
            }
        }
    }, {
        namespaces: ['jig_ccomp']
    });
};
const getEntryItemPrint = (item) => {
    let itemPrint = item.typeId.split(":")[1];
    if (item.isStackable) {
        itemPrint += ` x${item.amount}`;
    }
    return itemPrint;
};
const resolveItem = (entity) => {
    const itemComponent = entity.getComponent(EntityItemComponent.componentId);
    if (itemComponent) {
        return itemComponent.itemStack;
    }
    return undefined;
};
const isProtected = (entity) => {
    if (entity.hasTag(Config.radar.tags.chain_protect)) {
        return true;
    }
    if (entity.hasComponent(EntityEquippableComponent.componentId)) {
        const equipment = entity.getComponent(EntityEquippableComponent.componentId);
        if (equipment) {
            const items = [
                equipment.getEquipment(EquipmentSlot.Chest)?.typeId ?? '',
                equipment.getEquipment(EquipmentSlot.Feet)?.typeId ?? '',
                equipment.getEquipment(EquipmentSlot.Head)?.typeId ?? '',
                equipment.getEquipment(EquipmentSlot.Legs)?.typeId ?? ''
            ];
            return items.some(item => item.includes('chainmail'));
        }
    }
    return false;
};
const toRadarEntry = (entity, fromEntity) => {
    if (entity.typeId == "jig_ccomp:computer_old" || entity.typeId == "jig_ccomp:computer_ok" || entity.typeId == "jig_ccomp:computer_good" || entity.typeId == "jig_ccomp:computer_lap" || entity.typeId == "jig_ccomp:portable") {
        return {
            id: entity.typeId,
            item: undefined,
            nametag: "",
            health: entity.getComponent(EntityHealthComponent.componentId).currentValue,
            dimension: entity.dimension.id,
            location: entity.location,
            distance: getDistance(entity.location, fromEntity.location),
            protected: isProtected(entity)
        };
    }
    else
        return {
            id: entity.typeId,
            item: resolveItem(entity),
            nametag: entity.nameTag,
            health: entity.getComponent(EntityHealthComponent.componentId)?.currentValue,
            dimension: entity.dimension.id,
            location: entity.location,
            distance: getDistance(entity.location, fromEntity.location),
            protected: isProtected(entity)
        };
};
const getComputers = (fromEntity, closest) => {
    const entities = getEntities(fromEntity, {
        location: fromEntity.location,
        families: ['jig_ccomp:computer'],
    }, closest);
    return entities.map(entity => {
        const radarEntry = toRadarEntry(entity, fromEntity);
        return radarEntry;
    });
};
const getEverything = (fromEntity, closest) => {
    const entities = getEntities(fromEntity, {
        location: fromEntity.location,
    }, closest);
    return entities.map(entity => {
        const radarEntry = toRadarEntry(entity, fromEntity);
        return radarEntry;
    });
};
const getMob = (fromEntity, closest) => {
    const entities = getEntities(fromEntity, {
        location: fromEntity.location,
        families: ['mob'],
    }, closest);
    return entities.map(entity => {
        const radarEntry = toRadarEntry(entity, fromEntity);
        return radarEntry;
    });
};
const getMonsters = (fromEntity, closest) => {
    const entities = getEntities(fromEntity, {
        location: fromEntity.location,
        families: ['monster'],
    }, closest);
    return entities.map(entity => {
        const radarEntry = toRadarEntry(entity, fromEntity);
        return radarEntry;
    });
};
const getVillager = (fromEntity, closest) => {
    const entitiesVillager = getEntities(fromEntity, {
        location: fromEntity.location,
        families: ['villager'],
    }, closest);
    const entitiesWanderingTrader = getEntities(fromEntity, {
        location: fromEntity.location,
        families: ['wandering_trader'],
    }, closest);
    const allEntities = [...entitiesVillager, ...entitiesWanderingTrader];
    return allEntities.map(entity => {
        const radarEntry = toRadarEntry(entity, fromEntity);
        return radarEntry;
    });
};
const getTamed = (fromEntity, count) => {
    const entities = getEntities(fromEntity, {
        location: fromEntity.location
    });
    const tamedEntities = [];
    for (const entity of entities) {
        if (entity.hasComponent(EntityIsTamedComponent.componentId)) {
            const radarEntry = toRadarEntry(entity, fromEntity);
            tamedEntities.push(radarEntry);
            if (tamedEntities.length >= count) {
                return tamedEntities;
            }
        }
    }
    return tamedEntities;
};
const getRidable = (fromEntity, count) => {
    const entities = getEntities(fromEntity, {
        location: fromEntity.location,
        excludeTypes: ['player'],
        excludeFamilies: ['monster'],
    });
    const ridableEntities = [];
    const componentId = "minecraft:rideable";
    for (const entity of entities) {
        if (entity.hasComponent(componentId)) {
            const radarEntry = toRadarEntry(entity, fromEntity);
            ridableEntities.push(radarEntry);
            if (ridableEntities.length >= count) {
                return ridableEntities;
            }
        }
    }
    return ridableEntities;
};
const getPlayers = (fromEntity, closest) => {
    const entities = getEntities(fromEntity, {
        location: fromEntity.location,
        families: ['player'],
    }, closest);
    return entities.map(entity => {
        const radarEntry = toRadarEntry(entity, fromEntity);
        return radarEntry;
    });
};
const getItems = (fromEntity, closest) => {
    const entities = getEntities(fromEntity, {
        location: fromEntity.location,
        type: "minecraft:item",
    }, closest);
    return entities.map(entity => toRadarEntry(entity, fromEntity));
};
/**
 * Gets the entities, performing additional filter on the dimension.
 * This is due to a possible bug in minecraft Dimension#getEntities where it finds stuff in any dimension.
 * @param fromEntity
 * @param options
 * @param max
 */
const getEntities = (fromEntity, options, max) => {
    let entities = fromEntity.dimension.getEntities({
        ...options,
        excludeTags: [
            ...(options.excludeTags ?? []),
            Config.radar.tags.radar_protect
        ]
    });
    // Only entities in the dimension
    entities = entities.filter(e => e.dimension === fromEntity.dimension);
    if (max) {
        entities = entities.slice(0, max);
    }
    return entities;
};

const WEATHER_KEYS = {
    [WeatherType.Clear]: "jig_ccomp:weather.clear",
    [WeatherType.Rain]: "jig_ccomp:weather.raining",
    [WeatherType.Thunder]: "jig_ccomp:weather.thunder",
};
const MOON_PHASE_KEYS = {
    [MoonPhase.FullMoon]: "jig_ccomp:moon.full-moon",
    [MoonPhase.WaningGibbous]: "jig_ccomp:moon.waning-gibbous",
    [MoonPhase.FirstQuarter]: "jig_ccomp:moon.first-quarter",
    [MoonPhase.WaningCrescent]: "jig_ccomp:moon.waning-crescent",
    [MoonPhase.NewMoon]: "jig_ccomp:moon.new-moon",
    [MoonPhase.WaxingCrescent]: "jig_ccomp:moon.waxing-crescent",
    [MoonPhase.LastQuarter]: "jig_ccomp:moon.last-quarter",
    [MoonPhase.WaxingGibbous]: "jig_ccomp:moon.waxing-gibbous",
};
const viewWeatherForecast = (options) => {
    const baseOptions = RAW_TEXT(TEXT("\n"), TRANSLATE("jig_ccomp:view-weather-forecast.subtitle"), TEXT("\n\n"), TRANSLATE("jig_ccomp:view-weather-forecast.moon-phase"), TEXT("\n"), getMoonPhaseLabel(options.time), TEXT("\n"), TRANSLATE(MOON_PHASE_KEYS[options.moonPhase]), TEXT("\n"), TEXT("\n"));
    if (options.allowUpdates) {
        return inputScriptDialogue(TRANSLATE("jig_ccomp:view-weather-forecast.title"))
            .addElement(inputDropdown("future", RAW_TEXT(baseOptions, TRANSLATE("jig_ccomp:view-weather-forecast.future"))).addOption(TRANSLATE(WEATHER_KEYS[WeatherType.Clear]), WeatherType.Clear)
            .addOption(TRANSLATE(WEATHER_KEYS[WeatherType.Rain]), WeatherType.Rain)
            .addOption(TRANSLATE(WEATHER_KEYS[WeatherType.Thunder]), WeatherType.Thunder)
            .setDefaultValueIndex(0));
    }
    else {
        return multiButtonScriptDialogue(TRANSLATE("jig_ccomp:view-weather-forecast.title"))
            .setBody(baseOptions)
            .addButton("close", TRANSLATE("jig_ccomp:view-weather-forecast.close"));
    }
};
const getMoonPhaseLabel = (time) => {
    if (time < 11834) {
        return TRANSLATE("jig_ccomp:view-weather-forecast.tonight");
    }
    else {
        return TRANSLATE("jig_ccomp:view-weather-forecast.current");
    }
};

const initWeather = () => {
    system.afterEvents.scriptEventReceive.subscribe((event) => {
        const player = event.initiator;
        const sourceEntity = event.sourceEntity;
        if (player && player instanceof Player && sourceEntity && CraftComputer.isComputer(sourceEntity)) {
            if (event.id === 'jig_ccomp:show_weather') {
                {
                    player.playSound('ambient.weather.thunder');
                }
                sourceEntity?.runCommand("/event entity @s jig_ccomp:typing");
                system.runTimeout(async () => {
                    const response = await viewWeatherForecast({
                        moonPhase: world.getMoonPhase(),
                        time: world.getTimeOfDay(),
                        allowUpdates: isAdminSettingsEnabled("weather")
                    }).open({ ...defaultScriptDialogueOptions, player });
                    if (response instanceof InputScriptDialogueResponse) {
                        sourceEntity?.runCommand("/event entity @s jig_ccomp:typing");
                        const { future } = response.values;
                        if (future == 'Clear') {
                            player.playSound('jig_common.starnoise');
                        }
                        if (future == 'Rain') {
                            player.playSound('bucket.fill_water');
                        }
                        if (future == 'Thunder') {
                            player.playSound('ambient.weather.thunder');
                        }
                        player.dimension.setWeather(future);
                    }
                }, 10);
            }
        }
    }, {
        namespaces: [
            "jig_ccomp"
        ]
    });
};

const subscribeFromComputerNpcDialogue = (callback, options) => {
    system.afterEvents.scriptEventReceive.subscribe((event) => {
        const player = event.initiator;
        const sourceEntity = event.sourceEntity;
        if (player && player instanceof Player && sourceEntity && CraftComputer.isComputer(sourceEntity)) {
            callback({
                id: event.id,
                computer: new CraftComputer(sourceEntity),
                player,
                message: event.message
            });
        }
    }, options);
};
const executeScriptDialogueFromNpcDialogue = (callback) => {
    system.runTimeout(callback, 10);
};

const PREFILLED_SUBJECT_PREFIX = 'jig_ccomp:jig.prefilled_il.subj_';
const PREFILLED_SUBJECT_COUNT = 12;
const PREFILLED_MESSAGE_PREFIX = 'jig_ccomp:jig.prefilled_il.body_';
const PREFILLED_MESSAGE_COUNT = 72;
const sendEmailDialogue = (targets, attachments, to, subject, usePredefinedContent) => {
    let toElement = inputDropdown("to", TRANSLATE("jig_ccomp:pc-email-send.inputs.to.label"));
    targets.forEach(t => {
        toElement = toElement.addOption(t.name, t.id);
    });
    const toIndex = targets.findIndex(t => t.id === to);
    toElement = toElement.setDefaultValueIndex(toIndex === -1 ? 0 : toIndex);
    let attachmentElement = inputDropdown("attachment", TRANSLATE("jig_ccomp:pc-email-send.inputs.attachment.label"));
    attachmentElement = attachmentElement.addOption(TRANSLATE("jig_ccomp:pc-email-send.inputs.attachment.none"), "none");
    attachments.forEach((t, index) => {
        attachmentElement = attachmentElement.addOption(RAW_TEXT(t.item.nameTag ? TEXT(t.item.nameTag) : ITEM(t.item), TEXT(` ${t.item.isStackable ? `x${t.item.amount}` : ''}`.trimEnd())), index.toString());
    });
    attachmentElement = attachmentElement.setDefaultValueIndex(0);
    let subjectDropdown = inputDropdown("subject", TRANSLATE("jig_ccomp:pc-email-send.inputs.subject.label"));
    for (let i = 0; i < PREFILLED_SUBJECT_COUNT; ++i) {
        const key = PREFILLED_SUBJECT_PREFIX + i.toString().padStart(2, "0");
        subjectDropdown = subjectDropdown.addOption(TRANSLATE(key), key);
    }
    let messageDropdown = inputDropdown("message", TRANSLATE("jig_ccomp:pc-email-send.inputs.message.label"));
    for (let i = 0; i < PREFILLED_MESSAGE_COUNT; ++i) {
        const key = PREFILLED_MESSAGE_PREFIX + i.toString().padStart(2, "0");
        messageDropdown = messageDropdown.addOption(TRANSLATE(key), key);
    }
    return inputScriptDialogue(TRANSLATE("jig_ccomp:pc-email-send.title"))
        .addElement(toElement)
        .addElement(attachmentElement)
        .addElement(usePredefinedContent ? subjectDropdown :
        inputText("subject", TRANSLATE("jig_ccomp:pc-email-send.inputs.subject.label"), TRANSLATE("jig_ccomp:pc-email-send.inputs.subject.placeholder"), subject))
        .addElement(usePredefinedContent ? messageDropdown :
        inputText("message", TRANSLATE("jig_ccomp:pc-email-send.inputs.message.label"), TRANSLATE("jig_ccomp:pc-email-send.inputs.message.placeholder")));
};

const readEmailDialogue = (emails, player, computer) => {
    return multiButtonScriptDialogue(TRANSLATE("jig_ccomp:pc-email-read.title"))
        .setBody(TRANSLATE("jig_ccomp:pc-email-read.body"))
        .addButtons(emails.map(e => ({
        name: e.id,
        text: RAW_TEXT(TEXT_OR_RAW_TEXT(e.from_player), TEXT(": "), TEXT_OR_RAW_TEXT(e.subject)),
        iconPath: emailIcon(e),
        callback: async (selected) => {
            computer.computer.triggerEvent('jig_ccomp:typing');
            const response = await dualButtonScriptDialogue(RAW_TEXT(TEXT("Subject: "), TEXT_OR_RAW_TEXT(e.subject)), {
                name: "more",
                text: TRANSLATE("jig_ccomp:pc-email-read.button.more"),
                callback: async (selected) => {
                    computer.computer.triggerEvent('jig_ccomp:typing');
                    const buttons = [];
                    if (e.sender_type !== "script" && e.sender_type !== "system") {
                        buttons.push({
                            name: "reply-custom",
                            text: TRANSLATE("jig_ccomp:pc-email-read-more.button.reply-custom"),
                            iconPath: "textures/jig/ccomp/myui/reply.png"
                        });
                        buttons.push({
                            name: "reply",
                            text: TRANSLATE("jig_ccomp:pc-email-read-more.button.reply"),
                            iconPath: "textures/jig/ccomp/myui/reply.png"
                        });
                    }
                    if (e.printable) {
                        buttons.push({
                            name: "print",
                            text: TRANSLATE("jig_ccomp:pc-email-read-more.button.print"),
                            iconPath: "textures/jig/ccomp/myui/print.png"
                        });
                    }
                    if (e.attachment !== undefined && e.attachment !== -1) {
                        buttons.push({
                            name: "download",
                            text: TRANSLATE("jig_ccomp:pc-email-read-more.button.download"),
                            iconPath: "textures/jig/ccomp/myui/attachment.png"
                        });
                    }
                    buttons.push({
                        name: "delete",
                        text: TRANSLATE("jig_ccomp:pc-email-read-more.button.delete"),
                        iconPath: "textures/jig/ccomp/myui/trash.png"
                    });
                    if (isEmailBlockeable(e)) {
                        buttons.push({
                            name: "block",
                            text: TRANSLATE("jig_ccomp:pc-email-read-more.button.block"),
                            iconPath: "textures/jig/ccomp/myui/block.png"
                        });
                    }
                    buttons.push({
                        name: "close",
                        text: TRANSLATE("jig_ccomp:pc-email-read.button.close")
                    });
                    const moreResponse = await multiButtonScriptDialogue(RAW_TEXT(TEXT("Subject: "), TEXT_OR_RAW_TEXT(e.subject)))
                        .setBody(TRANSLATE("jig_ccomp:pc-email-read-more.body"))
                        .addButtons(buttons)
                        .open({
                        ...defaultScriptDialogueOptions,
                        player
                    });
                    if (moreResponse instanceof ButtonDialogueResponse) {
                        return moreResponse.selected;
                    }
                }
            }, {
                name: "close",
                text: TRANSLATE("jig_ccomp:pc-email-read.button.close")
            })
                .setBody(RAW_TEXT(TRANSLATE("jig_ccomp:email.from"), TEXT(": "), TEXT_OR_RAW_TEXT(e.from_player), TEXT("\n"), TRANSLATE("jig_ccomp:email.mailed-by"), TEXT(": "), TEXT_OR_RAW_TEXT(e.from_computer_name), TEXT("\n"), ...getAttachmentContent(e), TEXT("\n"), TEXT_OR_RAW_TEXT(e.message)))
                .open({
                ...defaultScriptDialogueOptions,
                player
            });
            if (response instanceof ButtonDialogueResponse) {
                return response.callback;
            }
        }
    })));
};
const getAttachmentContent = (email) => {
    switch (email.attachment) {
        case undefined:
            return [];
        case -1:
            return [TRANSLATE('jig_ccomp:email.attachment-downloaded')];
        default:
            return [TRANSLATE('jig_ccomp:email.attachment')];
    }
};
const emailIcon = (email) => {
    const hasAttachment = email.attachment !== undefined && email.attachment >= 0;
    if (email.isRead) {
        if (hasAttachment) {
            return "textures/jig/ccomp/myui/email_read_a.png";
        }
        return "textures/jig/ccomp/myui/email_read.png";
    }
    if (hasAttachment) {
        return "textures/jig/ccomp/myui/email_unread_a.png";
    }
    return "textures/jig/ccomp/myui/email_unread.png";
};
const isEmailBlockeable = (email) => {
    if (email.sender_type === "system") {
        return false;
    }
    if (email.sender_type === "npc") {
        if (email.from_player_id === Config.JigarBoxNpcId) {
            return false;
        }
    }
    return true;
};

//dont fix my formatting, it makes it easier to copy and paste all the strings of the messages
const npcRecipients = { Sword_PC: { name: TRANSLATE("entity.villager.weapon"), introduction: { message: TRANSLATE("jig_ccomp:weapon.introducn.message") }, too_many: { message: TRANSLATE("jig_ccomp:weapon.toomanyem.message") }, favoriteMailResponses: [
            TRANSLATE("jig_ccomp:weapon.favorit.0.message"),
        ], forum: [{
                subject: TRANSLATE('jig_ccomp:weapon.forum_m.0.subject'),
                message: TRANSLATE('jig_ccomp:weapon.forum_m.0.message')
            }, {
                subject: TRANSLATE('jig_ccomp:weapon.forum_m.1.subject'),
                message: TRANSLATE('jig_ccomp:weapon.forum_m.1.message')
            }], random: [
            { message: TRANSLATE("jig_ccomp:weapon.randomx.0.message") },
            { message: TRANSLATE("jig_ccomp:weapon.randomx.1.message") },
        ], attachmentResponses: [
            TRANSLATE("jig_ccomp:weapon.attachm.0.message"),
            TRANSLATE("jig_ccomp:weapon.attachm.1.message"),
        ], inventoryFullResponses: [
            TRANSLATE("jig_ccomp:weapon.invfull.0.message"),
            TRANSLATE("jig_ccomp:weapon.invfull.1.message"),
        ], responses: [
            { message: TRANSLATE("jig_ccomp:weapon.message.0.message") },
            { message: TRANSLATE("jig_ccomp:weapon.message.1.message"), attachment: ["minecraft:coal", 8] },
            { message: TRANSLATE("jig_ccomp:weapon.message.2.message"), attachment: ["minecraft:stone_axe", 1] },
            { message: TRANSLATE("jig_ccomp:weapon.message.3.message") },
            { message: TRANSLATE("jig_ccomp:weapon.message.4.message"), attachment: ["minecraft:stone_shovel", 1] },
            { message: TRANSLATE("jig_ccomp:weapon.message.5.message") },
            { message: TRANSLATE("jig_ccomp:weapon.message.6.message") },
            { message: TRANSLATE("jig_ccomp:weapon.message.7.message") },
            { message: TRANSLATE("jig_ccomp:weapon.message.8.message"), attachment: ["minecraft:diamond_sword", 1] },
        ] }, Moobile_Device: { name: TRANSLATE("entity.cow.name"), introduction: { message: TRANSLATE("jig_ccomp:cowxxx.introducn.message") }, too_many: { message: TRANSLATE("jig_ccomp:cowxxx.toomanyem.message") }, favoriteMailResponses: [
            TRANSLATE("jig_ccomp:cowxxx.favorit.0.message"),
        ], forum: [{
                subject: TRANSLATE('jig_ccomp:cowxxx.forum_m.0.subject'),
                message: TRANSLATE('jig_ccomp:cowxxx.forum_m.0.message')
            }, {
                subject: TRANSLATE('jig_ccomp:cowxxx.forum_m.1.subject'),
                message: TRANSLATE('jig_ccomp:cowxxx.forum_m.1.message')
            }], random: [
            { message: TRANSLATE("jig_ccomp:cowxxx.randomx.0.message") },
            { message: TRANSLATE("jig_ccomp:cowxxx.randomx.1.message") },
        ], attachmentResponses: [
            TRANSLATE("jig_ccomp:cowxxx.attachm.0.message"),
            TRANSLATE("jig_ccomp:cowxxx.attachm.1.message"),
        ], inventoryFullResponses: [
            TRANSLATE("jig_ccomp:cowxxx.invfull.0.message"),
            TRANSLATE("jig_ccomp:cowxxx.invfull.1.message"),
        ], responses: [
            { message: TRANSLATE("jig_ccomp:cowxxx.message.0.message") },
            { message: TRANSLATE("jig_ccomp:cowxxx.message.1.message") },
            { message: TRANSLATE("jig_ccomp:cowxxx.message.2.message"), attachment: ["minecraft:milk_bucket", 1] },
            { message: TRANSLATE("jig_ccomp:cowxxx.message.3.message") },
            { message: TRANSLATE("jig_ccomp:cowxxx.message.4.message") },
            { message: TRANSLATE("jig_ccomp:cowxxx.message.5.message") },
            { message: TRANSLATE("jig_ccomp:cowxxx.message.6.message") },
            { message: TRANSLATE("jig_ccomp:cowxxx.message.7.message") },
            { message: TRANSLATE("jig_ccomp:cowxxx.message.8.message"), attachment: ["minecraft:beef", 3] },
        ] }, brAinMD: { name: TRANSLATE("entity.zombie.name"), introduction: { message: TRANSLATE("jig_ccomp:zombie.introducn.message") }, too_many: { message: TRANSLATE("jig_ccomp:zombie.toomanyem.message") }, favoriteMailResponses: [
            TRANSLATE("jig_ccomp:zombie.favorit.0.message"),
        ], forum: [{
                subject: TRANSLATE('jig_ccomp:zombie.forum_m.0.subject'),
                message: TRANSLATE('jig_ccomp:zombie.forum_m.0.message')
            }, {
                subject: TRANSLATE('jig_ccomp:zombie.forum_m.1.subject'),
                message: TRANSLATE('jig_ccomp:zombie.forum_m.1.message')
            }], random: [
            { message: TRANSLATE("jig_ccomp:zombie.randomx.0.message") },
            { message: TRANSLATE("jig_ccomp:zombie.randomx.1.message") },
        ], attachmentResponses: [
            TRANSLATE("jig_ccomp:zombie.attachm.0.message"),
            TRANSLATE("jig_ccomp:zombie.attachm.1.message"),
        ], inventoryFullResponses: [
            TRANSLATE("jig_ccomp:zombie.invfull.0.message"),
            TRANSLATE("jig_ccomp:zombie.invfull.1.message"),
        ], responses: [
            { message: TRANSLATE("jig_ccomp:zombie.message.0.message") },
            { message: TRANSLATE("jig_ccomp:zombie.message.1.message"), attachment: ["minecraft:rotten_flesh", 1] },
            { message: TRANSLATE("jig_ccomp:zombie.message.2.message") },
            { message: TRANSLATE("jig_ccomp:zombie.message.3.message") },
            { message: TRANSLATE("jig_ccomp:zombie.message.4.message") },
            { message: TRANSLATE("jig_ccomp:zombie.message.5.message") },
            { message: TRANSLATE("jig_ccomp:zombie.message.6.message"), attachment: ["minecraft:iron_shovel", 1] },
            { message: TRANSLATE("jig_ccomp:zombie.message.7.message") },
            { message: TRANSLATE("jig_ccomp:zombie.message.8.message") },
        ] }, SuperHot: { name: TRANSLATE("entity.blaze.name"), introduction: { message: TRANSLATE("jig_ccomp:blazex.introducn.message") }, too_many: { message: TRANSLATE("jig_ccomp:blazex.toomanyem.message") }, favoriteMailResponses: [
            TRANSLATE("jig_ccomp:blazex.favorit.0.message"),
        ], forum: [{
                subject: TRANSLATE('jig_ccomp:blazex.forum_m.0.subject'),
                message: TRANSLATE('jig_ccomp:blazex.forum_m.0.message')
            }, {
                subject: TRANSLATE('jig_ccomp:blazex.forum_m.1.subject'),
                message: TRANSLATE('jig_ccomp:blazex.forum_m.1.message')
            }], random: [
            { message: TRANSLATE("jig_ccomp:blazex.randomx.0.message"), attachment: ["minecraft:blaze_rod", 1] },
            { message: TRANSLATE("jig_ccomp:blazex.randomx.1.message") },
        ], attachmentResponses: [
            TRANSLATE("jig_ccomp:blazex.attachm.0.message"),
            TRANSLATE("jig_ccomp:blazex.attachm.1.message"),
        ], inventoryFullResponses: [
            TRANSLATE("jig_ccomp:blazex.invfull.0.message"),
            TRANSLATE("jig_ccomp:blazex.invfull.1.message"),
        ], responses: [
            { message: TRANSLATE("jig_ccomp:blazex.message.0.message") },
            { message: TRANSLATE("jig_ccomp:blazex.message.1.message") },
            { message: TRANSLATE("jig_ccomp:blazex.message.2.message"), attachment: ["minecraft:sugar", 1] },
            { message: TRANSLATE("jig_ccomp:blazex.message.3.message") },
            { message: TRANSLATE("jig_ccomp:blazex.message.4.message") },
            { message: TRANSLATE("jig_ccomp:blazex.message.5.message"), attachment: ["minecraft:bread", 4] },
            { message: TRANSLATE("jig_ccomp:blazex.message.6.message"), attachment: ["minecraft:cooked_cod", 1] },
            { message: TRANSLATE("jig_ccomp:blazex.message.7.message") },
            { message: TRANSLATE("jig_ccomp:blazex.message.8.message"), attachment: ["minecraft:cooked_porkchop", 2] },
        ] }, PigC: { name: TRANSLATE("entity.piglin.name"), introduction: { message: TRANSLATE("jig_ccomp:piglin.introducn.message") }, too_many: { message: TRANSLATE("jig_ccomp:piglin.toomanyem.message") }, favoriteMailResponses: [
            TRANSLATE("jig_ccomp:piglin.favorit.0.message"),
        ], forum: [{
                subject: TRANSLATE('jig_ccomp:piglin.forum_m.0.subject'),
                message: TRANSLATE('jig_ccomp:piglin.forum_m.0.message')
            }, {
                subject: TRANSLATE('jig_ccomp:piglin.forum_m.1.subject'),
                message: TRANSLATE('jig_ccomp:piglin.forum_m.1.message')
            }], random: [
            { message: TRANSLATE("jig_ccomp:piglin.randomx.0.message") },
            { message: TRANSLATE("jig_ccomp:piglin.randomx.1.message") },
        ], attachmentResponses: [
            TRANSLATE("jig_ccomp:piglin.attachm.0.message"),
            TRANSLATE("jig_ccomp:piglin.attachm.1.message"),
        ], inventoryFullResponses: [
            TRANSLATE("jig_ccomp:piglin.invfull.0.message"),
            TRANSLATE("jig_ccomp:piglin.invfull.1.message"),
        ], responses: [
            { message: TRANSLATE("jig_ccomp:piglin.message.0.message") },
            { message: TRANSLATE("jig_ccomp:piglin.message.1.message") },
            { message: TRANSLATE("jig_ccomp:piglin.message.2.message"), attachment: ["minecraft:gold_nugget", 1] },
            { message: TRANSLATE("jig_ccomp:piglin.message.3.message") },
            { message: TRANSLATE("jig_ccomp:piglin.message.4.message"), attachment: ["minecraft:golden_sword", 1] },
            { message: TRANSLATE("jig_ccomp:piglin.message.5.message"), attachment: ["minecraft:golden_boots", 1] },
            { message: TRANSLATE("jig_ccomp:piglin.message.6.message"), attachment: ["minecraft:gilded_blackstone", 3] },
            { message: TRANSLATE("jig_ccomp:piglin.message.7.message") },
            { message: TRANSLATE("jig_ccomp:piglin.message.8.message") },
        ] }, xXBuffagerXx: { name: TRANSLATE("entity.pillager.name"), introduction: { message: TRANSLATE("jig_ccomp:illage.introducn.message") }, too_many: { message: TRANSLATE("jig_ccomp:illage.toomanyem.message") }, favoriteMailResponses: [
            TRANSLATE("jig_ccomp:illage.favorit.0.message"),
        ], forum: [{
                subject: TRANSLATE('jig_ccomp:illage.forum_m.0.subject'),
                message: TRANSLATE('jig_ccomp:illage.forum_m.0.message')
            }, {
                subject: TRANSLATE('jig_ccomp:illage.forum_m.1.subject'),
                message: TRANSLATE('jig_ccomp:illage.forum_m.1.message')
            }], random: [
            { message: TRANSLATE("jig_ccomp:illage.randomx.0.message") },
            { message: TRANSLATE("jig_ccomp:illage.randomx.1.message") },
        ], attachmentResponses: [
            TRANSLATE("jig_ccomp:illage.attachm.0.message"),
            TRANSLATE("jig_ccomp:illage.attachm.1.message"),
        ], inventoryFullResponses: [
            TRANSLATE("jig_ccomp:illage.invfull.0.message"),
            TRANSLATE("jig_ccomp:illage.invfull.1.message"),
        ], responses: [
            { message: TRANSLATE("jig_ccomp:illage.message.0.message"), attachment: ["minecraft:saddle", 1] },
            { message: TRANSLATE("jig_ccomp:illage.message.1.message"), attachment: ["minecraft:iron_sword", 1] },
            { message: TRANSLATE("jig_ccomp:illage.message.2.message") },
            { message: TRANSLATE("jig_ccomp:illage.message.3.message") },
            { message: TRANSLATE("jig_ccomp:illage.message.4.message"), attachment: ["minecraft:flower_banner_pattern", 1] },
            { message: TRANSLATE("jig_ccomp:illage.message.5.message") },
            { message: TRANSLATE("jig_ccomp:illage.message.6.message"), attachment: ["minecraft:bell", 1] },
            { message: TRANSLATE("jig_ccomp:illage.message.7.message") },
            { message: TRANSLATE("jig_ccomp:illage.message.8.message"), attachment: ["minecraft:totem_of_undying", 1] },
        ] }, ManeFrame: { name: TRANSLATE("entity.horse.name"), introduction: { message: TRANSLATE("jig_ccomp:horsex.introducn.message") }, too_many: { message: TRANSLATE("jig_ccomp:horsex.toomanyem.message") }, favoriteMailResponses: [
            TRANSLATE("jig_ccomp:horsex.favorit.0.message"),
        ], forum: [{
                subject: TRANSLATE('jig_ccomp:horsex.forum_m.0.subject'),
                message: TRANSLATE('jig_ccomp:horsex.forum_m.0.message')
            }, {
                subject: TRANSLATE('jig_ccomp:horsex.forum_m.1.subject'),
                message: TRANSLATE('jig_ccomp:horsex.forum_m.1.message')
            }], random: [
            { message: TRANSLATE("jig_ccomp:horsex.randomx.0.message") },
            { message: TRANSLATE("jig_ccomp:horsex.randomx.1.message") },
        ], attachmentResponses: [
            TRANSLATE("jig_ccomp:horsex.attachm.0.message"),
            TRANSLATE("jig_ccomp:horsex.attachm.1.message"),
        ], inventoryFullResponses: [
            TRANSLATE("jig_ccomp:horsex.invfull.0.message"),
            TRANSLATE("jig_ccomp:horsex.invfull.1.message"),
        ], responses: [
            { message: TRANSLATE("jig_ccomp:horsex.message.0.message") },
            { message: TRANSLATE("jig_ccomp:horsex.message.1.message"), attachment: ["minecraft:sugar", 1] },
            { message: TRANSLATE("jig_ccomp:horsex.message.2.message"), attachment: ["minecraft:apple", 5] },
            { message: TRANSLATE("jig_ccomp:horsex.message.3.message") },
            { message: TRANSLATE("jig_ccomp:horsex.message.4.message") },
            { message: TRANSLATE("jig_ccomp:horsex.message.5.message"), attachment: ["minecraft:hay_block", 10] },
            { message: TRANSLATE("jig_ccomp:horsex.message.6.message"), attachment: ["minecraft:saddle", 1] },
            { message: TRANSLATE("jig_ccomp:horsex.message.7.message"), attachment: ["minecraft:wheat", 15] },
            { message: TRANSLATE("jig_ccomp:horsex.message.8.message"), attachment: ["minecraft:leather_horse_armor", 1] },
        ] }, RuggedPC: { name: TRANSLATE("entity.llama.name"), introduction: { message: TRANSLATE("jig_ccomp:llamax.introducn.message") }, too_many: { message: TRANSLATE("jig_ccomp:llamax.toomanyem.message") }, favoriteMailResponses: [
            TRANSLATE("jig_ccomp:llamax.favorit.0.message"),
        ], forum: [{
                subject: TRANSLATE('jig_ccomp:llamax.forum_m.0.subject'),
                message: TRANSLATE('jig_ccomp:llamax.forum_m.0.message')
            }, {
                subject: TRANSLATE('jig_ccomp:llamax.forum_m.1.subject'),
                message: TRANSLATE('jig_ccomp:llamax.forum_m.1.message')
            }], random: [
            { message: TRANSLATE("jig_ccomp:llamax.randomx.0.message") },
            { message: TRANSLATE("jig_ccomp:llamax.randomx.1.message") },
        ], attachmentResponses: [
            TRANSLATE("jig_ccomp:llamax.attachm.0.message"),
            TRANSLATE("jig_ccomp:llamax.attachm.1.message"),
        ], inventoryFullResponses: [
            TRANSLATE("jig_ccomp:llamax.invfull.0.message"),
            TRANSLATE("jig_ccomp:llamax.invfull.1.message"),
        ], responses: [
            { message: TRANSLATE("jig_ccomp:llamax.message.0.message"), attachment: ["minecraft:red_carpet", 1] },
            { message: TRANSLATE("jig_ccomp:llamax.message.1.message"), attachment: ["minecraft:chest", 1] },
            { message: TRANSLATE("jig_ccomp:llamax.message.2.message"), attachment: ["minecraft:emerald", 2] },
            { message: TRANSLATE("jig_ccomp:llamax.message.3.message") },
            { message: TRANSLATE("jig_ccomp:llamax.message.4.message"), attachment: ["minecraft:hay_block", 1] },
            { message: TRANSLATE("jig_ccomp:llamax.message.5.message") },
            { message: TRANSLATE("jig_ccomp:llamax.message.6.message"), attachment: ["minecraft:lead", 1] },
            { message: TRANSLATE("jig_ccomp:llamax.message.7.message") },
            { message: TRANSLATE("jig_ccomp:llamax.message.8.message"), attachment: ["minecraft:acacia_fence", 1] },
        ] }, WaterCooled: { name: TRANSLATE("entity.camel.name"), introduction: { message: TRANSLATE("jig_ccomp:camelx.introducn.message") }, too_many: { message: TRANSLATE("jig_ccomp:camelx.toomanyem.message") }, favoriteMailResponses: [
            TRANSLATE("jig_ccomp:camelx.favorit.0.message"),
        ], forum: [{
                subject: TRANSLATE('jig_ccomp:camelx.forum_m.0.subject'),
                message: TRANSLATE('jig_ccomp:camelx.forum_m.0.message')
            }, {
                subject: TRANSLATE('jig_ccomp:camelx.forum_m.1.subject'),
                message: TRANSLATE('jig_ccomp:camelx.forum_m.1.message')
            }], random: [
            { message: TRANSLATE("jig_ccomp:camelx.randomx.0.message") },
            { message: TRANSLATE("jig_ccomp:camelx.randomx.1.message") },
        ], attachmentResponses: [
            TRANSLATE("jig_ccomp:camelx.attachm.0.message"),
            TRANSLATE("jig_ccomp:camelx.attachm.1.message"),
        ], inventoryFullResponses: [
            TRANSLATE("jig_ccomp:camelx.invfull.0.message"),
            TRANSLATE("jig_ccomp:camelx.invfull.1.message"),
        ], responses: [
            { message: TRANSLATE("jig_ccomp:camelx.message.0.message") },
            { message: TRANSLATE("jig_ccomp:camelx.message.1.message"), attachment: ["minecraft:cactus", 2] },
            { message: TRANSLATE("jig_ccomp:camelx.message.2.message") },
            { message: TRANSLATE("jig_ccomp:camelx.message.3.message"), attachment: ["minecraft:saddle", 1] },
            { message: TRANSLATE("jig_ccomp:camelx.message.4.message"), attachment: ["minecraft:lead", 1] },
            { message: TRANSLATE("jig_ccomp:camelx.message.5.message") },
            { message: TRANSLATE("jig_ccomp:camelx.message.6.message") },
            { message: TRANSLATE("jig_ccomp:camelx.message.7.message"), attachment: ["minecraft:bamboo_fence", 1] },
            { message: TRANSLATE("jig_ccomp:camelx.message.8.message") },
        ] }, Croak102: { name: TRANSLATE("entity.frog.name"), introduction: { message: TRANSLATE("jig_ccomp:frogxx.introducn.message") }, too_many: { message: TRANSLATE("jig_ccomp:frogxx.toomanyem.message") }, favoriteMailResponses: [
            TRANSLATE("jig_ccomp:frogxx.favorit.0.message"),
        ], forum: [{
                subject: TRANSLATE('jig_ccomp:frogxx.forum_m.0.subject'),
                message: TRANSLATE('jig_ccomp:frogxx.forum_m.0.message')
            }, {
                subject: TRANSLATE('jig_ccomp:frogxx.forum_m.1.subject'),
                message: TRANSLATE('jig_ccomp:frogxx.forum_m.1.message')
            }], random: [
            { message: TRANSLATE("jig_ccomp:frogxx.randomx.0.message") },
            { message: TRANSLATE("jig_ccomp:frogxx.randomx.1.message") },
        ], attachmentResponses: [
            TRANSLATE("jig_ccomp:frogxx.attachm.0.message"),
            TRANSLATE("jig_ccomp:frogxx.attachm.1.message"),
        ], inventoryFullResponses: [
            TRANSLATE("jig_ccomp:frogxx.invfull.0.message"),
            TRANSLATE("jig_ccomp:frogxx.invfull.1.message"),
        ], responses: [
            { message: TRANSLATE("jig_ccomp:frogxx.message.0.message") },
            { message: TRANSLATE("jig_ccomp:frogxx.message.1.message") },
            { message: TRANSLATE("jig_ccomp:frogxx.message.2.message"), attachment: ["minecraft:slime_ball", 1] },
            { message: TRANSLATE("jig_ccomp:frogxx.message.3.message"), attachment: ["minecraft:waterlily", 1] },
            { message: TRANSLATE("jig_ccomp:frogxx.message.4.message"), attachment: ["minecraft:big_dropleaf", 1] },
            { message: TRANSLATE("jig_ccomp:frogxx.message.5.message") },
            { message: TRANSLATE("jig_ccomp:frogxx.message.6.message"), attachment: ["minecraft:verdant_froglight", 1] },
            { message: TRANSLATE("jig_ccomp:frogxx.message.7.message") },
            { message: TRANSLATE("jig_ccomp:frogxx.message.8.message"), attachment: ["minecraft:bucket", 1] },
        ] }, Newton: { name: TRANSLATE("entity.axolotl.name"), introduction: { message: TRANSLATE("jig_ccomp:axolot.introducn.message") }, too_many: { message: TRANSLATE("jig_ccomp:axolot.toomanyem.message") }, favoriteMailResponses: [
            TRANSLATE("jig_ccomp:axolot.favorit.0.message"),
        ], forum: [{
                subject: TRANSLATE('jig_ccomp:axolot.forum_m.0.subject'),
                message: TRANSLATE('jig_ccomp:axolot.forum_m.0.message')
            }, {
                subject: TRANSLATE('jig_ccomp:axolot.forum_m.1.subject'),
                message: TRANSLATE('jig_ccomp:axolot.forum_m.1.message')
            }], random: [
            { message: TRANSLATE("jig_ccomp:axolot.randomx.0.message") },
            { message: TRANSLATE("jig_ccomp:axolot.randomx.1.message") },
        ], attachmentResponses: [
            TRANSLATE("jig_ccomp:axolot.attachm.0.message"),
            TRANSLATE("jig_ccomp:axolot.attachm.1.message"),
        ], inventoryFullResponses: [
            TRANSLATE("jig_ccomp:axolot.invfull.0.message"),
            TRANSLATE("jig_ccomp:axolot.invfull.1.message"),
        ], responses: [
            { message: TRANSLATE("jig_ccomp:axolot.message.0.message"), attachment: ["minecraft:glow_berries", 1] },
            { message: TRANSLATE("jig_ccomp:axolot.message.1.message"), attachment: ["minecraft:water_bucket", 1] },
            { message: TRANSLATE("jig_ccomp:axolot.message.2.message") },
            { message: TRANSLATE("jig_ccomp:axolot.message.3.message"), attachment: ["minecraft:tropical_fish_bucket", 1] },
            { message: TRANSLATE("jig_ccomp:axolot.message.4.message") },
            { message: TRANSLATE("jig_ccomp:axolot.message.5.message"), attachment: ["minecraft:trident", 1] },
            { message: TRANSLATE("jig_ccomp:axolot.message.6.message") },
            { message: TRANSLATE("jig_ccomp:axolot.message.7.message") },
            { message: TRANSLATE("jig_ccomp:axolot.message.8.message") },
        ] }, DeditatedRAM: { name: TRANSLATE("entity.sheep.name"), introduction: { message: TRANSLATE("jig_ccomp:sheepx.introducn.message") }, too_many: { message: TRANSLATE("jig_ccomp:sheepx.toomanyem.message") }, favoriteMailResponses: [
            TRANSLATE("jig_ccomp:sheepx.favorit.0.message"),
        ], forum: [{
                subject: TRANSLATE('jig_ccomp:sheepx.forum_m.0.subject'),
                message: TRANSLATE('jig_ccomp:sheepx.forum_m.0.message')
            }, {
                subject: TRANSLATE('jig_ccomp:sheepx.forum_m.1.subject'),
                message: TRANSLATE('jig_ccomp:sheepx.forum_m.1.message')
            }], random: [
            { message: TRANSLATE("jig_ccomp:sheepx.randomx.0.message") },
            { message: TRANSLATE("jig_ccomp:sheepx.randomx.1.message") },
        ], attachmentResponses: [
            TRANSLATE("jig_ccomp:sheepx.attachm.0.message"),
            TRANSLATE("jig_ccomp:sheepx.attachm.1.message"),
        ], inventoryFullResponses: [
            TRANSLATE("jig_ccomp:sheepx.invfull.0.message"),
            TRANSLATE("jig_ccomp:sheepx.invfull.1.message"),
        ], responses: [
            { message: TRANSLATE("jig_ccomp:sheepx.message.0.message"), attachment: ["minecraft:white_wool", 1] },
            { message: TRANSLATE("jig_ccomp:sheepx.message.1.message"), attachment: ["minecraft:cyan_dye", 1] },
            { message: TRANSLATE("jig_ccomp:sheepx.message.2.message") },
            { message: TRANSLATE("jig_ccomp:sheepx.message.3.message"), attachment: ["minecraft:wheat", 1] },
            { message: TRANSLATE("jig_ccomp:sheepx.message.4.message"), attachment: ["minecraft:shears", 1] },
            { message: TRANSLATE("jig_ccomp:sheepx.message.5.message"), attachment: ["minecraft:name_tag", 1] },
            { message: TRANSLATE("jig_ccomp:sheepx.message.6.message") },
            { message: TRANSLATE("jig_ccomp:sheepx.message.7.message") },
            { message: TRANSLATE("jig_ccomp:sheepx.message.8.message") },
        ] }, FeatherLite: { name: TRANSLATE("entity.villager.fletcher"), introduction: { message: TRANSLATE("jig_ccomp:fletch.introducn.message") }, too_many: { message: TRANSLATE("jig_ccomp:fletch.toomanyem.message") }, favoriteMailResponses: [
            TRANSLATE("jig_ccomp:fletch.favorit.0.message"),
        ], forum: [{
                subject: TRANSLATE('jig_ccomp:fletch.forum_m.0.subject'),
                message: TRANSLATE('jig_ccomp:fletch.forum_m.0.message')
            }, {
                subject: TRANSLATE('jig_ccomp:fletch.forum_m.1.subject'),
                message: TRANSLATE('jig_ccomp:fletch.forum_m.1.message')
            }], random: [
            { message: TRANSLATE("jig_ccomp:fletch.randomx.0.message") },
            { message: TRANSLATE("jig_ccomp:fletch.randomx.1.message") },
        ], attachmentResponses: [
            TRANSLATE("jig_ccomp:fletch.attachm.0.message"),
            TRANSLATE("jig_ccomp:fletch.attachm.1.message"),
        ], inventoryFullResponses: [
            TRANSLATE("jig_ccomp:fletch.invfull.0.message"),
            TRANSLATE("jig_ccomp:fletch.invfull.1.message"),
        ], responses: [
            { message: TRANSLATE("jig_ccomp:fletch.message.0.message") },
            { message: TRANSLATE("jig_ccomp:fletch.message.1.message"), attachment: ["minecraft:feather", 10] },
            { message: TRANSLATE("jig_ccomp:fletch.message.2.message"), attachment: ["minecraft:arrow", 20] },
            { message: TRANSLATE("jig_ccomp:fletch.message.3.message") },
            { message: TRANSLATE("jig_ccomp:fletch.message.4.message"), attachment: ["minecraft:bow", 1] },
            { message: TRANSLATE("jig_ccomp:fletch.message.5.message") },
            { message: TRANSLATE("jig_ccomp:fletch.message.6.message") },
            { message: TRANSLATE("jig_ccomp:fletch.message.7.message"), attachment: ["minecraft:string", 1] },
            { message: TRANSLATE("jig_ccomp:fletch.message.8.message") },
        ] }, CoMooter: { name: TRANSLATE("entity.mooshroom.name"), introduction: { message: TRANSLATE("jig_ccomp:mooshr.introducn.message") }, too_many: { message: TRANSLATE("jig_ccomp:mooshr.toomanyem.message") }, favoriteMailResponses: [
            TRANSLATE("jig_ccomp:mooshr.favorit.0.message"),
        ], forum: [{
                subject: TRANSLATE('jig_ccomp:mooshr.forum_m.0.subject'),
                message: TRANSLATE('jig_ccomp:mooshr.forum_m.0.message')
            }, {
                subject: TRANSLATE('jig_ccomp:mooshr.forum_m.1.subject'),
                message: TRANSLATE('jig_ccomp:mooshr.forum_m.1.message')
            }], random: [
            { message: TRANSLATE("jig_ccomp:mooshr.randomx.0.message") },
            { message: TRANSLATE("jig_ccomp:mooshr.randomx.1.message") },
        ], attachmentResponses: [
            TRANSLATE("jig_ccomp:mooshr.attachm.0.message"),
            TRANSLATE("jig_ccomp:mooshr.attachm.1.message"),
        ], inventoryFullResponses: [
            TRANSLATE("jig_ccomp:mooshr.invfull.0.message"),
            TRANSLATE("jig_ccomp:mooshr.invfull.1.message"),
        ], responses: [
            { message: TRANSLATE("jig_ccomp:mooshr.message.0.message") },
            { message: TRANSLATE("jig_ccomp:mooshr.message.1.message") },
            { message: TRANSLATE("jig_ccomp:mooshr.message.2.message"), attachment: ["minecraft:milk_bucket", 1] },
            { message: TRANSLATE("jig_ccomp:mooshr.message.3.message") },
            { message: TRANSLATE("jig_ccomp:mooshr.message.4.message"), attachment: ["minecraft:brown_mushroom", 2] },
            { message: TRANSLATE("jig_ccomp:mooshr.message.5.message"), attachment: ["minecraft:red_mushroom", 2] },
            { message: TRANSLATE("jig_ccomp:mooshr.message.6.message") },
            { message: TRANSLATE("jig_ccomp:mooshr.message.7.message") },
            { message: TRANSLATE("jig_ccomp:mooshr.message.8.message"), attachment: ["minecraft:suspicious_stew", 1] },
        ] }, Egg86: { name: TRANSLATE("entity.chicken.name"), introduction: { message: TRANSLATE("jig_ccomp:chicke.introducn.message"), attachment: ["minecraft:egg", 1] }, too_many: { message: TRANSLATE("jig_ccomp:chicke.toomanyem.message") }, favoriteMailResponses: [
            TRANSLATE("jig_ccomp:chicke.favorit.0.message"),
        ], forum: [{
                subject: TRANSLATE('jig_ccomp:chicke.forum_m.0.subject'),
                message: TRANSLATE('jig_ccomp:chicke.forum_m.0.message')
            }, {
                subject: TRANSLATE('jig_ccomp:chicke.forum_m.1.subject'),
                message: TRANSLATE('jig_ccomp:chicke.forum_m.1.message')
            }], random: [
            { message: TRANSLATE("jig_ccomp:chicke.randomx.0.message") },
            { message: TRANSLATE("jig_ccomp:chicke.randomx.1.message") },
        ], attachmentResponses: [
            TRANSLATE("jig_ccomp:chicke.attachm.0.message"),
            TRANSLATE("jig_ccomp:chicke.attachm.1.message"),
        ], inventoryFullResponses: [
            TRANSLATE("jig_ccomp:chicke.invfull.0.message"),
            TRANSLATE("jig_ccomp:chicke.invfull.1.message"),
        ], responses: [
            { message: TRANSLATE("jig_ccomp:chicke.message.0.message"), attachment: ["minecraft:egg", 1] },
            { message: TRANSLATE("jig_ccomp:chicke.message.1.message"), attachment: ["minecraft:egg", 1] },
            { message: TRANSLATE("jig_ccomp:chicke.message.2.message"), attachment: ["minecraft:egg", 1] },
            { message: TRANSLATE("jig_ccomp:chicke.message.3.message"), attachment: ["minecraft:egg", 1] },
            { message: TRANSLATE("jig_ccomp:chicke.message.4.message"), attachment: ["minecraft:egg", 1] },
            { message: TRANSLATE("jig_ccomp:chicke.message.5.message"), attachment: ["minecraft:egg", 1] },
            { message: TRANSLATE("jig_ccomp:chicke.message.6.message"), attachment: ["minecraft:egg", 1] },
            { message: TRANSLATE("jig_ccomp:chicke.message.7.message"), attachment: ["minecraft:egg", 1] },
            { message: TRANSLATE("jig_ccomp:chicke.message.8.message"), attachment: ["minecraft:egg", 1] },
        ] }, BoomBox: { name: TRANSLATE("entity.creeper.name"), introduction: { message: TRANSLATE("jig_ccomp:creepe.introducn.message") }, too_many: { message: TRANSLATE("jig_ccomp:creepe.toomanyem.message") }, favoriteMailResponses: [
            TRANSLATE("jig_ccomp:creepe.favorit.0.message"),
        ], forum: [{
                subject: TRANSLATE('jig_ccomp:creepe.forum_m.0.subject'),
                message: TRANSLATE('jig_ccomp:creepe.forum_m.0.message')
            }, {
                subject: TRANSLATE('jig_ccomp:creepe.forum_m.1.subject'),
                message: TRANSLATE('jig_ccomp:creepe.forum_m.1.message')
            }], random: [
            { message: TRANSLATE("jig_ccomp:creepe.randomx.0.message"), attachment: ["minecraft:gunpowder", 1] },
            { message: TRANSLATE("jig_ccomp:creepe.randomx.1.message") },
        ], attachmentResponses: [
            TRANSLATE("jig_ccomp:creepe.attachm.0.message"),
            TRANSLATE("jig_ccomp:creepe.attachm.1.message"),
        ], inventoryFullResponses: [
            TRANSLATE("jig_ccomp:creepe.invfull.0.message"),
            TRANSLATE("jig_ccomp:creepe.invfull.1.message"),
        ], responses: [
            { message: TRANSLATE("jig_ccomp:creepe.message.0.message") },
            { message: TRANSLATE("jig_ccomp:creepe.message.1.message") },
            { message: TRANSLATE("jig_ccomp:creepe.message.2.message") },
            { message: TRANSLATE("jig_ccomp:creepe.message.3.message") },
            { message: TRANSLATE("jig_ccomp:creepe.message.4.message") },
            { message: TRANSLATE("jig_ccomp:creepe.message.5.message") },
            { message: TRANSLATE("jig_ccomp:creepe.message.6.message") },
            { message: TRANSLATE("jig_ccomp:creepe.message.7.message") },
            { message: TRANSLATE("jig_ccomp:creepe.message.8.message"), attachment: ["minecraft:gunpowder", 5] },
        ] }, PortBox: { name: TRANSLATE("entity.enderman.name"), introduction: { message: TRANSLATE("jig_ccomp:endman.introducn.message") }, too_many: { message: TRANSLATE("jig_ccomp:endman.toomanyem.message") }, favoriteMailResponses: [
            TRANSLATE("jig_ccomp:endman.favorit.0.message"),
        ], forum: [{
                subject: TRANSLATE('jig_ccomp:endman.forum_m.0.subject'),
                message: TRANSLATE('jig_ccomp:endman.forum_m.0.message')
            }, {
                subject: TRANSLATE('jig_ccomp:endman.forum_m.1.subject'),
                message: TRANSLATE('jig_ccomp:endman.forum_m.1.message')
            }], random: [
            { message: TRANSLATE("jig_ccomp:endman.randomx.0.message"), attachment: ["minecraft:ender_pearl", 1] },
            { message: TRANSLATE("jig_ccomp:endman.randomx.1.message") },
        ], attachmentResponses: [
            TRANSLATE("jig_ccomp:endman.attachm.0.message"),
            TRANSLATE("jig_ccomp:endman.attachm.1.message"),
        ], inventoryFullResponses: [
            TRANSLATE("jig_ccomp:endman.invfull.0.message"),
            TRANSLATE("jig_ccomp:endman.invfull.1.message"),
        ], responses: [
            { message: TRANSLATE("jig_ccomp:endman.message.0.message") },
            { message: TRANSLATE("jig_ccomp:endman.message.1.message") },
            { message: TRANSLATE("jig_ccomp:endman.message.2.message") },
            { message: TRANSLATE("jig_ccomp:endman.message.3.message") },
            { message: TRANSLATE("jig_ccomp:endman.message.4.message"), attachment: ["minecraft:carved_pumpkin", 1] },
            { message: TRANSLATE("jig_ccomp:endman.message.5.message") },
            { message: TRANSLATE("jig_ccomp:endman.message.6.message") },
            { message: TRANSLATE("jig_ccomp:endman.message.7.message") },
            { message: TRANSLATE("jig_ccomp:endman.message.8.message"), attachment: ["minecraft:ender_pearl", 2] },
        ] }, MiteyMac: { name: TRANSLATE("entity.endermite.name"), introduction: { message: TRANSLATE("jig_ccomp:endmit.introducn.message") }, too_many: { message: TRANSLATE("jig_ccomp:endmit.toomanyem.message") }, favoriteMailResponses: [
            TRANSLATE("jig_ccomp:endmit.favorit.0.message"),
        ], forum: [{
                subject: TRANSLATE('jig_ccomp:endmit.forum_m.0.subject'),
                message: TRANSLATE('jig_ccomp:endmit.forum_m.0.message')
            }, {
                subject: TRANSLATE('jig_ccomp:endmit.forum_m.1.subject'),
                message: TRANSLATE('jig_ccomp:endmit.forum_m.1.message')
            }], random: [
            { message: TRANSLATE("jig_ccomp:endmit.randomx.0.message") },
            { message: TRANSLATE("jig_ccomp:endmit.randomx.1.message") },
        ], attachmentResponses: [
            TRANSLATE("jig_ccomp:endmit.attachm.0.message"),
            TRANSLATE("jig_ccomp:endmit.attachm.1.message"),
        ], inventoryFullResponses: [
            TRANSLATE("jig_ccomp:endmit.invfull.0.message"),
            TRANSLATE("jig_ccomp:endmit.invfull.1.message"),
        ], responses: [
            { message: TRANSLATE("jig_ccomp:endmit.message.0.message") },
            { message: TRANSLATE("jig_ccomp:endmit.message.1.message") },
            { message: TRANSLATE("jig_ccomp:endmit.message.2.message") },
            { message: TRANSLATE("jig_ccomp:endmit.message.3.message") },
            { message: TRANSLATE("jig_ccomp:endmit.message.4.message") },
            { message: TRANSLATE("jig_ccomp:endmit.message.5.message") },
            { message: TRANSLATE("jig_ccomp:endmit.message.6.message") },
            { message: TRANSLATE("jig_ccomp:endmit.message.7.message") },
            { message: TRANSLATE("jig_ccomp:endmit.message.8.message"), attachment: ["minecraft:ender_eye", 5] },
        ] }, AMBee: { name: TRANSLATE("entity.bee.name"), introduction: { message: TRANSLATE("jig_ccomp:beexxx.introducn.message") }, too_many: { message: TRANSLATE("jig_ccomp:beexxx.toomanyem.message") }, favoriteMailResponses: [
            TRANSLATE("jig_ccomp:beexxx.favorit.0.message"),
        ], forum: [{
                subject: TRANSLATE('jig_ccomp:beexxx.forum_m.0.subject'),
                message: TRANSLATE('jig_ccomp:beexxx.forum_m.0.message')
            }, {
                subject: TRANSLATE('jig_ccomp:beexxx.forum_m.1.subject'),
                message: TRANSLATE('jig_ccomp:beexxx.forum_m.1.message')
            }], random: [
            { message: TRANSLATE("jig_ccomp:beexxx.randomx.0.message") },
            { message: TRANSLATE("jig_ccomp:beexxx.randomx.1.message") },
        ], attachmentResponses: [
            TRANSLATE("jig_ccomp:beexxx.attachm.0.message"),
            TRANSLATE("jig_ccomp:beexxx.attachm.1.message"),
        ], inventoryFullResponses: [
            TRANSLATE("jig_ccomp:beexxx.invfull.0.message"),
            TRANSLATE("jig_ccomp:beexxx.invfull.1.message"),
        ], responses: [
            { message: TRANSLATE("jig_ccomp:beexxx.message.0.message") },
            { message: TRANSLATE("jig_ccomp:beexxx.message.1.message") },
            { message: TRANSLATE("jig_ccomp:beexxx.message.2.message") },
            { message: TRANSLATE("jig_ccomp:beexxx.message.3.message") },
            { message: TRANSLATE("jig_ccomp:beexxx.message.4.message") },
            { message: TRANSLATE("jig_ccomp:beexxx.message.5.message"), attachment: ["minecraft:honey_bottle", 2] },
            { message: TRANSLATE("jig_ccomp:beexxx.message.6.message") },
            { message: TRANSLATE("jig_ccomp:beexxx.message.7.message") },
            { message: TRANSLATE("jig_ccomp:beexxx.message.8.message"), attachment: ["minecraft:honeycomb", 1] },
        ] }, PurpBox: { name: TRANSLATE("entity.shulker.name"), introduction: { message: TRANSLATE("jig_ccomp:shulke.introducn.message") }, too_many: { message: TRANSLATE("jig_ccomp:shulke.toomanyem.message") }, favoriteMailResponses: [
            TRANSLATE("jig_ccomp:shulke.favorit.0.message"),
        ], forum: [{
                subject: TRANSLATE('jig_ccomp:shulke.forum_m.0.subject'),
                message: TRANSLATE('jig_ccomp:shulke.forum_m.0.message')
            }, {
                subject: TRANSLATE('jig_ccomp:shulke.forum_m.1.subject'),
                message: TRANSLATE('jig_ccomp:shulke.forum_m.1.message')
            }], random: [
            { message: TRANSLATE("jig_ccomp:shulke.randomx.0.message"), attachment: ["minecraft:shulker_shell", 1] },
            { message: TRANSLATE("jig_ccomp:shulke.randomx.1.message") },
        ], attachmentResponses: [
            TRANSLATE("jig_ccomp:shulke.attachm.0.message"),
            TRANSLATE("jig_ccomp:shulke.attachm.1.message"),
        ], inventoryFullResponses: [
            TRANSLATE("jig_ccomp:shulke.invfull.0.message"),
            TRANSLATE("jig_ccomp:shulke.invfull.1.message"),
        ], responses: [
            { message: TRANSLATE("jig_ccomp:shulke.message.0.message"), attachment: ["minecraft:shulker_shell", 1] },
            { message: TRANSLATE("jig_ccomp:shulke.message.1.message") },
            { message: TRANSLATE("jig_ccomp:shulke.message.2.message") },
            { message: TRANSLATE("jig_ccomp:shulke.message.3.message") },
            { message: TRANSLATE("jig_ccomp:shulke.message.4.message") },
            { message: TRANSLATE("jig_ccomp:shulke.message.5.message") },
            { message: TRANSLATE("jig_ccomp:shulke.message.6.message"), attachment: ["minecraft:shulker_shell", 1] },
            { message: TRANSLATE("jig_ccomp:shulke.message.7.message") },
            { message: TRANSLATE("jig_ccomp:shulke.message.8.message") },
            { message: TRANSLATE("jig_ccomp:shulke.message.9.message") },
        ] }, [Config.JigarBoxNpcId]: { name: TRANSLATE("jig_ccomp:jignet_desk.name"), introduction: { message: TRANSLATE("jig_ccomp:helpds.introducn.message") }, too_many: { message: TRANSLATE("jig_ccomp:helpds.toomanyem.message") }, attachmentResponses: [
            TRANSLATE("jig_ccomp:helpds.attachm.0.message"),
            TRANSLATE("jig_ccomp:helpds.attachm.1.message"),
        ], inventoryFullResponses: [
            TRANSLATE("jig_ccomp:helpds.invfull.0.message"),
            TRANSLATE("jig_ccomp:helpds.invfull.1.message"),
        ], responses: [
            { message: TRANSLATE("jig_ccomp:helpds.message.0.message"), attachment: ["jig_common:wrench", 1] },
            { message: TRANSLATE("jig_ccomp:helpds.message.1.message") },
            { message: TRANSLATE("jig_ccomp:helpds.message.2.message"), attachment: ["minecraft:apple", 1] },
            { message: TRANSLATE("jig_ccomp:helpds.message.3.message") },
            { message: TRANSLATE("jig_ccomp:helpds.message.4.message") },
            { message: TRANSLATE("jig_ccomp:helpds.message.5.message") },
            { message: TRANSLATE("jig_ccomp:helpds.message.6.message") },
            { message: TRANSLATE("jig_ccomp:helpds.message.7.message") },
            { message: TRANSLATE("jig_ccomp:helpds.message.8.message") },
            { message: TRANSLATE("jig_ccomp:helpds.message.9.message") },
            { message: TRANSLATE("jig_ccomp:helpds.message.10.message") },
        ] }
};

const NPCS = 'jig_ccomp:NPCS';
const JIGARBOX = Config.JigarBoxNpcId;
const MAXFAVORITEEMAILS = 10;
const defaultNpcData = {
    [JIGARBOX]: {
        nextResponse: -1,
        acceptsEmails: true,
        pendingEmail: false
    }
};
const addNpc = (player, npc, acceptsEmails) => {
    const npcs = getNpcs(player);
    if (!(npc in npcs)) {
        npcs[npc] = {
            nextResponse: -1,
            acceptsEmails,
            pendingEmail: false
        };
    }
    player.setDynamicProperty(NPCS, JSON.stringify(npcs));
};
const saveFavoriteEmail = (player, npc, subject, message) => {
    const npcs = getNpcs(player);
    if (!npcs[npc].favoriteEmails) {
        npcs[npc].favoriteEmails = [];
    }
    const favoriteEmail = npcs[npc].favoriteEmails;
    favoriteEmail.push({ subject, message });
    if (favoriteEmail.length > MAXFAVORITEEMAILS) {
        favoriteEmail.shift();
    }
    player.setDynamicProperty(NPCS, JSON.stringify(npcs));
};
const incrementNpcResponse = (player, npc) => {
    const npcs = getNpcs(player);
    npcs[npc].nextResponse++;
    player.setDynamicProperty(NPCS, JSON.stringify(npcs));
};
const markPendingEmail = (player, npc, pendingEmail) => {
    const npcs = getNpcs(player);
    npcs[npc].pendingEmail = pendingEmail;
    player.setDynamicProperty(NPCS, JSON.stringify(npcs));
};
const enableNpcMails = (player, npc) => {
    const npcs = getNpcs(player);
    npcs[npc].acceptsEmails = true;
    npcs[npc].pendingEmail = false;
    player.setDynamicProperty(NPCS, JSON.stringify(npcs));
};
const getEmaileableNpcs = (player) => {
    const npcs = getNpcs(player);
    return Object.fromEntries(Object.entries(npcs).filter(e => e[1].acceptsEmails));
};
const getNpcs = (player) => {
    return JSON.parse(player.getDynamicProperty(NPCS) ?? JSON.stringify(defaultNpcData));
};
const clearNpcs = (player) => {
    player.setDynamicProperty(NPCS, undefined);
};

const blockNpc = (npc) => {
    return multiButtonScriptDialogue(TRANSLATE("jig_ccomp:block-npc.title"))
        .setBody(RAW_TEXT(TRANSLATE("jig_ccomp:block-npc.body", "\n\n"), TEXT_OR_RAW_TEXT(npc), TRANSLATE("jig_ccomp:block-npc.body2", "\n\n")))
        .addButton("yes", TRANSLATE("jig_ccomp:yes"))
        .addButton("no", TRANSLATE("jig_ccomp:no"));
};

const blockPlayer = (player, computer) => {
    return multiButtonScriptDialogue(TRANSLATE("jig_ccomp:block-player.title"))
        .setBody(RAW_TEXT(TRANSLATE("jig_ccomp:block-player.body", "\n\n")))
        .addButton("player", RAW_TEXT(TRANSLATE("jig_ccomp:block-player"), TEXT_OR_RAW_TEXT(player)))
        .addButton("computer", RAW_TEXT(TRANSLATE("jig_ccomp:block-computer"), TEXT_OR_RAW_TEXT(computer)))
        .addButton("both", TRANSLATE("jig_ccomp:block-both"))
        .addButton("no", TRANSLATE("jig_ccomp:cancel"));
};

const blockScript = (script, computer) => {
    return multiButtonScriptDialogue(TRANSLATE("jig_ccomp:block-script.title"))
        .setBody(TRANSLATE("jig_ccomp:block-script.body", "\n\n"))
        .addButton("script", RAW_TEXT(TRANSLATE("jig_ccomp:block-script"), TEXT_OR_RAW_TEXT(script)))
        .addButton("computer", RAW_TEXT(TRANSLATE("jig_ccomp:block-scriptcomputer"), TEXT_OR_RAW_TEXT(computer)))
        .addButton("no", TRANSLATE("jig_ccomp:cancel"));
};

const PLAYER = TRANSLATE("jig_ccomp:player");
const COMPUTER = TRANSLATE("jig_ccomp:computer");
const SCRIPT = TRANSLATE("jig_ccomp:script");
const NPC = TRANSLATE("jig_ccomp:npc");
const unblock = (blocked) => {
    return multiButtonScriptDialogue(TRANSLATE("jig_ccomp:unblock-player.title"))
        .setBody(RAW_TEXT(TRANSLATE("jig_ccomp:unblock-player.body", "\n\n", "\n"), RAW_TEXT(getLabel(blocked), blocked.data.name), TEXT("?")))
        .addButton("yes", TRANSLATE("jig_ccomp:yes"))
        .addButton("no", TRANSLATE("jig_ccomp:no"));
};
const unblockList = (blocked) => {
    return multiButtonScriptDialogue(TRANSLATE("jig_ccomp:unblock-pc.title"))
        .setBody(TRANSLATE("jig_ccomp:unblock-pc.body"))
        .addButtons(blocked.map(b => ({
        name: b.id,
        text: RAW_TEXT(getLabel(b), TEXT(": "), b.data.name)
    })));
};
const getLabel = (block) => {
    if (block.block_type === "player") {
        if (block.sender_type === "npc") {
            return NPC;
        }
        else if (block.sender_type === "script") {
            return SCRIPT;
        }
        return PLAYER;
    }
    return COMPUTER;
};

const deleteEmailsDialogue = (emails) => {
    return multiButtonScriptDialogue(TRANSLATE("jig_ccomp:pc-delete-email.title"))
        .setBody(TRANSLATE("jig_ccomp:pc-delete-email.body", "\n"))
        .addButton('cancel', TRANSLATE("jig_ccomp:pc-delete-email.cancel"))
        .addButtons(emails.map(e => ({
        name: e.id,
        text: RAW_TEXT(TEXT("§4"), TEXT_OR_RAW_TEXT(e.from_player), TEXT(": "), TEXT_OR_RAW_TEXT(e.subject)),
    }), 'textures/jig/ccomp/myui/trash.png'));
};

// Used in both players and world
const POST_KEY = "jig_ccomp:posts";
const initForum = () => {
    subscribeFromComputerNpcDialogue(async (event) => {
        if (event.id === "jig_ccomp:forum") {
            event.computer.computer.triggerEvent('jig_ccomp:typing');
            cleanupPlayerData(event.player);
            if (!event.computer.getCreator()) {
                event.computer.setCreator(event.player.name);
            }
            forum: while (true) {
                const playerData = getPlayerData(event.player);
                const posts = getPlayerPosts(playerData, event.computer.getEmailBlocks());
                const response = await openForumDialogue(event.player, posts, playerData);
                if (response instanceof ButtonDialogueResponse) {
                    event.computer.computer.triggerEvent('jig_ccomp:typing');
                    let postId = undefined;
                    if (response.selected === "post") {
                        postId = await makePost(event.player, event.computer);
                    }
                    else {
                        postId = response.selected;
                    }
                    if (postId) {
                        while (true) {
                            const viewPostResponse = await viewPost(event.player, event.computer, getPostById(postId, event.player, event.computer));
                            if (viewPostResponse === "back") {
                                continue;
                            }
                            else if (viewPostResponse === "forum") {
                                continue forum;
                            }
                            break;
                        }
                    }
                }
                break;
            }
        }
    });
};
const makePost = async (player, computer) => {
    const response = await openMakePostDialogue(player);
    if (response instanceof InputScriptDialogueResponse) {
        computer.computer.triggerEvent('jig_ccomp:typing');
        const postId = makeRandomId();
        addPost({
            id: postId,
            op_id: player.name,
            op_name: TEXT(player.name),
            op_type: 'player',
            op_computer_id: computer.getId(),
            title: response.values.subject,
            message: response.values.message,
            replies: [],
            last_update: world.getAbsoluteTime(),
        });
        return postId;
    }
    return undefined;
};
const makeNpcPost = (npcId, npcName, npcPostForum) => {
    const postId = makeRandomId();
    addPost({
        id: postId,
        op_id: npcId,
        op_name: npcName,
        op_type: 'npc',
        op_computer_id: npcId,
        title: npcPostForum.subject,
        message: npcPostForum.message,
        replies: [],
        last_update: world.getAbsoluteTime(),
    });
    return postId;
};
const viewPost = async (player, computer, post) => {
    if (post === undefined) {
        if (await openPostNotFound(player)) {
            return "forum";
        }
    }
    else {
        if (post.op_type === "npc") {
            const unknownNpcs = getUnknownNpcs(player);
            if (unknownNpcs.includes(post.op_id)) {
                debug(`Adding npc: ${post.op_id}`);
                addNpc(player, post.op_id, true);
            }
        }
        updatePlayerDataReadPost(player, post.id);
        const response = await openReadPostDialogue(player, post);
        if (response instanceof ButtonDialogueResponse) {
            computer.computer.triggerEvent('jig_ccomp:typing');
            if (response.selected === "more") {
                return await optionsPost(player, computer, post);
            }
            else {
                return "forum";
            }
        }
    }
    return undefined;
};
const optionsPost = async (player, computer, post) => {
    const response = await openPostOptions(player, post.op_id);
    if (response instanceof ButtonDialogueResponse) {
        computer.computer.triggerEvent('jig_ccomp:typing');
        switch (response.selected) {
            case "reply":
                {
                    const replyResponse = await openReplyPostDialogue(player);
                    if (replyResponse instanceof InputScriptDialogueResponse) {
                        computer.computer.triggerEvent('jig_ccomp:typing');
                        addReply(post.id, {
                            title: replyResponse.values.subject,
                            message: replyResponse.values.message,
                            op_name: TEXT(player.name),
                            op_id: player.name
                        });
                        return "back";
                    }
                }
                break;
            case "reply-op":
                {
                    // Check if op's computer exists
                    let shouldSendEmail = false;
                    if (post.op_type === "player") {
                        if (craftComputers.isComputerAvailable(post.op_computer_id)) {
                            shouldSendEmail = true;
                        }
                    }
                    else if (post.op_type === "npc") {
                        shouldSendEmail = true;
                    }
                    if (shouldSendEmail) {
                        await sendEmail(player, computer, false, {
                            subject: prependToStringOrRawMessageThread(post.title, 'RE: ', false),
                            to: post.op_computer_id
                        });
                        return "back";
                    }
                    else {
                        const notFoundResponse = await openComputerNotFound(player);
                        if (notFoundResponse instanceof ButtonDialogueResponse) {
                            computer.computer.triggerEvent('jig_ccomp:typing');
                            return notFoundResponse.selected;
                        }
                    }
                }
                break;
            case "hide":
                {
                    updatePlayerDataHidePost(player, post.id);
                    return "forum";
                }
            case "block":
                {
                    const users = [];
                    users.push({
                        id: post.op_id,
                        name: post.op_name,
                        type: post.op_type,
                    });
                    users.push(...post.replies.map(r => ({
                        id: r.op_id,
                        name: r.op_name,
                        type: "player", // Only players can reply
                    })));
                    const blockedUsers = computer.getEmailBlocks().filter(block => block.sender_type === 'player');
                    const usersToDisplay = users
                        .filter((user, i) => { return users.findIndex(u => u.id == user.id) === i; }) // Only the first match will stay
                        .filter(user => user.id !== player.name) // Exclude the player
                        .filter(user => !blockedUsers.some(block => block.data.id === user.id)); // Exclude blocked users
                    const blockedReply = await openBlock(player, usersToDisplay);
                    if (blockedReply instanceof ButtonDialogueResponse) {
                        computer.computer.triggerEvent('jig_ccomp:typing');
                        if (blockedReply.selected === "back") {
                            return "back";
                        }
                        const blocked = users.find(u => u.id === blockedReply.selected);
                        let block = false;
                        if (blocked.type === "npc") {
                            const blockNpcResponse = await blockNpc(blocked.name).open({
                                player,
                                ...defaultScriptDialogueOptions
                            });
                            if (blockNpcResponse instanceof ButtonDialogueResponse) {
                                computer.computer.triggerEvent('jig_ccomp:typing');
                                if (blockNpcResponse.selected === "yes") {
                                    block = true;
                                }
                                else if (blockNpcResponse.selected === "no") {
                                    return "back";
                                }
                            }
                        }
                        else {
                            block = true;
                        }
                        if (block) {
                            computer.addBlock({
                                id: makeRandomId(),
                                sender_type: blocked.type,
                                block_type: 'player',
                                data: {
                                    id: blocked.id,
                                    name: TEXT_OR_RAW_TEXT(blocked.name),
                                }
                            });
                            if (blocked.type === "npc") {
                                computer.computer.runCommand(`/dialogue open @s "${player.name}" jig_ccomp:old_pc_blkn`);
                            }
                            else if (blocked.type === "player") {
                                computer.computer.runCommand(`/dialogue open @s "${player.name}" jig_ccomp:old_pc_blkp`);
                            }
                            else if (blocked.type === "script") {
                                computer.computer.runCommand(`/dialogue open @s "${player.name}" jig_ccomp:old_pc_blks`);
                            }
                        }
                        return undefined;
                    }
                }
                break;
            case "back":
                return "back";
        }
    }
};
const openForumDialogue = async (player, posts, playerData) => {
    return await multiButtonScriptDialogue(TRANSLATE('jig_ccomp:forum.title'))
        .setBody(TRANSLATE('jig_ccomp:forum.body'))
        .addButton('post', TRANSLATE('jig_ccomp:forum.make_title'), 'textures/jig/ccomp/myui/reply-thread.png')
        .addButtons(posts.map(p => ({
        name: p.id,
        text: p.title,
        iconPath: postIsRead(p, playerData) ? "textures/jig/ccomp/myui/post-old.png" : "textures/jig/ccomp/myui/post-new.png"
    }))).open({
        player,
        ...defaultScriptDialogueOptions
    });
};
const openMakePostDialogue = async (player) => {
    return await inputScriptDialogue(TRANSLATE('jig_ccomp:forum.make-post.title'))
        .addElement(inputText("subject", TRANSLATE("jig_ccomp:forum.make-post.subject.label"), TRANSLATE("jig_ccomp:forum.make-post.subject.placeholder"), ''))
        .addElement(inputText("message", TRANSLATE("jig_ccomp:forum.make-post.message.label"), TRANSLATE("jig_ccomp:forum.make-post.message.placeholder")))
        .open({
        player,
        ...defaultScriptDialogueOptions
    });
};
const openReplyPostDialogue = async (player) => {
    return await inputScriptDialogue(TRANSLATE('jig_ccomp:forum.reply-post.title'))
        .addElement(inputText("message", TRANSLATE("jig_ccomp:forum.reply-post.message.label"), TRANSLATE("jig_ccomp:forum.reply-post.message.placeholder")))
        .addElement(inputDropdown("subject", TRANSLATE("jig_ccomp:forum.reply-post.subject.label"))
        .addOption(':)', ':)')
        .addOption(':(', ':(')
        .addOption(':/', ':/')
        .addOption(":'(", ":'(")
        .addOption(":D", ":D")
        .addOption("XD", "XD")
        .addOption("^_^", "^_^")
        .addOption("-.-", "-.-")
        .addOption("o.O", "o.O")
        .addOption("OwO", "OwO"))
        .open({
        player,
        ...defaultScriptDialogueOptions
    });
};
const openReadPostDialogue = async (player, post) => {
    return await dualButtonScriptDialogue(TRANSLATE('jig_ccomp:read-post.title'), {
        name: 'more',
        text: TRANSLATE('jig_ccomp:read-post.more')
    }, {
        name: 'back',
        text: TRANSLATE('jig_ccomp:read-post.back')
    })
        .setBody(RAW_TEXT(
    //TRANSLATE('jig_ccomp:read-post.body-title'),
    //TEXT("\n"),
    TRANSLATE('jig_ccomp:read-post.body-subject'), TEXT_OR_RAW_TEXT(post.title), TEXT("§r\n"), 
    //TRANSLATE('jig_ccomp:read-post.body-message'),
    TEXT_OR_RAW_TEXT(post.message), TEXT("\n"), TRANSLATE('jig_ccomp:read-post.body-author'), post.op_name, ...post.replies.map(r => {
        if (r.playerBlocked) {
            return RAW_TEXT(TEXT("\n§8  -----   §r\n"), TEXT("§r"), TRANSLATE("jig_ccomp:read-post.blocked-reply"));
        }
        return RAW_TEXT(TEXT("\n§8  -----   §r\n"), TEXT("§r"), 
        //TRANSLATE('jig_ccomp:read-post.reply-body'),
        TEXT(r.message), TEXT("\n§r"), TEXT("§7"), r.op_name, TEXT("    §e§l"), TEXT(r.title));
    })))
        .open({
        player,
        ...defaultScriptDialogueOptions
    });
};
// This is an edge case - if the player is viewing forums titles, and clicks on and old
// post that has been deleted it won't load.
// I think the same could happen in computer emails, but less probable as the content is
// not as widely shared.
const openPostNotFound = async (player) => {
    return await multiButtonScriptDialogue(TRANSLATE('jig_ccomp:post-not-found.title'))
        .addButton("back", TRANSLATE("jig_ccomp:post-not-found.back"))
        .setBody(TRANSLATE("jig_ccomp:post-not-found.body"))
        .open({
        player,
        ...defaultScriptDialogueOptions
    });
};
const openComputerNotFound = async (player) => {
    return await multiButtonScriptDialogue(TRANSLATE('jig_ccomp:computer-not-found.title'))
        .addButton("back", TRANSLATE("jig_ccomp:computer-not-found.back"))
        .setBody(TRANSLATE("jig_ccomp:computer-not-found.body"))
        .open({
        player,
        ...defaultScriptDialogueOptions
    });
};
const openPostOptions = async (player, opId) => {
    let dialogue = multiButtonScriptDialogue(TRANSLATE('jig_ccomp:post-options.title'))
        .setBody(TRANSLATE('jig_ccomp:post-options.body'))
        .addButton('reply', TRANSLATE('jig_ccomp:post-options.reply-thread'), 'textures/jig/ccomp/myui/reply-thread.png');
    if (player.name !== opId) {
        dialogue = dialogue.addButton('reply-op', TRANSLATE('jig_ccomp:post-options.reply-op'), 'textures/jig/ccomp/myui/reply.png');
    }
    return await dialogue
        .addButton('hide', TRANSLATE('jig_ccomp:post-options.hide'), 'textures/jig/ccomp/myui/hide.png')
        .addButton('block', TRANSLATE('jig_ccomp:post-options.block'), 'textures/jig/ccomp/myui/block.png')
        .addButton('back', TRANSLATE('jig_ccomp:post-options.back'))
        .open({
        player,
        ...defaultScriptDialogueOptions
    });
};
const openBlock = async (player, users) => {
    return await multiButtonScriptDialogue(TRANSLATE('jig_ccomp:post-block.title'))
        .setBody(users.length === 0 ? TRANSLATE('jig_ccomp:post-block.body-empty') : TRANSLATE('jig_ccomp:post-block.body'))
        .addButtons(users.map(u => ({
        name: u.id,
        text: u.name,
        iconPath: 'textures/jig/ccomp/myui/block.png'
    })))
        .addButton('back', TRANSLATE('jig_ccomp:post-block.back'))
        .open({
        player,
        ...defaultScriptDialogueOptions
    });
};
//const openBlock = async (player: Player, users: Array<ForumUser>) => {
//	
//	const unblockedUsers = users.filter(user => !isPlayerBlocked('player', user.id, blocks));
//	
//	return await multiButtonScriptDialogue(TRANSLATE('jig_ccomp:post-block.title'))
//	  .setBody(unblockedUsers.length === 0 ? TRANSLATE('jig_ccomp:post-block.body-empty') : TRANSLATE('jig_ccomp:post-block.body'))
//	  .addButtons(unblockedUsers.map(u => ({
//		name: u.id,
//		text: u.name
//	  }),'textures/jig/ccomp/myui/block.png'))
//	  .addButton('back', TRANSLATE('jig_ccomp:post-block.back'))
//	  .open({
//		player,
//		...defaultScriptDialogueOptions
//	  });
// }
const getPosts = () => {
    return GET_PROP(world, POST_KEY, []);
};
const getPlayerPosts = (playerData, blocks) => {
    return getPosts().filter(p => {
        if (playerData.hidden_posts.includes(p.id)) {
            return false;
        }
        if (isPlayerBlocked(p.op_type, p.op_id, blocks)) {
            return false;
        }
        return true;
    });
};
const getPostById = (postId, player, computer) => {
    getPlayerData(player);
    const post = getPosts().find(p => p.id === postId);
    if (post) {
        post.replies = post.replies.map(r => {
            // replies are from players for now
            r.playerBlocked = isPlayerBlocked('player', r.op_id, computer.getEmailBlocks());
            return r;
        });
    }
    return post;
};
const getPlayerData = (player) => {
    return GET_PROP(player, POST_KEY, {
        hidden_posts: [],
        last_reads: {}
    });
};
const updatePlayerDataReadPost = (player, postId) => {
    const playerData = getPlayerData(player);
    playerData.last_reads[postId] = world.getAbsoluteTime();
    SET_PROP(player, POST_KEY, playerData);
};
const updatePlayerDataHidePost = (player, postId) => {
    const playerData = getPlayerData(player);
    playerData.hidden_posts.push(postId);
    SET_PROP(player, POST_KEY, playerData);
};
const postIsRead = (post, data) => {
    if (post.id in data.last_reads) {
        return post.last_update <= data.last_reads[post.id];
    }
    return false;
};
const addPost = (post) => {
    const posts = getPosts();
    posts.unshift(post);
    if (posts.length > Config.forum.post_capacity) {
        posts.splice(posts.length - 1, posts.length - Config.forum.post_capacity);
    }
    SET_PROP(world, POST_KEY, posts);
};
const addReply = (postId, reply) => {
    // Update updated time
    const posts = getPosts();
    const index = posts.findIndex(p => p.id === postId);
    if (index >= 0) {
        const post = posts[index];
        post.last_update = world.getAbsoluteTime();
        post.replies.push(reply);
        posts.splice(index, 1);
        posts.unshift(post);
        SET_PROP(world, POST_KEY, posts);
    }
};
const cleanupPlayerData = (player) => {
    const postIds = getPosts().map(p => p.id);
    const playerData = getPlayerData(player);
    for (const lastReadId of Object.keys(playerData.last_reads)) {
        if (!postIds.includes(lastReadId)) {
            delete playerData.last_reads[lastReadId];
        }
    }
    playerData.hidden_posts.filter(hiddenPostId => postIds.includes(hiddenPostId));
    SET_PROP(player, POST_KEY, playerData);
};

const RANDOM_PC_ID = "jig_ccomp:random-pc";
const NPC_PREFIX = "npc:";
const PLAYER_PREFIX = "player:";
const OFFLINE_COMPUTER = "offline:";
const initEmail = () => {
    subscribeFromComputerNpcDialogue((event) => {
        if (event.id === 'jig_ccomp:send_email') {
            const { use_predefined_content: usePredefinedContent } = JSON.parse(event.message);
            executeScriptDialogueFromNpcDialogue(() => sendEmail(event.player, event.computer, usePredefinedContent));
        }
        else if (event.id === 'jig_ccomp:read_emails') {
            executeScriptDialogueFromNpcDialogue(() => readEmails(event.player, event.computer));
        }
        else if (event.id === 'jig_ccomp:computer_blocks') {
            executeScriptDialogueFromNpcDialogue(() => unblockEmails(event.player, event.computer));
        }
        else if (event.id === 'jig_ccomp:delete_emails') {
            executeScriptDialogueFromNpcDialogue(() => deleteEmails(event.player, event.computer));
        }
    }, {
        namespaces: ['jig_ccomp']
    });
    system.afterEvents.scriptEventReceive.subscribe((event) => {
        if (event.id === 'jig_ccomp:clear_player_npc' && event.sourceEntity instanceof Player) {
            if (event.sourceEntity.hasTag("debug")) {
                clearNpcs(event.sourceEntity);
            }
        }
        else if (event.id === 'jig_ccomp:email_broadcast') {
            const broadcast = JSON.parse(event.message);
            const errors = [];
            // Validate fields before sending:
            if (typeof broadcast.addon !== "string") {
                errors.push(`addon needs to be string`);
            }
            if (typeof broadcast.computer !== "string") {
                errors.push(`computer needs to be string`);
            }
            try {
                if (broadcast.attachment) {
                    new ItemStack(broadcast.attachment, broadcast.attachmentCount ?? 1);
                }
            }
            catch (e) {
                errors.push(`Error while loading item ${broadcast.attachment} x${broadcast.attachmentCount ?? 1}`);
            }
            if (typeof broadcast.addon_name !== "object") {
                errors.push(`addon_name needs a RawMessage`);
            }
            const typeofComputerName = typeof broadcast.computer_name;
            if (typeofComputerName !== "string" && typeofComputerName !== "object") {
                errors.push(`computer_name needs to be string or RawMessage`);
            }
            const typeofSubject = typeof broadcast.subject;
            if (typeofSubject !== "string" && typeofSubject !== "object") {
                errors.push(`subject needs to be string or RawMessage`);
            }
            const typeofMessage = typeof broadcast.message;
            if (typeofMessage !== "string" && typeofMessage !== "object") {
                errors.push(`message needs to be string or RawMessage`);
            }
            if (errors.length > 0) {
                debug(`Found errors while sending a broadcast message. Aborting the message.\n Received: ${JSON.stringify(broadcast, undefined, 2)} \n Errors:\n- ${errors.join("\n- ")}`);
                return;
            }
            // All seems well, prefix the computer with the addon
            broadcast.computer = `${broadcast.addon}:${broadcast.computer}`;
            craftComputers.broadcastEmail(broadcast);
        }
    }, {
        namespaces: ['jig_ccomp']
    });
    const getPlayerByName = (name) => {
        if (name === undefined) {
            return undefined;
        }
        const players = world.getPlayers({
            name
        });
        if (players.length > 0) {
            return players[0];
        }
        return undefined;
    };
    world.afterEvents.entitySpawn.subscribe((event) => {
        if (CraftComputer.isComputer(event.entity)) {
            const computer = new CraftComputer(event.entity);
            const creator = getPlayerByName(computer.getCreator());
            if (creator) {
                const playerAlreadyContacted = GET_RAW_PROP(creator, Config.player.jignetContacted);
                if (typeof playerAlreadyContacted === "boolean") {
                    computer.setWelcomeSent(playerAlreadyContacted);
                }
            }
            if (!computer.getWelcomeSent()) {
                system.runTimeout(() => {
                    if (computer.computer.isValid()) {
                        if (creator?.isValid() === true && GET_RAW_PROP(creator, Config.player.jignetContacted) === true) {
                            computer.setWelcomeSent(true);
                            return;
                        }
                        void sendEmailFromJigarBox(computer);
                        computer.setWelcomeSent(true);
                        if (creator) {
                            SET_RAW_PROP(creator, Config.player.jignetContacted, true);
                        }
                    }
                }, 20 * 30);
            }
        }
    });
    system.runInterval(async () => {
        const availableNpcs = Object.keys(Object.fromEntries(Object.entries(npcRecipients)
            .filter(e => hasUnpostedForumEntries(e[0], e[1]))));
        if (availableNpcs.length > 0) {
            const chosenNpc = availableNpcs[makeRandomNumber(0, availableNpcs.length - 1)];
            const npc = npcRecipients[chosenNpc];
            if (npc.forum) {
                const postIndex = makeRandomNumber(0, npc.forum.length - 1);
                makeNpcPost(chosenNpc, npc.name, npc.forum[postIndex]);
                markForumPosted(chosenNpc, postIndex);
            }
        }
    }, Config.forum.random_npc_post_ticks);
};
const sendEmail = async (player, computer, usePredefinedContent, sendEmailOptions) => {
    if (!computer.getCreator()) {
        computer.setCreator(player.name);
    }
    // Todo: Add input handling, show a message and redraw with whatever the user had
    //  if additional input is needed
    const playerInventory = player.getComponent(EntityInventoryComponent.componentId);
    const emailAttachments = [];
    if (playerInventory?.container) {
        const container = playerInventory.container;
        // Get all inventory except equipped
        for (let i = 0; i < container.size; ++i) {
            const item = container.getItem(i);
            if (item) {
                emailAttachments.push({
                    item,
                    index: i,
                    container
                });
            }
        }
        computer.computer.triggerEvent('jig_ccomp:typing');
    }
    const allowMobs = isAdminSettingsEnabled("mobEmails");
    const connectedComputers = craftComputers.getComputers().filter(c => c.getId() !== computer.getId());
    const connectedNpcs = allowMobs ? Object.keys(getEmaileableNpcs(player)).filter(npc => npc in npcRecipients) : [Config.JigarBoxNpcId];
    const storedComputers = craftComputers.getReachableStoredComputerRecipients();
    if (sendEmailOptions?.to) {
        if (connectedComputers.find(c => c.getId() === sendEmailOptions.to)) {
            sendEmailOptions.to = `${PLAYER_PREFIX}${sendEmailOptions.to}`;
        }
        else if (connectedNpcs.find(npc => npc === sendEmailOptions.to)) {
            sendEmailOptions.to = `${NPC_PREFIX}${sendEmailOptions.to}`;
        }
        else if (storedComputers.find(c => c.id === sendEmailOptions.to)) {
            sendEmailOptions.to = `${OFFLINE_COMPUTER}${sendEmailOptions.to}`;
        }
    }
    const sortEmailRecipientByName = (p1, p2) => {
        const name1 = typeof p1.name === "string" ? p1.name.toLowerCase() : p1.name;
        const name2 = typeof p2.name === "string" ? p2.name.toLowerCase() : p2.name;
        if (name1 < name2) {
            return -1;
        }
        else if (name1 > name2) {
            return 1;
        }
        return 0;
    };
    const sortEmailRecipientById = (p1, p2) => {
        if (p1.id < p2.id) {
            return -1;
        }
        else if (p1.id > p2.id) {
            return 1;
        }
        return 0;
    };
    const emailRecipients = connectedComputers.map(c => ({
        name: `${c.getName()} (${c.getId().substring(0, 4)})`,
        id: `${PLAYER_PREFIX}${c.getId()}`,
    })).sort(sortEmailRecipientByName);
    const storedReachableComputers = craftComputers.getReachableStoredComputerRecipients();
    if (storedReachableComputers.length > 0) {
        emailRecipients.push(...storedReachableComputers.map(c => ({
            id: `${OFFLINE_COMPUTER}${c.id}`,
            name: `${c.name} (${c.id.substring(0, 4)})`
        })));
    }
    if (allowMobs) {
        if (getUnknownNpcs(player).length > 0) {
            emailRecipients.push({
                id: `${NPC_PREFIX}${RANDOM_PC_ID}`,
                name: TRANSLATE("jig_ccomp:random-pc"),
            });
        }
    }
    emailRecipients.push(...connectedNpcs.map(npc => ({
        id: `${NPC_PREFIX}${npc}`,
        name: npcRecipients[npc].name,
    })).sort(sortEmailRecipientById));
    const response = await sendEmailDialogue(emailRecipients, emailAttachments, sendEmailOptions?.to, sendEmailOptions?.subject, usePredefinedContent).open({ ...defaultScriptDialogueOptions, player });
    computer.computer.triggerEvent('jig_ccomp:typing');
    if (response instanceof InputScriptDialogueResponse) {
        const { to, subject, message, attachment } = response.values;
        const emailSubject = usePredefinedContent ? TRANSLATE(subject) : subject;
        const emailMessage = usePredefinedContent ? TRANSLATE(message) : message;
        let itemToAttach;
        if (attachment !== 'none') {
            itemToAttach = emailAttachments[parseInt(attachment)];
            world.scoreboard.getObjective('jig_ccomp:jig_computer.addon_stats')?.addScore('jig_ccomp:attachments_sent', 1);
        }
        if (to.startsWith(NPC_PREFIX)) {
            const target = to.substring(NPC_PREFIX.length);
            sendNpcEmail(target, player, computer, emailSubject, emailMessage, itemToAttach);
        }
        else {
            let target;
            let isOnline;
            if (to.startsWith(PLAYER_PREFIX)) {
                target = to.substring(PLAYER_PREFIX.length);
                isOnline = true;
            }
            else {
                target = to.substring(OFFLINE_COMPUTER.length);
                isOnline = false;
            }
            let emailSent;
            const targetEmail = {
                sender_type: 'player',
                from_computer_id: computer.getId(),
                from_computer_name: computer.getName(),
                from_player_id: player.name,
                from_player: player.name,
                subject: emailSubject,
                message: emailMessage,
                printable: !usePredefinedContent,
                isRead: false,
            };
            if (isOnline) {
                const targetComputer = craftComputers.getComputer(target);
                emailSent = targetComputer.addEmail(targetEmail, itemToAttach);
            }
            else {
                if (itemToAttach) {
                    computer.addEmail({
                        sender_type: 'system',
                        from_computer_id: 'jig_ccomp:system',
                        from_computer_name: TRANSLATE("jig_ccomp:system_email_computer"),
                        from_player_id: 'jig_ccomp:system',
                        from_player: TRANSLATE("jig_ccomp:system_email_player"),
                        subject: TRANSLATE("jig_ccomp:system_offline.subject"),
                        message: TRANSLATE("jig_ccomp:system_offline.message"),
                        printable: false,
                        isRead: false
                    });
                    return;
                }
                const blocks = GET_PROP(world, PROP(IDENTIFIER_BLOCKS(target)), []);
                if (evaluateBlocks(blocks, targetEmail)) {
                    emailSent = 'blocked';
                }
                else {
                    SET_PROP(world, IDENTIFIER_EMAILS(target), [
                        ...GET_PROP(world, IDENTIFIER_EMAILS(target), []),
                        {
                            ...targetEmail,
                            id: makeRandomId()
                        }
                    ]);
                    SET_RAW_PROP(world, IDENTIFIER_TRIGGER_EMAIL_NOTIFICATION(target), true);
                    emailSent = true;
                }
            }
            if (emailSent === 'blocked') {
                computer.addEmail({
                    sender_type: 'system',
                    from_computer_id: 'jig_ccomp:system',
                    from_computer_name: TRANSLATE("jig_ccomp:system_email_computer"),
                    from_player_id: 'jig_ccomp:system',
                    from_player: TRANSLATE("jig_ccomp:system_email_player"),
                    subject: TRANSLATE("jig_ccomp:system_blocked.subject"),
                    message: TRANSLATE("jig_ccomp:system_blocked.message"),
                    printable: false,
                    isRead: false
                });
            }
            if (emailSent === false) {
                computer.addEmail({
                    sender_type: 'system',
                    from_computer_id: 'jig_ccomp:system',
                    from_computer_name: TRANSLATE("jig_ccomp:system_email_computer"),
                    from_player_id: 'jig_ccomp:system',
                    from_player: TRANSLATE("jig_ccomp:system_email_player"),
                    subject: TRANSLATE("jig_ccomp:system_full.subject"),
                    message: TRANSLATE("jig_ccomp:system_full.message"),
                    printable: false,
                    isRead: false
                });
            }
            if (emailSent) {
                world.scoreboard.getObjective('jig_ccomp:jig_computer.addon_stats')?.addScore('jig_ccomp:emails_sent', 1);
                player.playSound('jig_ccomp.email_send');
                computer.computer.triggerEvent('jig_ccomp:screen_emas');
                computer.computer.triggerEvent('jig_ccomp:typing');
                if (isOnline) {
                    debug(`Player ${player.name} sent an email from computer ${computer.getName()} to computer ${craftComputers.getComputer(target).getName()} with the subject: ${subject}`);
                }
                else {
                    debug(`Player ${player.name} sent an email from computer ${computer.getName()} to offline computer identified by id ${target} with the subject: ${subject}`);
                }
            }
        }
    }
};
const readEmails = async (player, computer) => {
    if (!computer.getCreator()) {
        computer.setCreator(player.name);
    }
    let emails = computer.getEmails().reverse();
    const paper = "jig_ccomp:printed_paper";
    if (emails.length > 0) {
        while (true) {
            emails = computer.getEmails().reverse();
            // We don't have any other email
            if (emails.length === 0) {
                computer.computer.runCommand(`/dialogue open @s "${player.name}" jig_ccomp:old_pc_nome`);
                break;
            }
            computer.computer.triggerEvent('jig_ccomp:typing');
            const response = await readEmailDialogue(emails, player, computer).open({ ...defaultScriptDialogueOptions, player });
            computer.computer.triggerEvent('jig_ccomp:typing');
            if (response instanceof ButtonDialogueResponse) {
                computer.markAsRead(response.selected);
                const email = computer.getEmailById(response.selected);
                const unknownNpcs = getUnknownNpcs(player);
                if (unknownNpcs.includes(email.from_computer_id)) {
                    debug(`Adding npc: ${email.from_computer_id}`);
                    addNpc(player, email.from_computer_id, true);
                }
            }
            // Todo: Add helper method to make it easier to get nested data
            if (response instanceof ButtonDialogueResponse && response.callback) { // Select action
                const selectedEmail = response.selected;
                const selectionAction = response.callback;
                if (selectionAction === "download") {
                    if (computer.downloadAttachment(selectedEmail, player)) {
                        debug(`File downloaded from computer: ${computer.getName()} to player: ${player.name}`);
                    }
                    break;
                }
                else if (selectionAction === "delete") {
                    computer.deleteEmail(selectedEmail);
                    debug(`Email deleted from computer: ${computer.getName()}`);
                    player.playSound('jig_ccomp.email_delete');
                    computer.computer.triggerEvent('jig_ccomp:typing');
                    continue;
                }
                else if (selectionAction === "print") {
                    const email = computer.getEmails().find(e => e.id === selectedEmail);
                    printContent(`Email from: ${email.from_player}`, `${email.subject}\n\n${email.message}`, computer, player, paper);
                    player.playSound('jig_common.pcbloop');
                    computer.computer.triggerEvent('jig_ccomp:typing');
                    break;
                }
                else if (selectionAction === "reply" || selectionAction === "reply-custom") {
                    const useTemplates = selectionAction === "reply";
                    player.playSound('jig_common.pcbloop');
                    const email = computer.getEmails().find(e => e.id === selectedEmail);
                    await sendEmail(player, computer, useTemplates, {
                        to: email.from_computer_id,
                        subject: prependToStringOrRawMessage(email.subject, 'RE:', false)
                    });
                    computer.computer.triggerEvent('jig_ccomp:typing');
                    break;
                }
                else if (selectionAction === "block") {
                    const email = computer.getEmails().find(e => e.id === selectedEmail);
                    let blocked = undefined;
                    switch (email.sender_type) {
                        case "player":
                        case "script":
                            blocked = computer.getBlocked(email.sender_type, "computer", email.from_computer_id)
                                || computer.getBlocked(email.sender_type, "player", email.from_player_id);
                            break;
                        case "npc":
                            blocked = computer.getBlocked(email.sender_type, "player", email.from_player_id);
                    }
                    if (blocked) {
                        const unblockResponse = await unblock(blocked).open({ player });
                        if (unblockResponse instanceof ButtonDialogueResponse) {
                            if (unblockResponse.selected === "yes") {
                                computer.removeBlock(blocked.id);
                                computer.computer.runCommand(`/dialogue open @s "${player.name}" jig_ccomp:old_pc_ulkg`);
                            }
                        }
                    }
                    else {
                        if (email.sender_type === "player") {
                            const blockReply = await blockPlayer(email.from_player, email.from_computer_name).open({
                                ...defaultScriptDialogueOptions,
                                player
                            });
                            if (blockReply instanceof ButtonDialogueResponse) {
                                switch (blockReply.selected) {
                                    case "both":
                                        computer.addPlayerBlock(email);
                                        computer.addComputerBlock(email);
                                        computer.computer.runCommand(`/dialogue open @s "${player.name}" jig_ccomp:old_pc_blkb`);
                                        break;
                                    case "computer":
                                        computer.addComputerBlock(email);
                                        computer.computer.runCommand(`/dialogue open @s "${player.name}" jig_ccomp:old_pc_blkc`);
                                        break;
                                    case "player":
                                        computer.addPlayerBlock(email);
                                        computer.computer.runCommand(`/dialogue open @s "${player.name}" jig_ccomp:old_pc_blkp`);
                                        break;
                                }
                            }
                        }
                        else if (email.sender_type === "npc") {
                            const blockReply = await blockNpc(TEXT_OR_RAW_TEXT(email.from_player)).open({
                                ...defaultScriptDialogueOptions,
                                player
                            });
                            if (blockReply instanceof ButtonDialogueResponse) {
                                if (blockReply.selected === "yes") {
                                    computer.addPlayerBlock(email);
                                    computer.computer.runCommand(`/dialogue open @s "${player.name}" jig_ccomp:old_pc_blkn`);
                                }
                            }
                        }
                        else if (email.sender_type === "script") {
                            const blockReply = await blockScript(TEXT_OR_RAW_TEXT(email.from_player), TEXT_OR_RAW_TEXT(email.from_computer_name)).open({
                                ...defaultScriptDialogueOptions,
                                player
                            });
                            if (blockReply instanceof ButtonDialogueResponse) {
                                switch (blockReply.selected) {
                                    case "computer":
                                        computer.addComputerBlock(email);
                                        computer.computer.runCommand(`/dialogue open @s "${player.name}" jig_ccomp:old_pc_blkc`);
                                        break;
                                    case "script":
                                        computer.addPlayerBlock(email);
                                        computer.computer.runCommand(`/dialogue open @s "${player.name}" jig_ccomp:old_pc_blks`);
                                        break;
                                }
                            }
                        }
                    }
                }
            }
            computer.computer.triggerEvent('jig_ccomp:typing');
            computer.computer.runCommand(`/dialogue open @s "${player.name}" jig_ccomp:old_pc_emai`);
            break;
        }
    }
    else {
        computer.computer.runCommand(`/dialogue open @s "${player.name}" jig_ccomp:old_pc_noem`);
    }
};
const sendNpcEmail = (target, player, computer, originalSubject, message, attachment) => {
    if (target === RANDOM_PC_ID) {
        // Pick one from the not know list
        const npcs = getUnknownNpcs(player);
        target = npcs[makeRandomNumber(0, npcs.length - 1)];
        addNpc(player, target, false);
        debug(`Sending email to random mob.`);
    }
    debug(`Selected npc is ${target}`);
    world.scoreboard.getObjective('jig_ccomp:jig_computer.addon_stats')?.addScore('jig_ccomp:emails_sent', 1);
    player.playSound('jig_ccomp.email_send');
    computer.computer.triggerEvent('jig_ccomp:screen_emas');
    computer.computer.triggerEvent('jig_ccomp:typing');
    let responseDelay = makeRandomNumber(2000, 8000);
    if (player.hasTag("jig_ccomp:fast-emails")) {
        responseDelay = 20;
    }
    if (attachment) {
        responseDelay *= 0.5;
        if (attachment.container && attachment.index !== undefined) {
            attachment.container.setItem(attachment.index);
        }
    }
    const replier = npcRecipients[target];
    const npcs = getNpcs(player);
    const pendingEmail = npcs[target].pendingEmail;
    const email = {
        sender_type: 'npc',
        from_computer_name: target,
        from_computer_id: target,
        from_player_id: target,
        from_player: replier.name,
        subject: prependToStringOrRawMessage(originalSubject, "RE: ", true),
        message: '',
        printable: false,
        isRead: false,
        isTooFastEmail: false,
    };
    if (pendingEmail) {
        // Fetch the future operations, if it is a "isNotTooFastEmail" return
        // Else we need to fetch the future operations and patch it to send a too fast email
        const operations = getOperations(computer.computer);
        const operation = operations.find(o => {
            if (o.operation === "send-email-mob" /* COMPUTER_SEND_EMAIL_FROM_MOB */) {
                const [operation_target, _playerName, _email, _attachment] = o.params;
                if (target === operation_target) {
                    return true;
                }
            }
            return false;
        });
        if (operation) {
            const [_target, playerName, email, _attachment] = operation.params;
            if (email.isTooFastEmail) {
                // all is set, nothing to do
                return;
            }
            else {
                // patch the operation
                system.clearRun(operation.timerId);
                removeOperation(computer.computer, operation);
                const craftEmail = email;
                craftEmail.isTooFastEmail = true;
                const npcMessage = npcRecipients[target].too_many;
                craftEmail.message = npcMessage.message;
                operation.params = [target, playerName, craftEmail, npcMessage.attachment];
                addOperation(computer.computer, operation);
                runPendingOperations(computer.computer);
                return;
            }
        }
    }
    else {
        markPendingEmail(player, target, true);
    }
    const npcMessage = getNextMessage(target, player);
    const emailMessage = [npcMessage.message];
    const npcRecipient = npcRecipients[target];
    let favoriteEmailOfNPC;
    if (npcMessage.attachment && npcMessage.attachment[0] == Config.favoriteEmailAttachtment) {
        favoriteEmailOfNPC = getFavoriteEmailOfNPC(target, player);
    }
    if (attachment) {
        if (npcRecipient.attachmentResponses && npcRecipient.attachmentResponses.length > 0) {
            emailMessage.unshift(npcRecipient.attachmentResponses[makeRandomNumber(0, npcRecipient.attachmentResponses.length - 1)], TEXT("\n"));
        }
    }
    if (typeof message === "string" && message && typeof originalSubject === "string" && originalSubject) {
        saveFavoriteEmail(player, target, originalSubject, message);
    }
    email.message = RAW_TEXT(...emailMessage);
    const playerName = player.name;
    scheduleOperation(computer.computer, {
        operation: "send-email-mob" /* COMPUTER_SEND_EMAIL_FROM_MOB */,
        timeFromNow: responseDelay,
        params: [target, playerName, email, npcMessage.attachment, favoriteEmailOfNPC]
    });
    scheduleOperation(player, {
        operation: "enable-npc" /* PLAYER_ENABLE_NPC */,
        timeFromNow: responseDelay,
        params: [target]
    });
};
const sendEmailFromJigarBox = async (computer) => {
    const target = Config.JigarBoxNpcId;
    const replier = npcRecipients[target];
    const email = {
        sender_type: 'npc',
        from_computer_name: target,
        from_computer_id: target,
        from_player_id: target,
        from_player: replier.name,
        subject: TRANSLATE('jig_ccomp:helpds.subject.xx.message'),
        message: TRANSLATE('jig_ccomp:helpds.message.xx.message'),
        printable: false,
        isRead: false,
        isTooFastEmail: false,
    };
    computer.addEmail(email);
};
const unblockEmails = async (player, computer) => {
    const blocks = computer.getEmailBlocks();
    if (blocks.length > 0) {
        const response = await unblockList(computer.getEmailBlocks()).open({
            ...defaultScriptDialogueOptions,
            player
        });
        if (response instanceof ButtonDialogueResponse) {
            computer.removeBlock(response.selected);
            computer.computer.runCommand(`/dialogue open @s "${player.name}" jig_ccomp:old_pc_ulkg`);
        }
    }
    else {
        computer.computer.runCommand(`/dialogue open @s "${player.name}" jig_ccomp:old_pc_ulkx`);
    }
};
const deleteEmails = async (player, computer) => {
    // First check
    computer.computer.triggerEvent('jig_ccomp:typing');
    if (computer.getEmails().length === 0) {
        computer.computer.runCommand(`/dialogue open @s "${player.name}" jig_ccomp:old_pc_noem`);
        return;
    }
    while (player.isValid()) {
        const emails = computer.getEmails();
        if (emails.length === 0) {
            computer.computer.runCommand(`/dialogue open @s "${player.name}" jig_ccomp:old_pc_nome`);
            break;
        }
        const response = await deleteEmailsDialogue(emails).open({
            player,
            ...defaultScriptDialogueOptions
        });
        if (response instanceof ButtonDialogueResponse) {
            computer.computer.triggerEvent('jig_ccomp:typing');
            if (response.selected === "cancel") {
                break;
            }
            else {
                player.playSound('jig_ccomp.email_delete');
                computer.deleteEmail(response.selected);
            }
        }
        else {
            break;
        }
    }
};
const getFavoriteEmailOfNPC = (from, player) => {
    const npcs = getNpcs(player);
    const npc = npcs[from];
    if (npc.favoriteEmails && npc.favoriteEmails.length > 0) {
        return npc.favoriteEmails[makeRandomNumber(0, npc.favoriteEmails.length - 1)];
    }
};
const getNextMessage = (from, player) => {
    const npcs = getNpcs(player);
    const npc = npcs[from];
    const npcRecipient = npcRecipients[from];
    if (npc.nextResponse === -1) {
        incrementNpcResponse(player, from);
        return npcRecipient.introduction;
    }
    else if (npc.nextResponse >= npcRecipient.responses.length) {
        if (npc.favoriteEmails && npc.favoriteEmails.length > 0) {
            let sendFavoriteEmail = (makeRandomNumber(0, 100) > 80) ? true : false;
            if (sendFavoriteEmail) {
                if (npcRecipient.favoriteMailResponses && npcRecipient.favoriteMailResponses.length > 0) {
                    return {
                        attachment: [Config.favoriteEmailAttachtment, 0],
                        message: npcRecipient.favoriteMailResponses[makeRandomNumber(0, npcRecipient.favoriteMailResponses.length - 1)]
                    };
                }
            }
        }
        if (npcRecipient.random && npcRecipient.random.length > 0) {
            return npcRecipient.random[makeRandomNumber(0, npcRecipient.random.length - 1)];
        }
        else {
            return npcRecipient.responses[makeRandomNumber(0, npcRecipient.responses.length - 1)];
        }
    }
    else {
        incrementNpcResponse(player, from);
        return npcRecipient.responses[npc.nextResponse];
    }
};
const getUnknownNpcs = (player) => {
    const knowNpcs = Object.keys(getNpcs(player));
    return Object.keys(npcRecipients).filter(r => !knowNpcs.includes(r));
};
const prependToStringOrRawMessage = (stringOrRawMessage, prepend, usePrefill) => {
    if (typeof stringOrRawMessage === 'string') {
        return prepend + stringOrRawMessage;
    }
    else {
        if (usePrefill) { // Once we lift this requirement, we can remove this path
            return RAW_TEXT(TEXT(prepend), stringOrRawMessage);
        }
        else {
            return 'RE: Template';
        }
    }
};
const prependToStringOrRawMessageThread = (stringOrRawMessage, prepend, usePrefill) => {
    if (typeof stringOrRawMessage === 'string') {
        return prepend + stringOrRawMessage;
    }
    else {
        if (usePrefill) { // Once we lift this requirement, we can remove this path
            return RAW_TEXT(TEXT(prepend), stringOrRawMessage);
        }
        else {
            return 'RE: Forum Thread';
        }
    }
};
const operationSendEmailFromMob = (entity, params) => {
    const [target, playerName, email, attachment, favoriteEmail] = params;
    const computer = new CraftComputer(entity);
    debug(`Npc ${target} just answered to ${playerName}!`);
    let emailAttachment = undefined;
    if (attachment) {
        if (attachment[0] === Config.favoriteEmailAttachtment && favoriteEmail) {
            const paper = "jig_ccomp:printed_paper";
            const printedPaper = printContentFavoriteEmail(`Email from: ${playerName}`, `${favoriteEmail.subject}\n\n${favoriteEmail.message}`, computer, paper);
            emailAttachment = {
                item: printedPaper
            };
        }
        else {
            emailAttachment = {
                item: new ItemStack(attachment[0], attachment[1])
            };
        }
    }
    const emailSent = computer.addEmail(email, emailAttachment, false);
    if (emailSent === false) {
        let message = TRANSLATE("jig_ccomp:npc_full_inventory_default.message");
        const npcRecipient = npcRecipients[target];
        if (npcRecipient.inventoryFullResponses && npcRecipient.inventoryFullResponses.length > 0) {
            message = npcRecipient.inventoryFullResponses[makeRandomNumber(0, npcRecipient.inventoryFullResponses.length - 1)];
        }
        email.message = message;
        computer.addEmail(email, emailAttachment);
    }
};
const operationEnableNpc = (entity, params) => {
    const [npc] = params;
    enableNpcMails(entity, npc);
};
const getNpcUsedProp = (npcId) => PROP('npc-used-forum', npcId);
const hasUnpostedForumEntries = (npcId, data) => {
    if (data.forum && data.forum.length > 0) {
        const used = GET_PROP(world, getNpcUsedProp(npcId), []);
        return data.forum.some((_v, index) => !used.includes(index));
    }
    return false;
};
const markForumPosted = (npcId, postIndex) => {
    const used = GET_PROP(world, getNpcUsedProp(npcId), []);
    used.push(postIndex);
    SET_PROP(world, getNpcUsedProp(npcId), used);
};

var Email = /*#__PURE__*/Object.freeze({
  __proto__: null,
  getUnknownNpcs: getUnknownNpcs,
  initEmail: initEmail,
  operationEnableNpc: operationEnableNpc,
  operationSendEmailFromMob: operationSendEmailFromMob,
  prependToStringOrRawMessage: prependToStringOrRawMessage,
  prependToStringOrRawMessageThread: prependToStringOrRawMessageThread,
  sendEmail: sendEmail
});

const viewPortableOptions = (options) => {
    return inputScriptDialogue(TRANSLATE("jig_ccomp:portable.options.title"))
        // .addElement(inputToggle("name_tag", TRANSLATE("jig_ccomp:portable.options.name_tag"), options.name_tag))
        .addElement(inputToggle("coordinates", TRANSLATE("jig_ccomp:portable.options.coordinates"), options.coordinates))
        .addElement(inputToggle("biome", TRANSLATE("jig_ccomp:portable.options.biome"), options.biome))
        .addElement(inputToggle("light_level", TRANSLATE("jig_ccomp:portable.options.light_level"), options.light_level))
        .addElement(inputToggle("chunk", TRANSLATE("jig_ccomp:portable.options.chunk_border"), options.chunk))
        .addElement(inputToggle("schunk", TRANSLATE("jig_ccomp:portable.options.schunk_border"), options.schunk))
        .addElement(inputToggle("entityparticles", TRANSLATE("jig_ccomp:portable.options.entityparticles", "\n"), options.entityparticles));
};

const initPortable = () => {
    system.afterEvents.scriptEventReceive.subscribe(async (event) => {
        if (event.sourceEntity && event.id === 'jig_ccomp:portable_settings') {
            const players = event.sourceEntity.dimension.getEntities({
                location: event.sourceEntity.location,
                type: "minecraft:player",
                closest: 1
            });
            if (players.length > 0 && players[0] instanceof Player) {
                const player = players[0];
                //const initial_name_tag = event.sourceEntity.getProperty("jig_ccomp:name_tag") as boolean;
                const initial_coordinates = event.sourceEntity.getProperty("jig_ccomp:show_coordinates");
                const initial_light_level = event.sourceEntity.getProperty("jig_ccomp:show_lightlevel");
                const initial_biome = event.sourceEntity.getProperty("jig_ccomp:show_biome");
                const initial_chunk_border = event.sourceEntity.getProperty("jig_ccomp:show_chunk_border");
                const initial_schunk_border = event.sourceEntity.getProperty("jig_ccomp:show_schunk_border");
                const initial_entity_particles = event.sourceEntity.getProperty("jig_ccomp:show_entity_particles");
                player.playSound('jig_common.beeps.say');
                const response = await viewPortableOptions({
                    // name_tag: initial_name_tag
                    biome: initial_biome,
                    coordinates: initial_coordinates,
                    light_level: initial_light_level,
                    chunk: initial_chunk_border,
                    schunk: initial_schunk_border,
                    entityparticles: initial_entity_particles
                }).open({ ...defaultScriptDialogueOptions, player });
                if (response instanceof InputScriptDialogueResponse) {
                    player.playSound('jig_common.beeps.say');
                    const { 
                    //name_tag,
                    coordinates, light_level, biome, chunk, schunk, entityparticles } = response.values;
                    //event.sourceEntity.setProperty("jig_ccomp:name_tag", name_tag.value as boolean);
                    event.sourceEntity.setProperty("jig_ccomp:show_coordinates", coordinates);
                    event.sourceEntity.setProperty("jig_ccomp:show_lightlevel", light_level);
                    event.sourceEntity.setProperty("jig_ccomp:show_biome", biome);
                    event.sourceEntity.setProperty("jig_ccomp:show_chunk_border", chunk);
                    event.sourceEntity.setProperty("jig_ccomp:show_schunk_border", schunk);
                    event.sourceEntity.setProperty("jig_ccomp:show_entity_particles", entityparticles);
                    //if(name_tag.value as boolean === true){event.sourceEntity.triggerEvent("jig_ccomp:always_show")}else{event.sourceEntity.triggerEvent("jig_ccomp:sometimes_show")}
                }
            }
        }
    }, {
        namespaces: ["jig_ccomp"]
    });
};

const viewCreativeProtect = () => {
    return dualButtonScriptDialogue(RAW_TEXT(TRANSLATE("jig_ccomp:view-creative-protect.title")), {
        name: "fix-it",
        text: TRANSLATE("jig_ccomp:fix-game-mode")
    }, {
        name: "do-nothing",
        text: TRANSLATE("jig_ccomp:do-nothing"),
    }).setBody(RAW_TEXT(TRANSLATE("jig_ccomp:view-creative-protect.body", "\n\n")));
};

const initCreativeProtect = () => {
    system.afterEvents.scriptEventReceive.subscribe(async (event) => {
        //open this on playerinteractwithentitybeforeevent (or after tbh)
        if (event.sourceEntity && event.id === 'jig_ccomp:creative-protect') {
            const players = event.sourceEntity.dimension.getEntities({
                location: event.sourceEntity.location,
                type: "minecraft:player",
                gameMode: GameMode.creative,
                maxDistance: 8
            });
            event.sourceEntity.runCommand('/tellraw @a[r=8] {"rawtext":[{"translate":"jig_ccomp:computer_interacted_creative"}]}');
            event.sourceEntity.runCommand('/playsound mob.bat.death @a[r=8] ~ ~ ~ 1 0.5');
            if (players.length > 0) {
                players.forEach((player) => {
                    if (player instanceof Player) {
                        const asyncFunction = async () => {
                            const response = await viewCreativeProtect().open({ ...defaultScriptDialogueOptions, player });
                            if (response instanceof ButtonDialogueResponse) {
                                if (response.selected === "fix-it") {
                                    player.runCommand("/gamemode survival");
                                }
                            }
                        };
                        void asyncFunction();
                    }
                });
            }
        }
    }, {
        namespaces: ["jig_ccomp"]
    });
};

const viewTut_Book_1 = () => {
    return multiButtonScriptDialogue(RAW_TEXT(TRANSLATE("jig_ccomp:book.name")))
        .setBody(RAW_TEXT(TRANSLATE("jig_ccomp:info_book.page_01.a", "\n\n"), TEXT('§r\n\n'), TRANSLATE("jig_ccomp:info_book.page_02.a", "\n"), TEXT('§r\n\n'), TRANSLATE("jig_ccomp:info_book.page_03.a", "\n"), TEXT('§r\n\n'), TRANSLATE("jig_ccomp:info_book.page_04.a", "\n"), TEXT('§r\n\n'), TRANSLATE("jig_ccomp:info_book.page_05.a", "\n"), TEXT('§r\n\n'), TRANSLATE("jig_ccomp:info_book.page_06.a", "\n"), TEXT('§r\n\n'), TRANSLATE("jig_ccomp:info_book.page_07.a", "\n"), TEXT('§r\n\n'), TRANSLATE("jig_ccomp:info_book.page_08.a", "\n"), TEXT('§r\n\n'), TRANSLATE("jig_ccomp:info_book.page_09.a", "\n"), TEXT('§r\n\n'), TRANSLATE("jig_ccomp:info_book.page_10.a", "\n"), TEXT('§r\n\n'), TRANSLATE("jig_ccomp:info_book.page_11.a", "\n"), TEXT('§r\n\n'), TRANSLATE("jig_ccomp:info_book.page_12.a", "\n"), TEXT('§r\n\n'), TRANSLATE("jig_ccomp:info_book.page_13.a", "\n"), TEXT('§r\n\n'), TRANSLATE("jig_ccomp:info_book.page_14.a", "\n"), TEXT('§r\n\n'), TRANSLATE("jig_ccomp:info_book.page_15.a", "\n"), TEXT('§r\n\n'), TRANSLATE("jig_ccomp:info_book.page_16.a", "\n"), TEXT('§r\n\n'), TRANSLATE("jig_ccomp:info_book.page_17.a", "\n"), TEXT('§r\n\n'), TRANSLATE("jig_ccomp:info_book.page_19.a", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n"), TEXT('§r\n\n'), TRANSLATE("jig_ccomp:info_book.page_20.a", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n"), TEXT('§r\n\n'), TRANSLATE("jig_ccomp:info_book.page_21.a", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n"), TEXT('§r\n\n'), TRANSLATE("jig_ccomp:info_book.page_22.a", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n", "\n"), TEXT('§r\n\n'), TRANSLATE("jig_ccomp:info_book.page_18.a", "\n\n", "\n", "\n\n")))
        .addButton('update', TRANSLATE("jig_ccomp:butt.recent_update"))
        .addButton('exit', TRANSLATE("jig_ccomp:butt.old_pc_exit"));
};
const viewTut_Book_2 = () => {
    return multiButtonScriptDialogue(RAW_TEXT(TRANSLATE("item.jig_ccomp:guide_book_u")))
        .setBody(RAW_TEXT(TRANSLATE("jig_ccomp:info_u.page_a1", "\n"), TEXT('§r\n\n'), TRANSLATE("jig_ccomp:info_u.page_a2", "\n"), TEXT('§r\n\n'), TRANSLATE("jig_ccomp:info_u.page_a3", "\n"), TEXT('§r\n\n'), TRANSLATE("jig_ccomp:info_u.page_a4", "\n"), TEXT('§r\n\n'), TRANSLATE("jig_ccomp:info_u.page_a5", "\n"), TEXT('§r\n\n'), TRANSLATE("jig_ccomp:info_u.page_a6", "\n"), TEXT('§r\n\n'), TRANSLATE("jig_ccomp:info_u.page_01", "\n"), TEXT('§r\n\n'), TRANSLATE("jig_ccomp:info_u.page_02", "\n"), TEXT('§r\n\n'), TRANSLATE("jig_ccomp:info_u.page_03", "\n"), TEXT('§r\n\n'), TRANSLATE("jig_ccomp:info_u.page_04", "\n"), TEXT('§r\n\n'), TRANSLATE("jig_ccomp:info_u.page_05", "\n"), TEXT('§r\n\n'), TRANSLATE("jig_ccomp:info_u.page_06", "\n"), TEXT('§r\n\n'), TRANSLATE("jig_ccomp:info_u.page_07", "\n"), TEXT('§r\n\n'), TRANSLATE("jig_ccomp:info_u.page_09", "\n")))
        .addButton('exit', TRANSLATE("jig_ccomp:butt.old_pc_exit"));
};

const GUIDE_BOOK = "jig_ccomp:guide_book";
const GUIDE_BOOK_U = "jig_ccomp:guide_book_u";
const initTut_Book = () => {
    world.afterEvents.itemUse.subscribe(async (event) => {
        let player = event.source;
        const { source, itemStack } = event;
        if (itemStack.typeId === GUIDE_BOOK) {
            player.playSound('jig_common.beeps.say');
            const response = await viewTut_Book_1().open({ ...defaultScriptDialogueOptions, player });
            if (response instanceof ButtonDialogueResponse) {
                player.playSound('jig_common.beeps.say');
                // Check if the response was 'update' and open viewTut_Book_2 accordingly
                if (response.selected === 'update') {
                    await viewTut_Book_2().open({ ...defaultScriptDialogueOptions, player });
                }
            }
        }
        if (itemStack.typeId === GUIDE_BOOK_U) {
            player.playSound('jig_common.beeps.say');
            const response = await viewTut_Book_2().open({ ...defaultScriptDialogueOptions, player });
            if (response instanceof InputScriptDialogueResponse) {
                player.playSound('jig_common.beeps.say');
            }
        }
    });
};

const inventory = (textAdventure) => {
    const inventory = textAdventure.state.inventory;
    if (inventory.length === 0) {
        return TRANSLATE("jig_ccomp:text-adventure.ui.no-items");
    }
    return RAW_TEXT(TRANSLATE("jig_ccomp:text-adventure.ui.items"), TEXT("\n"), ...inventory.map(i => RAW_TEXT(TEXT(`- ${i.count}x `), textAdventure.items[i.id].description, TEXT("\n"))));
};
const getTargets = (targets) => {
    return targets.filter(t => !t.hidden).map(t => RAW_TEXT(TEXT(" - "), t.description, TEXT("\n")));
};
const getExits = (textAdventure, exits) => {
    return exits.map(e => RAW_TEXT(TEXT(" - "), e.where, TEXT(": "), e.description ?? textAdventure.rooms[e.room].distance_description ?? TEXT(""), TEXT("\n")));
};

const room = (textAdventure) => {
    const currentRoom = textAdventure.rooms[textAdventure.state.currentRoom];
    return multiButtonScriptDialogue(TRANSLATE("jig_ccomp:text-adventure.ui.room.title"))
        .setBody(RAW_TEXT(currentRoom.description, TEXT("\n"), TRANSLATE("jig_ccomp:text-adventure.ui.room.targets"), TEXT("\n"), ...getTargets(currentRoom.targets), TEXT("\n"), TRANSLATE("jig_ccomp:text-adventure.ui.room.exits"), TEXT("\n"), ...getExits(textAdventure, currentRoom.exits), TEXT("\n"), inventory(textAdventure))).addButton("interact", TRANSLATE("jig_ccomp:text-adventure.ui.interact"))
        .addButton("walk", TRANSLATE("jig_ccomp:text-adventure.ui.walk"))
        .addButton("close", "jig_ccomp:text-adventure.ui.close");
};

const status = (textAdventure) => {
    return dualButtonScriptDialogue(TRANSLATE("jig_ccomp:text-adventure.ui.status.title"), {
        name: "continue",
        text: "jig_ccomp:text-adventure.ui.continue",
    }, {
        name: "close",
        text: "jig_ccomp:text-adventure.ui.close",
    }).setBody(RAW_TEXT(...textAdventure.state.nextText, TEXT("\n"), textAdventure.state.isAlive ? inventory(textAdventure) : TRANSLATE("jig_ccomp:text-adventure.ui.status.dead")));
};

const filterByItem = (textAdventure, collection) => {
    return collection.filter(what => {
        if (!hasItems(textAdventure, what.hasItems)) {
            return false;
        }
        if (!doesNotHaveItems(textAdventure, what.doesNotHaveItem)) {
            return false;
        }
        return true;
    });
};
const doesNotHaveItems = (textAdventure, items) => {
    if (items) {
        for (const item of items) {
            if (hasItem(textAdventure, item)) {
                return false;
            }
        }
    }
    return true;
};
const hasItems = (textAdventure, items) => {
    if (items) {
        for (const item of items) {
            if (!hasItem(textAdventure, item)) {
                return false;
            }
        }
    }
    return true;
};
const hasItem = (textAdventure, item) => {
    if (typeof item === 'string') {
        return !!textAdventure.state.inventory.find(i => i.id === item);
    }
    else {
        return !!textAdventure.state.inventory.find(i => i.id === item.id && i.count >= item.count);
    }
};

const subRoomInteract = (textAdventure) => {
    const currentRoom = textAdventure.rooms[textAdventure.state.currentRoom];
    return multiButtonScriptDialogue(TRANSLATE("jig_ccomp:text-adventure.ui.interact.title"))
        .setBody(RAW_TEXT(...getTargets(currentRoom.targets)))
        .addButtons(currentRoom.targets.filter(t => !t.hidden).map(t => ({
        name: t.id,
        text: t.description,
    })));
};

const subRoomWalk = (textAdventure) => {
    const currentRoom = textAdventure.rooms[textAdventure.state.currentRoom];
    return multiButtonScriptDialogue(TRANSLATE("jig_ccomp:text-adventure.ui.walk.title"))
        .setBody(RAW_TEXT(...getExits(textAdventure, filterByItem(textAdventure, currentRoom.exits))))
        .addButtons(currentRoom.exits.map(e => ({
        name: e.id,
        text: e.where,
    })));
};

const makeTextAdventure = (input) => ({
    rooms: input.rooms,
    items: input.items,
    state: {
        inventory: input.inventory ?? [],
        isAlive: true,
        nextText: input.initialText ?? [],
        currentRoom: input.initialRoom,
    }
});
const validateInventory = (inventory, items) => {
    inventory.forEach(item => {
        if (!(item.id in items)) {
            throw new Error(`Item not defined ${item.id}`);
        }
    });
};
const SAY = (what) => ({
    action: (state, context) => {
        state.nextText.push(what);
        state.nextText.push(TEXT("\n"));
    }
});
const DIE = () => ({
    action: (state, context) => {
        state.isAlive = false;
    }
});
const GET_ITEM = (inventoryItemId, count = 1) => ({
    action: (state, context) => {
        const item = state.inventory.find(i => i.id === inventoryItemId);
        if (item) {
            item.count += count;
        }
        else {
            state.inventory.push({
                id: inventoryItemId,
                count
            });
            validateInventory(state.inventory, context.items);
        }
    }
});
const setTargetHidden = (targetId, room, hidden) => {
    room.targets.find(t => t.id === targetId).hidden = hidden;
};
const HIDE_TARGET = (targetId) => ({
    action: (state, context) => {
        setTargetHidden(targetId ?? context.target.id, context.room, true);
    }
});
const playTextAdventure = async (textAdventure, player) => {
    // game loop
    while (true) {
        if (textAdventure.state.nextText.length > 0) {
            const statusResponse = await status(textAdventure).open({ ...defaultScriptDialogueOptions, player });
            textAdventure.state.nextText = [];
            if (!(statusResponse instanceof ButtonDialogueResponse) || statusResponse.selected === "close") {
                break;
            }
        }
        if (!textAdventure.state.isAlive) {
            break;
        }
        const response = await room(textAdventure).open({ ...defaultScriptDialogueOptions, player });
        if (!(response instanceof ButtonDialogueResponse) || response.selected === "close") {
            break;
        }
        if (response instanceof ButtonDialogueResponse) {
            if (response.selected === "interact") {
                const responseInteract = await subRoomInteract(textAdventure).open({ ...defaultScriptDialogueOptions, player });
                if (responseInteract instanceof ButtonDialogueResponse) {
                    const selected = responseInteract.selected;
                    evaluateInteraction(textAdventure, textAdventure.rooms[textAdventure.state.currentRoom].targets.find(t => t.id === selected));
                }
            }
            else if (response.selected === 'walk') {
                const responseWalk = await subRoomWalk(textAdventure).open({ ...defaultScriptDialogueOptions, player });
                if (responseWalk instanceof ButtonDialogueResponse) {
                    const selected = responseWalk.selected;
                    evaluateWalk(textAdventure, textAdventure.rooms[textAdventure.state.currentRoom].exits.find(e => e.id === selected));
                }
            }
        }
    }
    return textAdventure;
};
const evaluateWalk = (textAdventure, exit) => {
    textAdventure.state.currentRoom = exit.room;
};
const evaluateInteraction = (textAdventure, target) => {
    const interacts = filterByItem(textAdventure, target.interact);
    if (interacts.length > 0) {
        const interact = interacts[0];
        for (const { action } of interact.actions) {
            action(textAdventure.state, {
                room: textAdventure.rooms[textAdventure.state.currentRoom],
                target,
                interact,
                items: textAdventure.items
            });
        }
    }
};

const SOUTH = {
    id: 'south',
    where: TRANSLATE("jig_ccomp:text-adventure.minecraft.exit.south")
};
const NORTH = {
    id: 'north',
    where: TRANSLATE("jig_ccomp:text-adventure.minecraft.exit.north")
};
const EAST = {
    id: 'east',
    where: TRANSLATE("jig_ccomp:text-adventure.minecraft.exit.east")
};
const WEST = {
    id: 'west',
    where: TRANSLATE("jig_ccomp:text-adventure.minecraft.exit.west")
};
const ZOMBIE = {
    id: "zombie",
    description: TRANSLATE("jig_ccomp:text-adventure.minecraft.targets.zombie"),
    interact: [
        {
            // Todo: Kill if you have a stone_sword
            actions: [SAY(TRANSLATE("jig_ccomp:text-adventure.minecraft.targets.zombie.fail")), DIE()]
        }
    ]
};
const CREEPER = {
    id: "creeper",
    description: TRANSLATE("jig_ccomp:text-adventure.minecraft.targets.creeper"),
    interact: [
        {
            // Todo: Kill if you have an iron_sword
            actions: [SAY(TRANSLATE("jig_ccomp:text-adventure.minecraft.targets.creeper.fail")), DIE()]
        }
    ]
};
const SLIME = {
    id: "slime",
    description: TRANSLATE("jig_ccomp:text-adventure.minecraft.targets.slime"),
    interact: [
        {
            // Todo: Kill if you have a stone_sword
            actions: [SAY(TRANSLATE("jig_ccomp:text-adventure.minecraft.targets.slime.fail")), DIE()]
        }
    ]
};
const WITCH = {
    id: "witch",
    description: TRANSLATE("jig_ccomp:text-adventure.minecraft.targets.witch"),
    interact: [
        {
            // Todo: Kill if you have a diamond_sword
            actions: [SAY(TRANSLATE("jig_ccomp:text-adventure.minecraft.targets.witch.fail")), DIE()]
        }
    ]
};
const CRAFTING_TABLE = {
    id: "crafting-table",
    description: TRANSLATE("jig_ccomp:text-adventure.minecraft.targets.crafting-table"),
    interact: [
        {
            hasItems: ["wood"],
            doesNotHaveItem: ["wooden_pickaxe"],
            actions: [GET_ITEM("wooden_pickaxe"), SAY(TRANSLATE("jig_ccomp:text-adventure.minecraft.targets.crafting-table-reward_1"))]
        },
        {
            hasItems: [{
                    id: 'cobblestone',
                    count: 2
                }, {
                    id: 'wood',
                    count: 1
                }],
            doesNotHaveItem: ["stone_sword"],
            actions: [GET_ITEM("stone_sword"), SAY(TRANSLATE("jig_ccomp:text-adventure.minecraft.targets.crafting-table-reward_2"))]
        },
        {
            hasItems: [{
                    id: 'coal',
                    count: 1
                }, {
                    id: 'wood',
                    count: 1
                }],
            doesNotHaveItem: ["torch"],
            actions: [GET_ITEM("torch"), SAY(TRANSLATE("jig_ccomp:text-adventure.minecraft.targets.crafting-table-reward_3"))]
        },
        {
            actions: [SAY(TRANSLATE("jig_ccomp:text-adventure.minecraft.targets.crafting-table-fail"))]
        }
    ]
};
const TREE = {
    id: "tree",
    description: TRANSLATE("jig_ccomp:text-adventure.minecraft.targets.tree"),
    interact: [
        {
            actions: [
                GET_ITEM("wood"), SAY(TRANSLATE("jig_ccomp:text-adventure.minecraft.targets.tree-reward"))
            ]
        }
    ]
};
const STONE = {
    description: TRANSLATE("jig_ccomp:text-adventure.minecraft.targets.stone"),
    interact: [
        {
            hasItems: ["wooden_pickaxe"],
            actions: [
                GET_ITEM("cobblestone"), SAY(TRANSLATE("jig_ccomp:text-adventure.minecraft.targets.stone-reward")), HIDE_TARGET()
            ]
        },
        {
            actions: [SAY(TRANSLATE("jig_ccomp:text-adventure.minecraft.targets.stone-fail"))]
        }
    ]
};
const COAL_ORE = {
    description: TRANSLATE("jig_ccomp:text-adventure.minecraft.targets.coal_ore"),
    interact: [
        { hasItems: ["wooden_pickaxe"],
            actions: [
                GET_ITEM("coal"), SAY(TRANSLATE("jig_ccomp:text-adventure.minecraft.targets.coal_ore-reward")), HIDE_TARGET()
            ]
        },
        {
            actions: [SAY(TRANSLATE("jig_ccomp:text-adventure.minecraft.targets.coal_ore-fail"))]
        }
    ]
};
const IRON_ORE = {
    description: TRANSLATE("jig_ccomp:text-adventure.minecraft.targets.iron_ore"),
    interact: [
        { hasItems: ["stone_pickaxe"],
            actions: [
                GET_ITEM("raw_iron"), SAY(TRANSLATE("jig_ccomp:text-adventure.minecraft.targets.iron_ore-reward")), HIDE_TARGET()
            ]
        },
        {
            actions: [SAY(TRANSLATE("jig_ccomp:text-adventure.minecraft.targets.iron_ore-fail"))]
        }
    ]
};
const newMinecraftAdventure = () => makeTextAdventure({
    initialRoom: 'a1',
    initialText: [
        TRANSLATE("jig_ccomp:text-adventure.minecraft.introduction")
    ],
    inventory: [],
    rooms: {
        a1: {
            description: TRANSLATE("jig_ccomp:text-adventure.minecraft.room.a1.description"),
            distance_description: TRANSLATE("jig_ccomp:text-adventure.minecraft.room.a1.distance_description"),
            targets: [
                {
                    ...TREE
                },
                {
                    ...CREEPER
                },
                {
                    ...ZOMBIE
                }
            ],
            exits: [
                {
                    ...EAST,
                    room: 'a2'
                },
                {
                    ...SOUTH,
                    room: 'b1'
                }
            ]
        },
        a2: {
            description: TRANSLATE("jig_ccomp:text-adventure.minecraft.room.a2.description"),
            distance_description: TRANSLATE("jig_ccomp:text-adventure.minecraft.room.a2.distance_description"),
            targets: [
                {
                    id: 'stone-01',
                    ...STONE
                },
                {
                    id: 'stone-02',
                    ...STONE
                },
                {
                    id: 'stone-03',
                    ...STONE
                },
                {
                    id: 'stone-04',
                    ...STONE
                },
                {
                    id: 'stone-05',
                    ...STONE
                }
            ],
            exits: [
                {
                    ...EAST,
                    room: 'a3'
                },
                {
                    ...SOUTH,
                    room: 'b2'
                },
                {
                    ...WEST,
                    room: 'a1'
                }
            ]
        },
        a3: {
            description: TRANSLATE("jig_ccomp:text-adventure.minecraft.room.a3.description"),
            distance_description: TRANSLATE("jig_ccomp:text-adventure.minecraft.room.a3.distance_description"),
            targets: [
                {
                    id: 'stone-01',
                    ...STONE
                },
                {
                    id: "coal_ore",
                    ...COAL_ORE
                },
                {
                    id: 'iron_ore-1',
                    ...IRON_ORE
                },
                {
                    id: 'stone-02',
                    ...STONE
                },
                {
                    id: 'iron_ore-2',
                    ...IRON_ORE
                },
                {
                    id: 'iron_ore-3',
                    ...IRON_ORE
                },
            ],
            exits: [
                {
                    ...SOUTH,
                    room: 'b3'
                },
                {
                    ...WEST,
                    room: 'a2'
                }
            ]
        },
        b1: {
            description: TRANSLATE("jig_ccomp:text-adventure.minecraft.room.b1.description"),
            distance_description: TRANSLATE("jig_ccomp:text-adventure.minecraft.room.b1.distance_description"),
            targets: [
                {
                    id: "mangrove",
                    description: TRANSLATE("jig_ccomp:text-adventure.minecraft.targets.mangrove"),
                    interact: [
                        {
                            actions: [
                                GET_ITEM("wood"), SAY(TRANSLATE("jig_ccomp:text-adventure.minecraft.targets.mangrove-reward"))
                            ]
                        }
                    ]
                },
                {
                    ...SLIME
                }
            ],
            exits: [
                {
                    ...NORTH,
                    room: 'a1'
                }, {
                    ...EAST,
                    room: 'b2'
                }
            ]
        },
        b2: {
            description: TRANSLATE("jig_ccomp:text-adventure.minecraft.room.b2.description"),
            distance_description: TRANSLATE("jig_ccomp:text-adventure.minecraft.room.b2.distance_description"),
            targets: [
                {
                    ...CRAFTING_TABLE
                },
                {
                    ...WITCH
                }
            ],
            exits: [
                {
                    ...NORTH,
                    room: 'a2'
                }, {
                    ...EAST,
                    room: 'b3',
                    doesNotHaveItem: ["torch"]
                }, {
                    ...EAST,
                    id: "east-torch",
                    room: 'b3_alt',
                    hasItems: ["torch"]
                }, {
                    ...WEST,
                    room: 'b1'
                }
            ]
        },
        b3: {
            description: TRANSLATE("jig_ccomp:text-adventure.minecraft.room.b3.description"),
            distance_description: TRANSLATE("jig_ccomp:text-adventure.minecraft.room.b3.distance_description"),
            targets: [],
            exits: [
                {
                    ...NORTH,
                    room: 'a3'
                }
            ]
        },
        b3_alt: {
            description: TRANSLATE("jig_ccomp:text-adventure.minecraft.room.b3_alt.description"),
            distance_description: TRANSLATE("jig_ccomp:text-adventure.minecraft.room.b3_alt.distance_description"),
            targets: [],
            exits: [
                {
                    ...NORTH,
                    room: 'a3'
                }
            ]
        }
    },
    items: {
        "wood": {
            description: TRANSLATE("jig_ccomp:text-adventure.minecraft.item.wood")
        },
        "wooden_pickaxe": {
            description: TRANSLATE("jig_ccomp:text-adventure.minecraft.item.wooden_pickaxe")
        },
        "stone_pickaxe": {
            description: TRANSLATE("jig_ccomp:text-adventure.minecraft.item.stone_pickaxe")
        },
        "cobblestone": {
            description: TRANSLATE("jig_ccomp:text-adventure.minecraft.item.cobblestone")
        },
        "stone_sword": {
            description: TRANSLATE("jig_ccomp:text-adventure.minecraft.item.stone_sword")
        },
        "iron_sword": {
            description: TRANSLATE("jig_ccomp:text-adventure.minecraft.item.iron_sword")
        },
        "diamond_sword": {
            description: TRANSLATE("jig_ccomp:text-adventure.minecraft.item.diamond_sword")
        },
        "raw_iron": {
            description: TRANSLATE("jig_ccomp:text-adventure.minecraft.item.raw_iron")
        },
        "iron_shovel": {
            description: TRANSLATE("jig_ccomp:text-adventure.minecraft.item.iron_shovel")
        },
        "rotten_flesh": {
            description: TRANSLATE("jig_ccomp:text-adventure.minecraft.item.rotten_flesh")
        },
        "gunpowder": {
            description: TRANSLATE("jig_ccomp:text-adventure.minecraft.item.gunpowder")
        },
        "slime": {
            description: TRANSLATE("jig_ccomp:text-adventure.minecraft.item.slime")
        },
        "coal": {
            description: TRANSLATE("jig_ccomp:text-adventure.minecraft.item.coal")
        },
        "torch": {
            description: TRANSLATE("jig_ccomp:text-adventure.minecraft.item.torch")
        },
    }
});

const viewPcIdDialogue = (current_name, current_player) => {
    return multiButtonScriptDialogue(TRANSLATE("jig_ccomp:pc-id-view.title"))
        .setBody(RAW_TEXT(TRANSLATE("jig_ccomp:pc-id-view.body", "\n\n"), TEXT(`\n\n`), TEXT(`"${current_name}"`), TEXT(`\n\n`), TRANSLATE("jig_ccomp:computer_id_player"), TEXT(`\n\n`), TEXT(`"${current_player}"`)))
        .addButtons([
        {
            name: "close",
            text: TRANSLATE("jig_ccomp:pc-email-read.button.close")
        }
    ]);
};

const viewBugScoreDialogue = (bug_score) => {
    return multiButtonScriptDialogue(TRANSLATE("jig_ccomp:bug-score-view.title"))
        .setBody(RAW_TEXT(TEXT(`§a`), TRANSLATE("jig_ccomp:bug-score-view.body"), TEXT(`§r\n\n`), TEXT(`§q\\(")/              §n\\(")/              §u\\(")/`), TEXT(`\n`), TEXT(`§q-( )-               §n-( )-              §u-( )-`), TEXT(`\n`), TEXT(`§q/(_)\\              §n/(_)\\              §u/(_)\\`), TEXT(`\n`), TEXT(`\n`), TRANSLATE("jig_ccomp:bug-score-view.your-score"), TEXT(`${bug_score}`), TEXT(`\n\n`)
    //,TEXT(`\n\n`), TRANSLATE("jig_ccomp:bug-score-view.high-score"),TEXT(`: ${high_score}`)
    ))
        .addButtons([
        {
            name: "close",
            text: TRANSLATE("jig_ccomp:pc-email-read.button.close")
        }
    ]);
};

const viewComputerOptionsDialogue = (options) => {
    return inputScriptDialogue(TRANSLATE("jig_ccomp:pc-settings.title"))
        .addElement(inputToggle("startup", TRANSLATE("jig_ccomp:pc-settings.startup"), options.startup))
        .addElement(inputToggle("shutdown", TRANSLATE("jig_ccomp:pc-settings.shutdow"), options.shutdown))
        .addElement(inputToggle("notification", TRANSLATE("jig_ccomp:pc-settings.notific"), options.notification))
        .addElement(inputToggle("typing_sound", TRANSLATE("jig_ccomp:pc-settings.typing_sound"), options.typing_sound))
        .addElement(inputToggle("typing_animation", TRANSLATE("jig_ccomp:pc-settings.typing_animation"), options.typing_animation));
};

const CREATIVE_PLAYER_DETECTOR_INTERVAL = 5;
const CREATIVE_PLAYER_ENTITY_OPTIONS = {
    families: ['jig_ccomp:real_computer']
};
const creativePlayerDetector = () => {
    for (const dimension of [MinecraftDimensionTypes.overworld, MinecraftDimensionTypes.nether, MinecraftDimensionTypes.theEnd]) {
        try {
            const computers = world.getDimension(dimension).getEntities(CREATIVE_PLAYER_ENTITY_OPTIONS);
            for (const computer of computers) {
                const isActive = computer.getProperty('jig_ccomp:is_active');
                if (!isActive) {
                    continue;
                }
                const isOn = computer.getProperty('jig_ccomp:is_on');
                if (computer.dimension.getEntities({
                    closest: 1,
                    location: computer.location,
                    maxDistance: 8,
                    type: "minecraft:player",
                    gameMode: GameMode.creative
                }).length > 0) {
                    // We have at least a player in creative mode
                    if (isOn) {
                        computer.triggerEvent('jig:player_nearby_creative');
                    }
                }
                else {
                    if (!isOn) {
                        computer.triggerEvent('jig:players_nearby_not_creative');
                    }
                }
            }
        }
        catch (err) {
            // swallow any error
            debug(`Error while running the creative player detector: ${err}`);
        }
    }
    system.runTimeout(creativePlayerDetector, CREATIVE_PLAYER_DETECTOR_INTERVAL);
};
const initCreativePlayerDetector = () => {
    // Runs about every 5 ticks plus the time it takes to run
    creativePlayerDetector();
};

var MinesweeperOutcome;
(function (MinesweeperOutcome) {
    MinesweeperOutcome["WON"] = "won";
    MinesweeperOutcome["LOST"] = "lost";
    MinesweeperOutcome["PLAYING"] = "playing";
})(MinesweeperOutcome || (MinesweeperOutcome = {}));
var MinesweeperAction;
(function (MinesweeperAction) {
    MinesweeperAction["OPEN"] = "open";
    MinesweeperAction["FLAG"] = "flag";
})(MinesweeperAction || (MinesweeperAction = {}));
class Minesweeper {
    turns;
    /** [row][column]*/
    _board;
    columns;
    rows;
    mines;
    constructor(rows, columns, mines) {
        this.turns = 0;
        this.columns = columns;
        this.rows = rows;
        this._board = [...Array(rows)].map(() => [...Array(columns)].map(() => ({
            isMine: false,
            isOpen: false,
            isFlagged: false,
            minesAround: 0
        })));
        this.mines = mines;
        this.setupMines(mines);
    }
    get board() {
        return this._board;
    }
    playTurn(row, column, action, player) {
        world.scoreboard.getObjective('jig_ccomp:jig_computer.addon_stats')?.addScore('jig_ccomp:tnt_sweeper_turns', 1);
        this.turns++;
        const cell = this._board[row][column];
        if (cell.isOpen) {
            return MinesweeperOutcome.PLAYING;
        }
        if (action === MinesweeperAction.FLAG) {
            cell.isFlagged = !cell.isFlagged;
        }
        else {
            if (cell.isMine) {
                if (!player.hasTag("jig_ccomp:cheat_sweeper") && !player.hasTag("jig_ccomp:really_cheat_sweeper")) {
                    return MinesweeperOutcome.LOST;
                }
                if (player.hasTag("jig_ccomp:cheat_sweeper") && !player.hasTag("jig_ccomp:really_cheat_sweeper")) {
                    cell.isFlagged = !cell.isFlagged;
                }
                if (player.hasTag("jig_ccomp:really_cheat_sweeper")) {
                    return MinesweeperOutcome.WON;
                }
            }
            this.openValid(row, column);
            if (this.playerWon()) {
                return MinesweeperOutcome.WON;
            }
        }
        return MinesweeperOutcome.PLAYING;
    }
    openValid(row, column) {
        if (row >= 0 && row < this.rows && column >= 0 && column < this.columns) {
            const cell = this._board[row][column];
            if (!cell.isOpen && !cell.isFlagged) {
                cell.isOpen = true;
                if (cell.minesAround === 0) {
                    this.openValid(row - 1, column - 1);
                    this.openValid(row - 1, column);
                    this.openValid(row - 1, column + 1);
                    this.openValid(row, column - 1);
                    this.openValid(row, column + 1);
                    this.openValid(row + 1, column - 1);
                    this.openValid(row + 1, column);
                    this.openValid(row + 1, column + 1);
                }
            }
        }
    }
    playerWon() {
        for (const row of this._board) {
            for (const cell of row) {
                if (!cell.isOpen && !cell.isMine) {
                    return false;
                }
            }
        }
        return true;
    }
    setupMines(mines) {
        let minesCreated = 0;
        while (minesCreated < mines) {
            const i = makeRandomNumber(0, this.rows - 1);
            const j = makeRandomNumber(0, this.columns - 1);
            if (!this._board[i][j].isMine) {
                this._board[i][j].isMine = true;
                this.updateMineCountIfAny(i - 1, j - 1);
                this.updateMineCountIfAny(i - 1, j);
                this.updateMineCountIfAny(i - 1, j + 1);
                this.updateMineCountIfAny(i, j - 1);
                this.updateMineCountIfAny(i, j + 1);
                this.updateMineCountIfAny(i + 1, j - 1);
                this.updateMineCountIfAny(i + 1, j);
                this.updateMineCountIfAny(i + 1, j + 1);
                minesCreated++;
            }
        }
    }
    updateMineCountIfAny(row, column) {
        if (row >= 0 && row < this.rows && column >= 0 && column < this.columns) {
            this._board[row][column].minesAround++;
        }
    }
}
const initMineSweeper = () => {
    subscribeFromComputerNpcDialogue(async (event) => {
        let difficulty = "a";
        if (event.id === "jig_ccomp:minesweeper") {
            world.scoreboard.getObjective('jig_ccomp:jig_computer.addon_stats')?.addScore(`jig_ccomp:tnt_sweeper_new`, 1);
            let sizeX = 7;
            let sizeZ = 5;
            let multiplier = 1;
            const multiplierMap = {
                'a': 1,
                'b': 1,
                'c': 1,
                'd': 1.5,
                'e': 1.5,
                'f': 2,
                'g': 2.5,
                'h': 3,
                'i': 4
                // Add more mappings as needed
            };
            if (event.message) {
                // Use a regular expression to match the number and the letter
                const regex = /^(\d+)([a-zA-Z])?$/;
                const match = event.message.match(regex);
                if (match && match[2]) {
                    difficulty = match[2].toLowerCase();
                }
                if (match) {
                    sizeX = parseInt(match[1]);
                    sizeZ = parseInt(match[1]);
                    sizeZ = Math.min(sizeX, 8);
                    // Determine the multiplier based on the letter
                    if (match[2]) {
                        const letter = match[2].toLowerCase();
                        multiplier = multiplierMap[letter] || 1; // Default to sizeX/7 if the letter is not in the map
                    }
                    else {
                        multiplier = 1;
                    }
                }
            }
            await openMineSweeper(event.player, sizeX, sizeZ, multiplier, event.computer, difficulty);
        }
    }, {
        namespaces: ['jig_ccomp']
    });
};
const indexToColumn = (index) => {
    return index + 1;
};
const indexToRow = (index) => {
    return (index + 10).toString(36).toUpperCase();
};
const boardToRawMessage = (minesweeper, printAllBoard) => {
    return [
        TEXT("§8x§0-" + [...Array(minesweeper.board[0].length)].map((_, index) => {
            const column = indexToColumn(index);
            let color;
            if (index >= 8) {
                color = index % 2 === 0 ? "§b" : "§9"; // Use §b for even and §e for odd indices >= 10
            }
            else {
                color = "§b"; // Use §b for indices < 10
            }
            return index >= 8 ? `${color}${column}` : `${color}${column}§0-`;
        }).join("") + "§r\n"),
        ...minesweeper.board.flatMap((row, index) => {
            const rowLabel = indexToRow(index);
            const space = rowLabel === "I" ? " " : "";
            const printRow = [TEXT(`§d${space}${rowLabel}§r`)];
            for (const cell of row) {
                printRow.push(TEXT("§0-"));
                if (!printAllBoard && cell.isFlagged) {
                    printRow.push(TEXT("§aF§r"));
                }
                else if (!printAllBoard && !cell.isOpen) {
                    printRow.push(TEXT("§r-"));
                }
                else if (cell.isMine) {
                    printRow.push(TEXT("§cT§r"));
                }
                else if (cell.minesAround) {
                    printRow.push(TEXT(`§6${cell.minesAround}§r`));
                }
                else {
                    printRow.push(TEXT("§70"));
                }
            }
            printRow.push(TEXT("\n"));
            return printRow;
        })
    ];
};
const newGame = (sizeX, sizeZ, multiplier) => {
    return new Minesweeper(sizeZ, sizeX, Math.ceil((sizeZ * sizeX) * 0.10 * multiplier));
};
const openMineSweeper = async (player, sizeX, sizeZ, multiplier, computer, difficulty) => {
    let minesweeper = newGame(sizeX, sizeZ, multiplier);
    while (true) {
        const board = boardToRawMessage(minesweeper, false);
        let rowDropdown = inputDropdown("row", TRANSLATE("jig_ccomp:minesweeper.row"));
        for (let i = 0; i < minesweeper.rows; ++i) {
            rowDropdown = rowDropdown.addOption(TEXT(`${indexToRow(i)}`), i);
        }
        const response = await inputScriptDialogue(TRANSLATE(`jig_ccomp:minesweeper.title${multiplier}`))
            .addElement(inputDropdown("action", RAW_TEXT(...board))
            .addOption(TRANSLATE("jig_ccomp:minesweeper.action.open"), "open")
            .addOption(TRANSLATE("jig_ccomp:minesweeper.action.flag"), "flag")
            .addOption(TRANSLATE("jig_ccomp:minesweeper.action.quit"), "quit"))
            .addElement(inputSlider("column", TRANSLATE("jig_ccomp:minesweeper.column"), 1, minesweeper.columns, 1, 1))
            .addElement(rowDropdown)
            .open({
            player,
            ...defaultScriptDialogueOptions
        });
        if (response instanceof InputScriptDialogueResponse) {
            let action;
            if (response.values.action === "open") {
                action = MinesweeperAction.OPEN;
            }
            else if (response.values.action === "flag") {
                action = MinesweeperAction.FLAG;
            }
            else {
                break;
            }
            const outcome = minesweeper.playTurn(response.values.row, response.values.column - 1, action, player);
            let score;
            switch (difficulty) {
                case "a":
                    score = "";
                    break;
                case "b":
                    score = "";
                    break;
                case "c":
                    score = "";
                    break;
                case "d":
                    score = "";
                    break;
                case "e":
                    score = "_u";
                    break;
                case "f":
                    score = "_u";
                    break;
                case "g":
                    score = "_u";
                    break;
                case "h":
                    score = "_u";
                    break;
                case "i":
                    score = "_i";
                    break;
                default:
                    score = "";
                    break;
            }
            if (outcome === MinesweeperOutcome.PLAYING) {
                const worldSoundOptions = {
                    pitch: 1.0,
                    volume: 0.5,
                };
                player.playSound("random.pop2", worldSoundOptions);
                continue;
            }
            else {
                if (outcome === MinesweeperOutcome.LOST) {
                    world.scoreboard.getObjective('jig_ccomp:jig_computer.addon_stats')?.addScore(`jig_ccomp:tnt_sweeper${score}_lost`, 1);
                    computer.computer.dimension.spawnParticle("minecraft:large_explosion", computer.computer.location);
                    computer.computer.dimension.spawnParticle("minecraft:lava_particle", computer.computer.location);
                    player.playSound("random.explode");
                }
                else {
                    world.scoreboard.getObjective('jig_ccomp:jig_computer.addon_stats')?.addScore(`jig_ccomp:tnt_sweeper${score}_won`, 1);
                    player.playSound("random.levelup");
                }
                let finishResponse;
                if (outcome === MinesweeperOutcome.WON) {
                    finishResponse = await multiButtonScriptDialogue(TRANSLATE(`jig_ccomp:minesweeper.finish.title${multiplier}`))
                        .setBody(RAW_TEXT(TRANSLATE(`jig_ccomp:minesweeper.finish.body-${outcome}`), TEXT("\n"), ...boardToRawMessage(minesweeper, true)))
                        .addButton("play", TRANSLATE("jig_ccomp:minesweeper.finish.play"))
                        .addButton("print", TRANSLATE("jig_ccomp:minesweeper.finish.print"))
                        .addButton("quit", TRANSLATE("jig_ccomp:minesweeper.finish.quit"))
                        .open({
                        player,
                        ...defaultScriptDialogueOptions
                    });
                }
                else {
                    finishResponse = await multiButtonScriptDialogue(TRANSLATE(`jig_ccomp:minesweeper.finish.title${multiplier}`))
                        .setBody(RAW_TEXT(TRANSLATE(`jig_ccomp:minesweeper.finish.body-${outcome}`), TEXT("\n"), ...boardToRawMessage(minesweeper, true)))
                        .addButton("play", TRANSLATE("jig_ccomp:minesweeper.finish.play"))
                        .addButton("quit", TRANSLATE("jig_ccomp:minesweeper.finish.quit"))
                        .open({
                        player,
                        ...defaultScriptDialogueOptions
                    });
                }
                if (finishResponse instanceof ButtonDialogueResponse) {
                    if (finishResponse.selected === "play") {
                        minesweeper = newGame(sizeX, sizeZ, multiplier);
                        continue;
                    }
                    if (finishResponse.selected === "print") {
                        let paper, message;
                        switch (difficulty) {
                            case "a":
                                paper = "jig_ccomp:printed_paper";
                                message = "But just the demo...";
                                break;
                            case "b":
                                paper = "jig_ccomp:tnt_paper_1";
                                message = "Difficulty: Easy!";
                                break;
                            case "c":
                                paper = "jig_ccomp:tnt_paper_1";
                                message = "Difficulty: Medium!";
                                break;
                            case "d":
                                paper = "jig_ccomp:tnt_paper_1";
                                message = "Difficulty: Hard!";
                                break;
                            case "e":
                                paper = "jig_ccomp:tnt_paper_2";
                                message = "Difficulty: Very Hard!";
                                break;
                            case "f":
                                paper = "jig_ccomp:tnt_paper_2";
                                message = "Difficulty: Extreme!";
                                break;
                            case "g":
                                paper = "jig_ccomp:tnt_paper_2";
                                message = "Difficulty: Ultimate!";
                                break;
                            case "h":
                                paper = "jig_ccomp:tnt_paper_2";
                                message = "Difficulty: Ultimate +!";
                                break;
                            case "i":
                                paper = "jig_ccomp:tnt_paper_3";
                                message = "Difficulty: Impossible!";
                                break;
                            default:
                                paper = "jig_ccomp:tnt_paper_1";
                                message = "Unknown Difficulty!";
                                break;
                        }
                        printContent(player.name + `'s §eTNT Sweeper Award`, `This certifies that you completed TNT Sweeper! ${message}`, computer, player, paper);
                        computer.computer.runCommand("/event entity @s jig_ccomp:typing");
                        break;
                    }
                }
            }
        }
        break;
    }
};

const light_level_update = "jig_ccomp:update_light_level";
const random_print = "jig_ccomp:random_print";
const bug_smash_print = "jig_ccomp:bug_smash_print";
const id_print = "jig_ccomp:id_print";
const trivia_jig_print = "jig_ccomp:trivia_jig_print";
const trivia_min_print = "jig_ccomp:trivia_min_print";
const tnt_print_1 = "jig_ccomp:tnt_print_1";
const tnt_print_2 = "jig_ccomp:tnt_print_2";
const tnt_print_3 = "jig_ccomp:tnt_print_3";
system.afterEvents.scriptEventReceive.subscribe((event) => {
    if (event.id === 'jig:text-adventure-test' && event.sourceEntity instanceof Player) {
        playTextAdventure(newMinecraftAdventure(), event.sourceEntity);
    }
});
world.afterEvents.playerSpawn.subscribe(async (event) => {
    event.player.runCommand("function jig/ccomp/init");
    event.player.runCommand("function jig/ccomp/book_giver");
});
system.afterEvents.scriptEventReceive.subscribe(function (event) {
    if (event.id === random_print) {
        const player = event.sourceEntity;
        const sourceEntity = event.sourceEntity;
        if (player && player instanceof Player && sourceEntity && CraftComputer.isComputer(sourceEntity)) {
            const computer = new CraftComputer(sourceEntity);
            let paper = "jig_ccomp:printed_paper";
            let item_name = "This is a test";
            let item_lore = "This is a lore test, does this work even if it isnt done through the stuff? but also, what determines if its a subject line or the email body";
            printContent(`fg` + item_name, `fdg` + item_lore, computer, player, paper);
        }
    }
    if (event.id === id_print) {
        const player = event.initiator;
        const sourceEntity = event.sourceEntity;
        if (player && player instanceof Player && sourceEntity && CraftComputer.isComputer(sourceEntity)) {
            sourceEntity.id;
            let paper = "jig_ccomp:printed_paper";
            const computer = new CraftComputer(sourceEntity);
            printContent(`Computer ID Printout`, `${computer.getName()}`, computer, player, paper);
        }
    }
    if (event.id === bug_smash_print) {
        const player = event.initiator;
        const sourceEntity = event.sourceEntity;
        if (player && player instanceof Player && sourceEntity && CraftComputer.isComputer(sourceEntity)) {
            let this_game = world.scoreboard.getObjective('jig_ccomp:bug_smash')?.getScore(player) || 0;
            let paper = "jig_ccomp:printed_paper";
            let s = "";
            if (this_game !== 1) {
                s = "s";
            }
            if (this_game < 10) {
                paper = "jig_ccomp:old_bug_paper";
            }
            if (this_game >= 10 && this_game <= 99) {
                paper = "jig_ccomp:ok_bug_paper";
            }
            if (this_game >= 100) {
                paper = "jig_ccomp:good_bug_paper";
            }
            const computer = new CraftComputer(sourceEntity);
            printContent(player.name + `'s §bBug Smash§f Scorecard`, `You §csmashed§5 §l` + this_game + `§r§o§5 bug` + s + `!`, computer, player, paper);
        }
    }
    if (event.id === trivia_jig_print) {
        const player = event.initiator;
        const sourceEntity = event.sourceEntity;
        if (player && player instanceof Player && sourceEntity && CraftComputer.isComputer(sourceEntity)) {
            let paper = "jig_ccomp:trivia_jig_paper";
            const computer = new CraftComputer(sourceEntity);
            printContent(player.name + `'s §aJigarbov Award`, `This certifies that you completed 10 Jigarbov Trivia questions in a row!`, computer, player, paper);
        }
    }
    if (event.id === trivia_min_print) {
        const player = event.initiator;
        const sourceEntity = event.sourceEntity;
        if (player && player instanceof Player && sourceEntity && CraftComputer.isComputer(sourceEntity)) {
            let paper = "jig_ccomp:trivia_min_paper";
            const computer = new CraftComputer(sourceEntity);
            printContent(player.name + `'s §eMinecraft Award`, `This certifies that you completed 10 Minecraft Trivia questions in a row!`, computer, player, paper);
        }
    }
    if (event.id === tnt_print_1) {
        const player = event.initiator;
        const sourceEntity = event.sourceEntity;
        const difficulty = event.message;
        if (player && player instanceof Player && sourceEntity && CraftComputer.isComputer(sourceEntity)) {
            let paper = "jig_ccomp:tnt_paper_1";
            const computer = new CraftComputer(sourceEntity);
            printContent(player.name + `'s §eTNT Sweeper Award`, `This certifies that you completed the ${difficulty} difficulty!`, computer, player, paper);
        }
    }
    if (event.id === tnt_print_2) {
        const player = event.initiator;
        const sourceEntity = event.sourceEntity;
        const difficulty = event.message;
        if (player && player instanceof Player && sourceEntity && CraftComputer.isComputer(sourceEntity)) {
            let paper = "jig_ccomp:tnt_paper_2";
            const computer = new CraftComputer(sourceEntity);
            printContent(player.name + `'s §eTNT Sweeper Ultimate Award`, `This certifies that you completed the ${difficulty} difficulty!`, computer, player, paper);
        }
    }
    if (event.id === tnt_print_3) {
        const player = event.initiator;
        const sourceEntity = event.sourceEntity;
        const difficulty = event.message;
        if (player && player instanceof Player && sourceEntity && CraftComputer.isComputer(sourceEntity)) {
            let paper = "jig_ccomp:tnt_paper_3";
            const computer = new CraftComputer(sourceEntity);
            printContent(player.name + `'s §eTNT Sweeper Impossible Award`, `This certifies that you completed the ${difficulty} difficulty!`, computer, player, paper);
        }
    }
    if (event.id === light_level_update) {
        if (event.sourceEntity instanceof Entity) {
            const round = (value) => {
                return Math.round(value * 100) / 100;
            };
            let showcoordinates = event.sourceEntity.getProperty("jig_ccomp:show_coordinates");
            let showlightlevel = event.sourceEntity.getProperty("jig_ccomp:show_lightlevel");
            let showbiome = event.sourceEntity.getProperty("jig_ccomp:show_biome");
            let showChunkBorder = event.sourceEntity.getProperty("jig_ccomp:show_chunk_border");
            let showSChunkBorder = event.sourceEntity.getProperty("jig_ccomp:show_schunk_border");
            let showEntityParticles = event.sourceEntity.getProperty("jig_ccomp:show_entity_particles");
            let coordinates = event.sourceEntity.location;
            if (showEntityParticles) {
                // Determine target entities based on areaEffect flag
                const targetEntities = event.sourceEntity.dimension.getEntities({
                    location: event.sourceEntity.location,
                    maxDistance: 20,
                    excludeFamilies: ["player"]
                });
                const targetPlayers = event.sourceEntity.dimension.getEntities({
                    location: event.sourceEntity.location,
                    maxDistance: 20,
                    families: ["player"]
                });
                // Apply effects to targeted entities
                for (const entity of targetEntities) {
                    entity.dimension.spawnParticle("jig_ccomp:digital", entity.location);
                }
                // Apply effects to targeted entities
                for (const other of targetPlayers) {
                    other.dimension.spawnParticle("jig_ccomp:digital_red", other.location);
                }
            }
            function isEdgeOfChunkX(x, chunkSize) {
                return Math.abs(x) % chunkSize === 0 || Math.abs(x) % chunkSize === chunkSize - 1;
            }
            function isEdgeOfChunkY(y, chunkSize) {
                return Math.abs(y) % chunkSize === 0 || Math.abs(y) % chunkSize === chunkSize - 1;
            }
            function isEdgeOfChunkZ(z, chunkSize) {
                return Math.abs(z) % chunkSize === 0 || Math.abs(z) % chunkSize === chunkSize - 1;
            }
            const chunkSize = 16;
            let chunk_border_x = isEdgeOfChunkX(Math.trunc(coordinates.x), chunkSize) ? "X" : "";
            let chunk_border_y = isEdgeOfChunkY(Math.trunc(coordinates.y), chunkSize) ? "Yes" : "No";
            let chunk_border_z = isEdgeOfChunkZ(Math.trunc(coordinates.z), chunkSize) ? " Z" : "";
            let chunk_border_no = "";
            if (chunk_border_x === "" && chunk_border_z === "") {
                chunk_border_no = "No";
            }
            let light_level = event.sourceEntity.getProperty("jig_ccomp:light_level");
            let biome = event.sourceEntity.getProperty("jig_ccomp:biome");
            switch (biome) {
                //all of these should be rawtext strings too
                case 0:
                    biome = 'Unknown';
                    break;
                case 1:
                    biome = 'Basalt Delta';
                    break;
                case 2:
                    biome = 'Beach';
                    break;
                case 3:
                    biome = 'Birch';
                    break;
                case 4:
                    biome = 'Caves';
                    break;
                case 5:
                    biome = 'Cherry Grove';
                    break;
                case 6:
                    biome = 'Crimson Forest';
                    break;
                case 7:
                    biome = 'Dark Oak Forest';
                    break;
                case 8:
                    biome = 'Deep Ocean';
                    break;
                case 9:
                    biome = 'Deep Dark';
                    break;
                case 10:
                    biome = 'Desert';
                    break;
                case 11:
                    biome = 'Dripstone Caves';
                    break;
                case 12:
                    biome = 'Edge';
                    break;
                case 13:
                    biome = 'Extreme Hills';
                    break;
                case 14:
                    biome = 'Flower Forest';
                    break;
                case 15:
                    biome = 'Forest';
                    break;
                case 16:
                    biome = 'Frozen';
                    break;
                case 17:
                    biome = 'Frozen Peaks';
                    break;
                case 18:
                    biome = 'Grove';
                    break;
                case 19:
                    biome = 'Hills';
                    break;
                case 20:
                    biome = 'Ice';
                    break;
                case 21:
                    biome = 'Ice Plains';
                    break;
                case 22:
                    biome = 'Jagged Peaks';
                    break;
                case 23:
                    biome = 'Jungle';
                    break;
                case 24:
                    biome = 'Lakes';
                    break;
                case 25:
                    biome = 'Lush Cave';
                    break;
                case 26:
                    biome = 'Mangrove Swamp';
                    break;
                case 27:
                    biome = 'Mega';
                    break;
                case 28:
                    biome = 'Mesa';
                    break;
                case 29:
                    biome = 'Mooshroom Island';
                    break;
                case 30:
                    biome = 'Mountain';
                    break;
                case 31:
                    biome = 'Mutated Biome';
                    break;
                case 32:
                    biome = 'Nether';
                    break;
                case 33:
                    biome = 'Nether Wastes';
                    break;
                case 34:
                    biome = 'Netherwart Forest';
                    break;
                case 35:
                    biome = 'Ocean';
                    break;
                case 36:
                    biome = 'Plains';
                    break;
                case 37:
                    biome = 'Plateau';
                    break;
                case 38:
                    biome = 'River';
                    break;
                case 39:
                    biome = 'Roofed Forest';
                    break;
                case 40:
                    biome = 'Savanna';
                    break;
                case 41:
                    biome = 'Shore';
                    break;
                case 42:
                    biome = 'Snowy Slopes';
                    break;
                case 43:
                    biome = 'Soulsand Valley';
                    break;
                case 44:
                    biome = 'Stone Beach';
                    break;
                case 45:
                    biome = 'Swamp';
                    break;
                case 46:
                    biome = 'Taiga';
                    break;
                case 47:
                    biome = 'The End';
                    break;
                case 48:
                    biome = 'Warped Forest';
                    break;
            }
            let key = "";
            if (showcoordinates) {
                key = key + `x: ${round(coordinates.x)}, y: ${round(coordinates.y)}, z: ${round(coordinates.z)}`;
            }
            if (showbiome) {
                key = key + "\nBiome: " + biome;
            }
            if (showlightlevel) {
                key = key + "\nLight Level: " + light_level;
            }
            if (showChunkBorder) {
                key = key + "\nCB: " + chunk_border_x + chunk_border_z + chunk_border_no;
            }
            if (showSChunkBorder) {
                key = key + "\nSubCB: " + chunk_border_y;
            }
            event.sourceEntity.nameTag = key;
        }
    }
});
// Split initialization
system.runTimeout(() => {
    initWorldInfo();
    initRadars();
    initWeather();
    initEmail();
    initPortable();
}, 1);
system.runTimeout(() => {
    initCreativeProtect();
    initTut_Book();
    initCreativePlayerDetector();
    initComputers();
    initForum();
    initAdminConfigMenu();
}, 1);
system.runTimeout(() => {
    initMineSweeper();
}, 1);
world.afterEvents.entitySpawn.subscribe((event) => {
    craftComputers.addComputer(event.entity);
});
world.afterEvents.entityLoad.subscribe((event) => {
    craftComputers.addComputer(event.entity);
});
world.beforeEvents.entityRemove.subscribe((event) => {
    craftComputers.removeComputer(event.removedEntity);
});
const viewBugScore = async (player, computer) => {
    computer.computer.triggerEvent('jig_ccomp:typing');
    let this_game = world.scoreboard.getObjective('jig_ccomp:bug_smash')?.getScore(player) || 0;
    await viewBugScoreDialogue(this_game).open({ ...defaultScriptDialogueOptions, player });
    computer.computer.runCommand(`/dialogue open @s "${player.name}" jig_ccomp:bug_smash_miss`);
};
const viewComputerOptions = async (player, computer) => {
    computer.computer.triggerEvent('jig_ccomp:typing');
    const initial_startup = computer.computer.hasTag("jig_ccomp.mute_startup") || false;
    const initial_shutdown = computer.computer.hasTag("jig_ccomp.mute_shutdown") || false;
    const initial_notification = computer.computer.hasTag("jig_ccomp.mute_notification") || false;
    const initial_typing_sound = computer.computer.hasTag("jig_ccomp.mute_typing_sound") || false;
    const initial_typing_animation = computer.computer.getProperty("jig_ccomp:mute_keyboard_jiggle") || false;
    player.playSound('jig_common.beeps.say');
    const response = await viewComputerOptionsDialogue({
        startup: !initial_startup,
        shutdown: !initial_shutdown,
        notification: !initial_notification,
        typing_sound: !initial_typing_sound,
        typing_animation: !initial_typing_animation
    }).open({ ...defaultScriptDialogueOptions, player });
    if (response instanceof InputScriptDialogueResponse) {
        player.playSound('jig_common.beeps.say');
        const { startup, shutdown, notification, typing_sound, typing_animation } = response.values;
        if (startup === false) {
            computer.computer.addTag("jig_ccomp.mute_startup");
        }
        else {
            computer.computer.removeTag("jig_ccomp.mute_startup");
        }
        if (shutdown === false) {
            computer.computer.addTag("jig_ccomp.mute_shutdown");
        }
        else {
            computer.computer.removeTag("jig_ccomp.mute_shutdown");
        }
        if (notification === false) {
            computer.computer.addTag("jig_ccomp.mute_notification");
        }
        else {
            computer.computer.removeTag("jig_ccomp.mute_notification");
        }
        if (typing_sound === false) {
            computer.computer.addTag("jig_ccomp.mute_typing_sound");
        }
        else {
            computer.computer.removeTag("jig_ccomp.mute_typing_sound");
        }
        if (typing_animation === false) {
            computer.computer.setProperty("jig_ccomp:mute_keyboard_jiggle", true);
        }
        else {
            computer.computer.setProperty("jig_ccomp:mute_keyboard_jiggle", false);
        }
    }
    computer.computer.runCommand(`/dialogue open @s "${player.name}" jig_ccomp:old_pc_pcid`);
};
const viewComputerId = async (player, computer) => {
    computer.computer.triggerEvent('jig_ccomp:typing');
    if (!computer.getCreator()) {
        computer.setCreator(player.name);
    }
    await viewPcIdDialogue(computer.getName(), computer.getCreator() ?? "undefined").open({ ...defaultScriptDialogueOptions, player });
    computer.computer.runCommand(`/dialogue open @s "${player.name}" jig_ccomp:old_ac_pcid`);
};
const setComputerId = async (player, computer) => {
    computer.computer.triggerEvent('jig_ccomp:typing');
    const response = await setPcIdDialogue(computer.getName()).open({ ...defaultScriptDialogueOptions, player });
    if (response instanceof InputScriptDialogueResponse) {
        const name = response.values['name'];
        if (name.length > 0) {
            debug(`Setting PC name to ${name} for computer ${computer.getId()}`);
            computer.setName(name);
            computer.computer.triggerEvent('jig_ccomp:typing');
            await viewPcIdDialogue(computer.getName(), computer.getCreator() ?? "undefined").open({ ...defaultScriptDialogueOptions, player });
            computer.computer.runCommand(`/dialogue open @s "${player.name}" jig_ccomp:old_ac_pcid`);
        }
    }
};
// Process dialogue scriptevents
system.afterEvents.scriptEventReceive.subscribe(async (event) => {
    const player = event.initiator;
    const sourceEntity = event.sourceEntity;
    if (player && player instanceof Player && sourceEntity && CraftComputer.isComputer(sourceEntity)) {
        const computer = new CraftComputer(sourceEntity);
        if (event.id === 'jig_ccomp:view_pc_id') {
            executeScriptDialogueFromNpcDialogue(() => viewComputerId(player, computer));
        }
        else if (event.id === 'jig_ccomp:set_pc_id') {
            executeScriptDialogueFromNpcDialogue(() => setComputerId(player, computer));
        }
        else if (event.id === 'jig_ccomp:bug_smash_view') {
            executeScriptDialogueFromNpcDialogue(() => viewBugScore(player, computer));
        }
        else if (event.id === 'jig_ccomp:computer_options') {
            executeScriptDialogueFromNpcDialogue(() => viewComputerOptions(player, computer));
        }
    }
});
system.afterEvents.scriptEventReceive.subscribe((event) => {
    if (event.sourceEntity && event.initiator) {
        if (event.id.endsWith('trivia_wrong')) {
            const triviaPrizes = isAdminSettingsEnabled('triviaPrizes');
            const nextNpcDialogue = triviaPrizes ? event.id : event.id + '_x';
            event.sourceEntity.runCommand(`dialogue open @s "${event.initiator.nameTag}" ${nextNpcDialogue}`);
        }
        else if (event.id.endsWith('_trivia_next')) {
            // Check if we wont or we want to display the next question
            const triviaId = event.id.slice(0, event.id.length - '_next'.length);
            if (getScoreOr(triviaId, event.initiator, 0) >= 10) {
                // won!
                const wonBase = triviaId + '_won';
                const triviaPrizes = isAdminSettingsEnabled('triviaPrizes');
                const nextNpcDialogue = triviaPrizes ? wonBase : wonBase + '_x';
                event.sourceEntity.runCommand(`dialogue open @s "${event.initiator.nameTag}" ${nextNpcDialogue}`);
            }
            else {
                // Assumes all trivia have 50 questions - This can be updated to be different per trivia
                const nextTrivia = makeRandomNumber(1, 50);
                const nextTriviaString = nextTrivia < 10 ? `0${nextTrivia}` : `${nextTrivia}`;
                event.sourceEntity.runCommand(`dialogue open @s "${event.initiator.nameTag}" ${triviaId}_${nextTriviaString}`);
            }
        }
    }
}, {
    namespaces: ['jig_ccomp']
});
const initDimension = (dimension) => {
    try {
        craftComputers.init(world.getDimension(dimension));
    }
    catch (e) {
        debug(`Error initializing computers on dimension ${dimension}: ${e}`);
    }
};
system.runTimeout(() => {
    initDimension("overworld");
    initDimension("nether");
    initDimension("the_end");
    world.afterEvents.playerSpawn.subscribe((event) => {
        if (event.initialSpawn && event.player.isValid()) {
            runPendingOperations(event.player);
        }
    });
    world.afterEvents.entityLoad.subscribe((event) => {
        if (event.entity.isValid()) {
            runPendingOperations(event.entity);
        }
    });
    craftComputers.getComputers().forEach(c => {
        runPendingOperations(c.computer);
    });
    world.getPlayers().filter(p => p.isValid()).forEach(p => {
        runPendingOperations(p);
    });
}, 50);
// Init world
if (GET_RAW_PROP(world, Config.world.installedTime) === undefined) {
    SET_RAW_PROP(world, Config.world.installedTime, world.getAbsoluteTime());
}
//# sourceMappingURL=main.js.map
